
/* ----------------------------------------------------------------------
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic
 
  Purpose: GebPP - a lite PPP client for post-procesing positioning
 
-*/

#include <cstring>
#include <iostream>
#include <iomanip>
#include <signal.h>
#include <functional>
#include <numeric>

#include "gset/gcfg_gebpp.h"

using namespace std;

static const char pgm[] = "geb-pp";
static const char ver[] = "0.9.0";
static const char rev[] = "$Rev: 417 $";
static const char dat[] = __DATE__;
static const char tim[] = __TIME__;
static const char own[] = "(gnss@pecny.cz)";
void catch_signal(int){ cout << "Program interrupted by Ctrl-C [SIGINT,2]\n"; exit(1); }

void usage(){
   
  cout << pgm  << " v[" << ver << "]"
       << endl << " - compiled: " << dat << " " << tim << " "  << own << " " << rev
       << endl
       << endl << "Usage: "
       << endl
       << endl << "    -h                     .. this help                          "
       << endl << "    -v int                 .. verbosity                          "
       << endl << "    -x xml-file            .. configuration                      "
       << endl << endl;
  exit(0);
}

/* -----
 * MAIN
-*/
int main(int argc, char** argv)
{
  clock_t clkt;
  signal(SIGINT,  catch_signal);

  t_gcfg_gebpp gset;
  t_glog glog;
    glog.mask("/dev/stdout");  glog.verb(0);  glog.cache_size(1); // single line cached
   
  string conf    =    "";  // xml configuration file

  // COMMAND-LINE ARGS
  for( int i=1; i<argc; i++ ){
    if (  !strcmp(argv[i],"--help") && i  <argc ){ usage(); exit(0); }
    else if ( !strcmp(argv[i],"-h") && i  <argc ){ usage(); exit(0); }
    else if ( !strcmp(argv[i],"-X") && i  <argc ){ gset.help(); exit(0); }
    else if ( !strcmp(argv[i],"-v") && i+1<argc ){ glog.verb(atoi(argv[++i])); }
    else if ( !strcmp(argv[i],"-x") && i+1<argc ){ conf    = argv[++i]; }

    else fprintf(stderr,"Unknown option: %s\n",argv[i]);
  }

  // READ XML SETTINGS
  if( conf.empty() ){ if( gset.read_istream(cin) ){ gset.help(); usage(); exit(1); }}
  else{               if( gset.read(conf)        ){ gset.help(); usage(); exit(1); }}

  gset.glog( &glog );
  string site = gset.strval("gen","rec");
  int sample  = dynamic_cast<t_gsetgen*>(&gset)->sampling();
  multimap<IFMT,string> inp = gset.inputs_all();   

  // COMMENTS
  glog.comment(1,"main","program = " + string(pgm));
  glog.comment(1,"main","version = " + string(ver));
  glog.comment(1,"main","setting = " + conf);
  glog.comment(1,"main","logfile = " + glog.mask());
  glog.comment(1,"main","verbose = " + int2str(glog.verb()));
  glog.comment(1,"main","sample  = " + int2str(sample));

  t_gallobs*    gobs  = 0;
  t_gallpcv*    gpcv  = 0;
  t_gpppfilter* gppp  = 0;
  t_gpreproc*   gpre  = 0;
  t_gallnav*    gorb  = 0;
  t_gdata*      gdata = 0;

  // DATA/PRODUCT SETUP
  if( gset.input_size("sp3")    > 0 &&
      gset.input_size("rinexc") > 0 &&
      gset.input_size("rinexo") > 0
  ){
     gorb = new t_gallprec;
     gobs = new t_gallobs;
  }else{  
     glog.comment(0,"main", "Error: missing input product or offline: sp3+clk+obs");
     glog.comment(0,"main", "Usage: ./" + string(pgm) + " --help");
     usage();
  }
   
  if( gset.input_size("atx") > 0 ) gpcv = new t_gallpcv;
       
  if( gpcv ) gpcv->glog(&glog);
  if( gorb ) gorb->glog(&glog);
  if( gobs ) gobs->glog(&glog);
   
  // DATA READING
  multimap<IFMT,string>::const_iterator itINP = inp.begin();
  for( size_t i = 0; i < inp.size() && itINP != inp.end(); ++itINP )
  {
    IFMT   ifmt(itINP->first);
    string path(itINP->second);
    t_gio*    gio    = 0;
    t_gcoder* gcoder = 0;

    // DEFINE DECODER
    if(       ifmt == SP3_INP     ){ gdata = gorb; gcoder = new t_sp3("",8172);
    }else if( ifmt == RINEXO_INP  ){ gdata = gobs; gcoder = new t_rinexo("",4096);
    }else if( ifmt == RINEXC_INP  ){ gdata = gorb; gcoder = new t_rinexc("",4096);
    }else if( ifmt == ATX_INP     ){ gdata = gpcv; gcoder = new t_atx("",4096);
    }else{ 
      glog.comment(0,"main","Error: unrecognized format " + int2str(ifmt) );
      gdata = 0;
    }

    if( path.substr(0,7) != "file://"  ){
      glog.comment(0,"main","path not file (skipped)!");
      continue;
    }
     
    // READ DATA FROM FILE
    if( gcoder ){

      gio = new t_gfile;
      gio->path(path);
    
      gcoder->glog(&glog);
      gcoder->gset(gset);
     
      gio->coder( gcoder );
      gio->add_data( "ID", gdata );
 	 
      clkt = clock();
      gio->run_read();
      clkt = clock() - clkt;
      glog.comment(0, "main", "READ: " + path + " time: " + dbl2str(clkt/1e6,1) + " sec");

      delete gio;
      delete gcoder;
    }
  }

  // OBS CONTENT LOG
  glog.comment(0, "main", "OBS: " + site
                                  + gobs->beg_obs(site).str_ymdhms(" beg: ")
                                  + gobs->end_obs(site).str_ymdhms(" end: ")
                                  + " #epo: " + dbl2str( gobs->nepochs(site) ));

  // PPP PROCESSING
  if( gobs->beg_obs(site) == t_gtime::last_time(t_gtime::GPS)  ||
      gobs->end_obs(site) == t_gtime::first_time(t_gtime::GPS) ||
      site.empty()
  ){
    glog.comment(0, "main", "No site/data for processing!");
  }else{
   
    glog.comment(0, "main", site + " PP-filter started ");
    gpre = new t_gpreproc(gobs, &gset);      gpre->glog(&glog); 
    gppp = new t_gpppfilter(site, &gset);    // glog set special glog
    gppp->setDAT(gobs, gorb);
    gppp->setPCV(gpcv);
     
    clkt = clock();
    gpre->preprocess(site, gobs->beg_obs(site), gobs->end_obs(site));
    gppp->processBatch(gobs->beg_obs(site), gobs->end_obs(site), sample);
    clkt = clock() - clkt;
   
    glog.comment(0, "main", site + " PP-filter ended: " + dbl2str(clkt/1e6,1) + " sec");
  }

  if( gobs ) delete gobs;
  if( gppp ) delete gppp;
  if( gpre ) delete gpre;
  if( gorb ) delete gorb;
  if( gpcv ) delete gpcv;  

}
