
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <stdlib.h>
#include <sstream>
#include <iomanip>

#include "gall/gallnav.h" 
 
using namespace std;


// constructor
// ----------
t_gallnav::t_gallnav() 
  : t_gdata(),
    _offset(0),
    _nepoch(t_gtime::GPS),
    _overwrite(false)
{
}


// destructor
// ----------
t_gallnav::~t_gallnav()
{
  map<string,t_map_ref>::iterator itsat;
  for( itsat = _mapsat.begin(); itsat != _mapsat.end(); ++itsat ){
    string sat = itsat->first;
    map<t_gtime,t_geph*>::iterator it;
    for( it = _mapsat[sat].begin(); it != _mapsat[sat].end(); ++it ) delete it->second;
    _mapsat[sat].clear();
  }
  _mapsat.clear();
}


// return position
// ----------
int t_gallnav::pos( string sat, const t_gtime& t, double  xyz[],
		                                  double  var[],
		                                  double  vel[] ) // [m]
{
  boost::mutex::scoped_lock lock(_mutex);
  t_geph* tmp = _find( sat, t );
  if( tmp == 0 ){	
    for(int i = 0; i<3; i++){
	                     xyz[i] = 0.0;
	           if( var ) var[i] = 0.0;
	           if( vel ) vel[i] = 0.0;
     }
     return -1;
   }     
   return tmp->pos( t, xyz, var, vel );
   
//  cout << "GALLNAV JSEM TU :: " << sat << endl; cout.flush();
//  return find( sat, t )->pos( t, xyz, var, vel );
}


// return clock corrections
// ----------
int t_gallnav::clk( string sat, const t_gtime& t, double*  clk,
		                                  double*  var,
		                                  double* dclk ) // [s]
{
  boost::mutex::scoped_lock lock(_mutex);
  t_geph* tmp = _find( sat, t );   
  if( tmp == 0 ){	
              *clk  = 0.0;
    if( var ) *var  = 0.0;
    if(dclk ) *dclk = 0.0;
    return -1;
   }     
   return tmp->clk( t, clk, var, dclk );
    
//  return this->find( sat, t )->clk( t, clk, var, dclk );
}


// return list of available satellites
// ----------
vector<string> t_gallnav::satellites()
{
  boost::mutex::scoped_lock lock(_mutex);
  vector<string> all_sat;
  map<string,t_map_ref>::const_iterator itPRN = _mapsat.begin();
   
  while( itPRN != _mapsat.end() ){
    all_sat.push_back(itPRN->first);
    itPRN++;
  }
  return all_sat;
}


// return first position for satellite
// ----------
t_gtime t_gallnav::beg_gnav( string prn )
{
  boost::mutex::scoped_lock lock(_mutex);

  if( _mapsat.find(prn) != _mapsat.end() )
    return _mapsat[prn].begin()->first;

  return t_gtime::last_time();
}


// return last position for satellite
// ----------
t_gtime t_gallnav::end_gnav( string prn )
{
  boost::mutex::scoped_lock lock(_mutex);

  if( _mapsat.find(prn) != _mapsat.end() )
    return _mapsat[prn].rbegin()->first;

  return t_gtime::first_time();
}


// add navigation message
// ----------
int t_gallnav::add( const t_gnav* nav )
{   
  if( nav->chk() < 0 ) return 1;
   
  boost::mutex::scoped_lock lock(_mutex);

  t_gtime ep(nav->epoch());
  string sat = nav->sat();
   
  if( _overwrite || _mapsat[sat].find(ep) == _mapsat[sat].end() ){

    // distinguish GNSS specific message
    if(       nav->id_type() == t_gdata::EPHGPS ){
       _mapsat[sat][ep] = new t_gnavgps( *(t_gnavgps*)(nav) );  // allocate in heap

    }else if( nav->id_type() == t_gdata::EPHGLO ){
//       _mapsat[sat][ep] = new t_gnavglo( *(t_gnavgps*)(nav) );
       // cout << " NASEL JSEM GLO" << nav->str_type() << "\n";  // allocate in heap
    }else{
       if( _log && _log->verb() >= 2 ){  _log->comment(2,"gallnav","warning: t_gnav record not identified!"); }
       return 1;
    }
  }
  if( _log && _log->verb() >= 3 ){
    ostringstream lg;
    lg << "add sat [" << nav->str_type()
       << "]: " << sat
       << " " << ep.str("%Y-%m-%d %H:%M:%S")
       << " iod: " << fixed << setw(3) <<  nav->iod();
    _log->comment(3,"gallnav",lg.str());
  }
 
    // problem of various compiler implementatino !!!!!
//    if( string(typeid(*nav).name()) == "9t_gnavglo" ){
//     cout << " NASEL JSEM GPS " << typeid(*nav).name()  << "\n";
//    }else{
//     cout << " NENASEL JSEM GPS" << typeid(*nav).name() << "\n";
//     }

  return 0;
}


// return number of epochs
// ----------
unsigned int t_gallnav::nepochs( const string& prn )
{
  boost::mutex::scoped_lock lock(_mutex);

  if( _mapsat.find(prn) != _mapsat.end() )
    return _mapsat[prn].size();
   
  return 0;
}



// clean function
// ----------
void t_gallnav::clean_outer( const t_gtime& beg, const t_gtime& end )
{
  if( end < beg ) return;
  boost::mutex::scoped_lock lock(_mutex);

#ifdef DEBUG   
  if( _log ) _log->comment(1,"gallnav","rtcm clean request: "
			              + beg.str("%Y-%m-%d %H:%M:%S[%T] - ")
                                      + end.str("%Y-%m-%d %H:%M:%S[%T]"));
#endif

  // loop over all satellites
  map<string,t_map_ref>::const_iterator itPRN = _mapsat.begin();
  while( itPRN != _mapsat.end() ){
    string prn = itPRN->first;

    // find and CLEAN all data (epochs) out of the specified period !
    map<t_gtime,t_geph*>::iterator it;
    map<t_gtime,t_geph*>::iterator itFirst = _mapsat[prn].begin();
    map<t_gtime,t_geph*>::iterator itLast  = _mapsat[prn].end();
    map<t_gtime,t_geph*>::iterator itBeg   = _mapsat[prn].lower_bound(beg);  // greater only   old: // greater|equal
    map<t_gtime,t_geph*>::iterator itEnd   = _mapsat[prn].upper_bound(end);  // greater only!
     
//  cout << "distance = " << distance(itBeg,itEnd) << " prn " << prn << endl;

    // delete before BEGIN request
    if( itBeg != itFirst ){
       
      // begin is last
      if( itBeg == itLast ){
        itBeg--;
        if( _log && _log->verb() >= 2 ) 
	    _log->comment(2,"gallnav","gnav removed before: " + itFirst->first.str("%Y-%m-%d %H:%M:%S[%T] ") + prn);
        it = itFirst;
	while( it != itLast ) // check for npointers() > 0 !!

	  if( (it->second)->npointers() == 0){ delete it->second;
                                              _mapsat[prn].erase(it++);
	  }else{ it++; }

      // begin is not last
      }else{
        if( _log && _log->verb() >=2 )
	    _log->comment(2,"gallnav","gnav removed before: " + itBeg->first.str("%Y-%m-%d %H:%M:%S[%T] ") + prn);
        it = itFirst;
	while( it != itBeg ) // check for npointers() > 0 !!
	  if( (it->second)->npointers() == 0){ delete it->second;
                                              _mapsat[prn].erase(it++);
	  }else{ it++; }
	
      }
    }

    // delete after END request
    if( itEnd != itLast ){
        if( _log && _log->verb() >= 2 ) 
	    _log->comment(2,"gallnav","gnav removed after : " + itEnd->first.str("%Y-%m-%d %H:%M:%S[%T] ") + prn );
        it = itEnd;
	while( it != itLast ) // check for npointers() > 0 !!
	  if( (it->second)->npointers() == 0){ delete it->second;
                                              _mapsat[prn].erase(it++);
	  }else{ it++; }

    }
    itPRN++;
  }
  lock.unlock();
}


// return list of available satellites
// ----------
t_geph* t_gallnav::_find( string sat, const t_gtime& t ){

  if( _mapsat.find(sat) == _mapsat.end() ) return 0;
  
  map<t_gtime,t_geph*>::iterator it = _mapsat[sat].lower_bound(t);  // greater|equal
  if( it == _mapsat[sat].end() ){
    if( _log && _log->verb() >= 3 ){
      string lg("gnav element not found: [" + sat + t.str("] %Y-%m-%d %H:%M:%S "));
      _log->comment(3,"gallnav",lg);
    }       
    return 0;
  }

  return it->second;
}

// return list of available satellites
// ----------
t_geph* t_gallnav::_find( string sat, int iod ){

  if( _mapsat.find(sat) == _mapsat.end() ) return 0;

  map<t_gtime,t_geph*>::const_iterator it = _mapsat[sat].begin();
  while( it != _mapsat[sat].end() ){
    t_gnav* pt_nav;
    pt_nav = 0;
    if( ( pt_nav = dynamic_cast<t_gnav*>(it->second) ) != NULL ){
      if( pt_nav->iod() == iod ) break; // found
    }
    it++;
  }

  // not found !
  if( it == _mapsat[sat].end() ){
    if( _log && _log->verb() >= 3 ){
      ostringstream lg;
      lg << "warning: gnav element not found "
	 << "[sat: " << sat
	 << " iod: " << iod << "]";
      _log->comment(3,"gallnav",lg.str());
    }
    return 0;
  }

  return it->second;
}

