
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <stdlib.h>
#include <sstream>

#include "gall/gallpcv.h"
 
using namespace std;

// constructor
// ----------
t_gallpcv::t_gallpcv() 
  : t_gdata(),
    _overwrite(false)
{
}


// destructor
// ----------
t_gallpcv::~t_gallpcv()
{

  map<string,t_map_num>::iterator itant  = _mappcv.begin();   
  for( itant = _mappcv.begin(); itant != _mappcv.end(); ++itant ){

    string ant = itant->first;

    map<string,t_map_epo>::iterator itnum;
    for( itnum = _mappcv[ant].begin(); itnum != _mappcv[ant].end(); ++itnum ){

      string num = itnum->first;
         
      map<t_gtime,t_gpcv*>::iterator it;
      for( it = _mappcv[ant][num].begin(); it != _mappcv[ant][num].end(); ++it ){
	   
        delete it->second;
      }
    }
  }
}


// find PCO+PCV model
// ----------
t_gpcv* t_gallpcv::find( string ant, string ser, const t_gtime& t )
{
  boost::mutex::scoped_lock lock(_mutex);
  return _find( ant, ser, t );
}

// get PCO (phase center offsets)
// zen/azi [rad]
// ----------
double t_gallpcv::pco( string ant, string ser, const t_gtime& t, double zen, double azi, GFRQ freq)
{
  boost::mutex::scoped_lock lock(_mutex);
  t_gpcv* tmp = _find( ant, ser, t );
  if( tmp == 0 ){	
     return 0.0;
   }
   return tmp->pco( zen, azi, freq );
}

  
// get PCV (phase center variations)
// zen/azi [rad]
// ----------
double t_gallpcv::pcv( string ant, string ser, const t_gtime& t, double zen, double azi, GFRQ freq )
{
  boost::mutex::scoped_lock lock(_mutex);
  t_gpcv* tmp = _find( ant, ser, t );
  if( tmp == 0 ){	
     return 0.0;
   }
   return tmp->pcv( zen, azi, freq );
}


// return position (virtual)
// ----------
int t_gallpcv::addpcv( const t_gpcv& pcv )
{
  boost::mutex::scoped_lock lock(_mutex);
  string  ant = pcv.antenn();
  string  rad = pcv.radome(); // NOT USED (included in antenna !)
  string  num = pcv.sernum();
  t_gtime beg = pcv.beg();
  t_gtime end = pcv.end();
   
  if( _mappcv.find(ant)           == _mappcv.end() ||
      _mappcv[ant].find(num)      == _mappcv[ant].end() ||
      _mappcv[ant][num].find(beg) == _mappcv[ant][num].end() ){

     // new instance
     _mappcv[ant][num][beg] = new t_gpcv(pcv);
#ifdef DEBUG
     cout << "NEW: " << ant << " " << num << " " << endl;
#endif
  }else{
      
    // exists, but overwrite ?
    if( _overwrite ){
#ifdef DEBUG
     cout << "DEL: " << ant << " " << num << " " << endl;
#endif       
     delete _mappcv[ant][num][beg];
     _mappcv[ant][num][beg] = new t_gpcv(pcv);
#ifdef DEBUG       
     cout << "RENEW: " << ant << " " << num << " " << endl;
#endif
    }else{
      if( _log && _log->verb() >= 1 ) _log->comment(1, "gallpcv", "already exists, not overwritten !\n" );
#ifdef DEBUG       
      cout << "OLD: " << ant << " " << num << " " << endl;
#endif
    }
  }

  if( _log && _log->verb() >= 2 ) 
      _log->comment(2,"gallpcv","add PCV " + ant + " " + num 
	                                   + beg.str_ymdhms(" ") + end.str_ymdhms(" "));
   
  return 1;
}


// return list of available antennas
// ----------
vector<string> t_gallpcv::antennas()
{ 
  boost::mutex::scoped_lock lock(_mutex);   
  vector<string> all_ant;
     
  map<string,t_map_num>::const_iterator itant = _mappcv.begin();   
//  if( itant == _mappcv.end() ){
//     return all_ant;
//  }
      
  for( itant = _mappcv.begin(); itant != _mappcv.end(); ++itant ){
  
    string ant = itant->first;

    map<string,t_map_epo>::iterator itnum;
    for( itnum = _mappcv[ant].begin(); itnum != _mappcv[ant].end(); ++itnum ){

      string num = itnum->first;
     
      all_ant.push_back(ant + " " + num);
    }
  }
  return all_ant;
}


// return list of available satellites
// ----------
t_gpcv* t_gallpcv::_find( string ant, string ser, const t_gtime& t )
{
  string ant0 = ant;
  string ser0 = "";
   
  // antenna not found in the list
  if( _mappcv.find(ant) == _mappcv.end() ){
     
    // try alternative name for receiver antenna
    if( ant.size() >= 19 )  ant0 = ant.replace(16,4,"NONE");

    if( _mappcv.find(ant0) == _mappcv.end() ){
      if( _log && _log->verb() >= 2 ) _log->comment(2,"gallpcv","pcv not found [" + ant + "]");
      return 0;
    }
  }
   
  // individual antenna
  map<string,t_map_epo>::iterator itser;
  if( ser.find("*") == string::npos ){
    for( itser = _mappcv[ant0].begin(); itser != _mappcv[ant0].end(); ++itser ){
      
      // antenna not found in the list
      if( _mappcv[ant0].find(ser) != _mappcv[ant0].end() ){     
	ser0 = ser;
        break;
      }
    }
  }
   
  // try to find the approximation 
  if( ser0 == "" && _mappcv[ant].find(ser) == _mappcv[ant].end() ){
     ser = "*";
     if( _log && _log->verb() >= 5 ){
        _log->comment(5,"gallpcv","pcv not found [" + ant + "," + ser + "]");
     }
  }

  // check/return individual and check time validity
  if( ser0 == ser ){
     
    map<t_gtime,t_gpcv*>::iterator it = _mappcv[ant][ser0].begin();
    while( it != _mappcv[ant][ser0].end() ){	    
      if(     (it->second)->beg() < t  &&
   	  t < (it->second)->end()      ){
	return it->second;	      
      }
      it++;
    }
  }

  // find type-specific and check time validity
  if( ser.find("*") != string::npos  ){
    for( itser = _mappcv[ant].begin(); itser != _mappcv[ant].end(); ++itser ){
       
      ser0 = itser->first;
 
      map<t_gtime,t_gpcv*>::iterator it = _mappcv[ant][ser0].begin();
      while( it != _mappcv[ant][ser0].end() ){	    
        if(     (it->second)->beg() < t  &&
   	    t < (it->second)->end()      ){
	  return it->second;	      
	}
	it++;
      }
    }
  }
  
  if( _log && _log->verb() >= 1 ) _log->comment(1,"gallpcv","pcv not found, even no alternative! [" + ant0 + "]");
  return 0;
}
