
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <stdlib.h>
#include <cstring>
#include <iostream>

#include "gcoders/gcoder.h" 
#include "gutils/gtypeconv.h"
 
using namespace std;

// constructor
// ----------
t_gcoder::t_gcoder( string version, int sz )
{
  _initver = version;
  _init();
   
  // after init!
//  _version = version; 
  _buffsz  = sz;
  _log     = 0;

// use malloc instead of new due to realocate function!
// ====================================================
  _buffer = (char*)malloc((_buffsz+1)*sizeof(char));  // due to realocate function!

//  if( data != 0 ) _data[ data_id ] = data;

}


/* ----------
 * destructor
 */
t_gcoder::~t_gcoder()
{
  // use free instead of delete due to realocate function!
  free(_buffer);
}


/* ----------
 * read settings
 */
void t_gcoder::gset(t_gsetgen& gen)
{
  _beg = gen.beg();
  _end = gen.end();
  _sys = gen.sys();
  _rec = gen.rec();
  _sat = gen.sat();
  _obs = gen.obs();
#ifdef DEBUG   
  cout << "GCODER settings:"
       << " _beg " << _beg.str("%Y-%m-%d %H:%M:%S")
       << " _end " << _end.str("%Y-%m-%d %H:%M:%S")
       << " _sys " << _sys.size() 
       << " _rec " << _rec.size()
       << " _sat " << _sat.size()
       << " _obs " << _obs.size() << endl; 
#endif
}


/* ----------
 * initialize
 */
void t_gcoder::_init()
{
  _version = _initver;
  _endpos  = 0;
}


/* ----------
 * clear
 */
void t_gcoder::clear()
{
  // use free instead of delete due to realocate function!
  free(_buffer);  
   
  _init();
   
// use malloc instead of new due to realocate function!
// ====================================================
  _buffer = (char*)malloc((_buffsz+1)*sizeof(char));  // due to realocate function!
}


/* ----------
 * get single line from the buffer
 */
int t_gcoder::_getline( string& str, int from_pos )
{
  str = "";
#ifdef DEBUG
  cout << "end:"  << _endpos << " from:" << from_pos << endl;
#endif
  if( _endpos == 0 || from_pos >= _endpos) return -1;

  string tmp(_buffer); 
  size_t ifirst = 0;

  if( (ifirst = tmp.find(crlf,from_pos)) == string::npos // &&
//    (ifirst = tmp.find(eof,tmp.size())) == string::npos        
//       tmp.c_str()[tmp.size()] != eof                         // end what about end of line ???
      )  return -1;
   
  str = tmp.substr( from_pos, ifirst+1 - from_pos); // + CRLF

  return str.length();
}


/* ----------
 * cummulate buffer
 */
int t_gcoder::_add2buffer( char* buff, int sz )
{
  if( sz == 0 and endpos() > 0 ) return endpos();
#ifdef DEBUG
  for(int i=0; i<sz; i++ ){
     cout << " IN " << i << " [" << *(buff+i) << "] ";
  }
#endif

  if( _endpos + sz > _buffsz ){
    int old = _buffsz;
     
    // INCREASE BUFFER USING REALLOC !
    _buffsz = (_endpos + sz + 1)*BUFFER_INCREASE_FAC;
    if(! (_buffer = (char *)realloc(_buffer, _buffsz*sizeof(char))) ){
      if( _log )  _log->comment(0,"gcoder","error: reallocation not succesful!");
      exit(1);
    }

    // DO NOT EXCEED MAXIMUM BUFFER SIZE !
    if( _buffsz > MAXIMUM_BUFFER_SIZE ){
      if( _log )  _log->comment(0,"gcoder","error: buffer limited to " + int2str(MAXIMUM_BUFFER_SIZE));
      _consume(sz);
    }else{	       
      if( _log )  _log->comment(2,"gcoder","buffer encreased from " + int2str(old) + " to " + int2str(_buffsz));
    }
  }
   
  for(int i=0; i<sz; i++, _endpos++ ){
     *(_buffer + _endpos) = *(buff+i);
  }

  _buffer[_endpos] = '\0'; // temporary set end of string (will be occasionally replace by add2buff)

#ifdef DEBUG             // HERE WE CAN PRINT _BUFFER SINCE IT ENDS WITH '\0'
  cout << "add2buffer : " << sz << " -> " << _endpos << " BUFF:" << _buffer << ":END\n"; cout.flush();
#endif

  return sz;
}



/* ----------
 * remove from buffer
 */
int t_gcoder::_consume( int bytes_to_eat )
{
  if( bytes_to_eat <= 0        ){ return 0; }
  if( bytes_to_eat >  _endpos or
      bytes_to_eat >  _buffsz  ){
     
    if( _log )  _log->comment(0,"gcoder","error: too many bytes to eat: " + int2str(bytes_to_eat)
                                                                    + " " + int2str(_buffsz)
                                                                    + " " + int2str(_endpos));
    else                cerr << "gcoder - error: too many bytes to eat: " + int2str(bytes_to_eat)
                                                                    + " " + int2str(_buffsz)
                                                                    + " " + int2str(_endpos) << endl;
    exit(1);
  }
   
  for( int eat = 0; eat + bytes_to_eat < _endpos; eat++ ){
    *(_buffer + eat) = *(_buffer + eat + bytes_to_eat );
  }
  _endpos -= bytes_to_eat;
  _buffer[_endpos] = '\0'; // temporary set end of string (will be occasionally replace by add2buff)

#ifdef DEBUG
  cout << "consumed: " << bytes_to_eat << " ... endpos = " << _endpos << "\n";
#endif  
  return bytes_to_eat;
}


// add specific data structure to the coder
// ----------
int t_gcoder::add_data( string data_id, t_gdata* data )
{
  if( data_id.empty() or data == 0 ) return -1;

  map<string,t_gdata*>::iterator it = _data.find(data_id);
  if( it != _data.end() ){
    if( _log ) _log->comment(1,"gcoder","warning: structure " + data_id + " already exists !" );
    return -1;
  }

#ifdef DEBUG
  cout << " gcoder: adding data structure: " << data_id << endl;
#endif
  _data[data_id] = data;
  return 1;
}


// remove specific data structure from the coder
// ----------
int t_gcoder::rem_data( string data_id )
{
  if( data_id.empty() ) return -1;

  map<string,t_gdata*>::iterator it = _data.find(data_id);
  if( it != _data.end() ) return -1;

#ifdef DEBUG
  cout << " gcoder: removing data structure: " << data_id << endl;
#endif

  _data.erase(it);
  return 1;
}
