
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <cstring>

#include "gcoders/rinexc.h"
#include "gall/gallprec.h"
#include "gutils/gtypeconv.h"
 
using namespace std;


/* ----------
 * constructor
 */
t_rinexc::t_rinexc( string version, int sz )
  : t_gcoder( version, sz )
{
  _gnsssys = 'G';
}


/* ----------
 * CLK-RINEX header
 */
int t_rinexc::decode_head( char* buff, int sz, vector<string>& errmsg)
{ 
  boost::mutex::scoped_lock lock(_mutex);
  if( t_gcoder::_add2buffer( buff, sz) == 0 ){ return 0; };

//  double a0, a1;
//  int t,w;

  string tmp;
  int consume = 0;
  int tmpsize = 0;
  while( ( tmpsize = t_gcoder::_getline( tmp )) >= 0 ){

    consume += tmpsize;
    if( tmp.find("RINEX VERSION",60) != string::npos ){          // first line
      if( tmp[20] != 'C' ){
	string ltmp("warning - not rinex clock data file");
        if( _log )
	      _log->comment(0,"rinexc",tmp);
	else  cerr << tmp;
	return -1;
      }
      _version = tmp.substr(0,9);
      if( _log and substitute(_version, " ", "") > 0 ){
	ostringstream ltmp;
        ltmp << "version = " << _version;
	_log->comment(1,"rinexc",ltmp.str());
      }
      _gnsssys = tmp[41];

    }else if( tmp.find("PGM / RUN BY / DATE",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading PGM / RUN BY / DATE");
       
    }else if( tmp.find("COMMENT",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading COMMENT");
       
    }else if( tmp.find("SYS / # / OBS TYPES",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading SYS / # / OBS TYPES");
       
    }else if( tmp.find("TIME SYSTEM ID",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading TIME SYSTEM ID");
              
    }else if( tmp.find("LEAP SECONDS",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading LEAP SECONDS");
       
    }else if( tmp.find("SYS / DCBS APPLIED",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading SYS DCBS APPLIED");
              
    }else if( tmp.find("SYS / PCVS APPLIED",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading SYS PCVS APPLIED");
              
    }else if( tmp.find("# / TYPES OF DATA",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading TYPES OF DATA");
       
    }else if( tmp.find("STATION NAME / NUM",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading STATION NAME");
              
    }else if( tmp.find("STATION CLK REF",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading STATION CLK REF");

    }else if( tmp.find("ANALYSIS CENTER",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading ANALYSIS CENTER");

    }else if( tmp.find("# OF CLK REF",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading # OF CLK REF");
       
    }else if( tmp.find("ANALYSIS CLK REF",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading ANALYSIS CLK REF");
       
    }else if( tmp.find("# OF SOLN STA / TRF",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading # OF SOLN STA / TRF");
       
    }else if( tmp.find("SOLN STA NAME / NUM",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading SOLN STA NAME / NUM");

    }else if( tmp.find("# OF SOLN SATS",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading # OF SOLN SATS");

    }else if( tmp.find("PRN LIST",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading PRN LIST");

/*
      a0 = str2dbl(tmp.substr( 3,  19 );
      a1 = str2dbl(tmp.substr( 22, 19 );
      w  = str2int(tmp.substr( 42,  8 ));
      t  = str2int(tmp.substr( 53,  6 ));
      if( _log ) _log->comment(2,"rinexc","reading DELTA-UTC");
#ifdef DEBUG
      cout << "--- set DELTA-UTC: " << a0 << " : " << a1 << " : " << t << " : " << w << endl;
#endif
*/

    }else if( tmp.find("END OF HEADER",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexc","reading END OF HEADER ");
      t_gcoder::_consume(tmpsize);
      return -1;
    }

    t_gcoder::_consume(tmpsize);
  }

  return consume;
}

   
/* ----------
 * CLK-RINEX body
 */
int t_rinexc::decode_data(char* buff, int sz, int& cnt, vector<string>& errmsg)
{
  t_gtime epoch(t_gtime::GPS);       // epoch
  double data[6];                    // clk/rms values (off,vel,acc)
  string  flg, id, line;
  int tmpsize = 0;                   // individual reading counter
  int consume = 0;                   // total read counter

  for(int i = 0; i<6; i++) data[i] = UNDEFVAL_CLK; // initialize

  boost::mutex::scoped_lock lock(_mutex);

  // complete main buffer
  if( t_gcoder::_add2buffer(buff, sz) == 0 ){ return 0; };

  while( ( tmpsize = t_gcoder::_getline( line ) ) >= 0 ){

    istringstream istr(line);
    istr.clear();

    int yr,mn,dd,hr,mi;
    double sc;
    int ncol = 1;

    // new record
    istr >> flg >> id >> yr >> mn >> dd >> hr >> mi >> sc >> ncol;
    if( istr.fail() ) break;
    epoch.from_ymd( yr, mn, dd, (hr*3600 + mi*60), sc );
     
    // loop over n-column
    for(int i = 0; i < ncol; i++){
       
      // first line
      istr >> data[i];
      if( istr.fail() ) break;

      // additional (second) line
      if( i > 2 ){
        if( ( tmpsize += t_gcoder::_getline( line ) ) < 0 ) break;

        istr >> data[i];
        if( istr.fail() ) break;
      }
    }
       
    if( (! _data.empty()) && flg.find("AS") != string::npos ){
    
      double clk[3] = { data[0], data[2], data[4] };
      double var[3] = { data[0], data[2], data[4] };

      // fill clk
      map<string,t_gdata*>::iterator it = _data.begin();
      while( it != _data.end() ){
        if( _log && _log->verb() >= 3 ){
           ostringstream ltmp;
           ltmp << "add sat clk " << id 
                << epoch.str_ymdhms(" ")
	        << scientific
                << setw(14) << clk[0] << setw(14) << var[0]
     	        << setw(14) << clk[1] << setw(14) << var[1]
	        << setw(14) << clk[2] << setw(14) << var[2];
            _log->comment(3,"rinexc",ltmp.str());
	    }
        ((t_gallprec*)it->second)->addclk( id, epoch, clk, var );
        it++;
      }
      cnt++;
    }
    consume += t_gcoder::_consume(tmpsize);
  }

  return consume;
}
