
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <cstring>

#include "gcoders/rinexo.h"
#include "gdata/gobsgnss.h"
 
using namespace std;


// constructor
// ----------
t_rinexo::t_rinexo( string version, int sz )
  : t_gcoder( version, sz )
{
  this->clear();
}


// clear
// ----------
void t_rinexo::clear()
{ 
  t_gcoder::clear();

  _epoch    = t_gtime::current_time(t_gtime::GPS);
  _gtime    = t_gtime::current_time(t_gtime::GPS);
  _first    = t_gtime::current_time(t_gtime::GPS);
  _last     = t_gtime::current_time(t_gtime::GPS);
  _line     = "";
  _tmpsize  = 0;
  _consume  = 0;
  _complete = true;
  _rnxsys   = 'G';
  _program  = "";
  _runby    = "";
  _markname = "";
  _marknumb = "";
  _marktype = "";
  _observer = "";
  _agency   = "";
  _recnumb  = "";
  _rectype  = "";
  _recvers  = "";
  _antnumb  = "";
  _anttype  = "";
  _strength = "DBHZ";
  _interval = 1;
  _leapsec  = 0;
  _numsats  = 0;
   
  _aprxyz = t_gtriple(0,0,0);
  _antxyz = t_gtriple(0,0,0);
  _antenu = t_gtriple(0,0,0);
   
  _pcosat.erase(_pcosat.begin(),_pcosat.end());
  _pcosys.erase(_pcosys.begin(),_pcosys.end());
  _pcoecc.erase(_pcoecc.begin(),_pcoecc.end());
  _mapobs.erase(_mapobs.begin(),_mapobs.end());

}

   
// OBS-RINEX header
//   - read individual lines (tmpsize once used and consumed)
//   - read block of lines at once (tmpsize cummulated and then consumed)
// ----------
int t_rinexo::decode_head( char* buff, int sz, vector<string>& errmsg)
{ 
  boost::mutex::scoped_lock lock(_mutex);
  if( t_gcoder::_add2buffer( buff, sz) == 0 ){ return 0; };

  bool complete = true;
  int  consume  = 0;
  int  tmpsize  = 0;
  _line = "";

  while( ( tmpsize = t_gcoder::_getline( _line )) >= 0 ){
  
    complete = true; 
    consume += tmpsize;
    // --------
    if( _line.find("RINEX VERSION",60) != string::npos ){          // first line
      if( _line[20] != 'O' ){      // A1
        if( _log ) _log->comment(0,"rinexo","warning: not rinex observation file");
	 // else               cerr << "rinexo - warning: not rinex observation file\n";
	    return -1;
      }
      _rnxsys = toupper(_line[40]);     // A1 (G=GPS, R=GLO, E=GAL, S=SBAS, M=Mix)
      _version = trim(_line.substr(0,9)); // F9.2
      substitute(_version, " ", "");
       
      if( _log and substitute(_version, " ", "") > 0 ){ 
	_log->comment(2,"rinexo", "reading VER: " + _version + " SYS: " + string(1,_rnxsys) );
      }
      
      if( _rnxsys == ' ' ){
        if( _log ) _log->comment(0,"rinexo","warning: RINEX system not defined, GPS set as default" );
        _rnxsys = 'G';
      }
	 
    // --------
    }else if( _line.find("PGM / RUN BY / DATE",60) != string::npos ){
      _program = trim(_line.substr(0,20));
      _runby   = trim(_line.substr(20,20));
      _gtime.tsys(t_gtime::UTC);
      if( _line.substr(56,3) != "UTC" ) _gtime.tsys(t_gtime::LOC);
       
      if(      _gtime.from_str("%Y%m%d %H%M%S",    _line.substr(40,15)) == 0 ) ;
      else if( _gtime.from_str("%Y-%m-%d %H-%M-%S",_line.substr(40,20)) == 0 ) ;
      else{    _gtime.first_time(t_gtime::UTC); }
       
      if( _log && _log->verb() >= 3) 
          _log->comment(3,"rinexo","PGM / RUN BY / DATE: " 
			       + _program + " " + _runby + _gtime.str_ymdhms(" ") );
       
    // --------
    }else if( _line.find("COMMENT",60) != string::npos ){
       
      _comment.push_back(_line.substr(0,60));
      if( _log ) _log->comment(3,"rinexo","COMMENT");
       
    // --------
    }else if( _line.find("MARKER NAME",60) != string::npos ){
              
      _markname = trim(_line.substr(0,4)); // (0,60));
      transform(_markname.begin(), _markname.end(), _markname.begin(), ::toupper);
     if( _log ) _log->comment(3,"rinexo","MARKER NAME: " + _markname);
      
    // --------
    }else if( _line.find("MARKER NUMBER",60) != string::npos ){
       
      _marknumb = trim(_line.substr(0,60));
      if( _log ) _log->comment(3,"rinexo","MARKER NUMB: " + _marknumb);

    // --------
    }else if( _line.find("MARKER TYPE",60) != string::npos ){

      _marktype = trim(_line.substr(0,20));
      if( _log ) _log->comment(3,"rinexo","MARKER TYPE: " + _marktype);

    // --------
    }else if( _line.find("OBSERVER / AGENCY",60) != string::npos ){

      _observer = trim(_line.substr(0,20));
      _agency   = _line.substr(20,40);
      if( _log ) _log->comment(3,"rinexo","OBSERVER / AGENCY: " 
                  + _observer + " " + _agency);

    // --------
    }else if( _line.find("REC # / TYPE / VERS",60) != string::npos ){

      _recnumb  = trim(_line.substr(0, 20));
      _rectype  = trim(_line.substr(20,20));
      _recvers  = trim(_line.substr(40,20));
      if( _log ) _log->comment(3,"rinexo","REC # / TYPE / VERS: "
			      + _recnumb + " " + _rectype + " " + _recvers);

    // --------
    }else if( _line.find("ANT # / TYPE",60) != string::npos ){

      _antnumb  = trim(_line.substr(0, 20));
      _anttype  = trim(_line.substr(20,20));
      if( _log ) _log->comment(3,"rinexo","ANT # / TYPE: "
			      + _antnumb + " " + _anttype);

    // --------
    }else if( _line.find("APPROX POSITION XYZ",60) != string::npos ){

      _aprxyz[0] = str2dbl(_line.substr( 0,14));
      _aprxyz[1] = str2dbl(_line.substr(14,14));
      _aprxyz[2] = str2dbl(_line.substr(28,14));

      if( _log ) _log->comment(3,"rinexo","APPROX POSITION XYZ: " 
                  + dbl2str(_aprxyz[0]) + " " + dbl2str(_aprxyz[1]) + " " + dbl2str(_aprxyz[2]));


    // --------
    }else if( _line.find("ANTENNA: X/Y/Z",60) != string::npos ){

      _antxyz[0] = str2dbl(_line.substr( 0,14));         
      _antxyz[1] = str2dbl(_line.substr(14,14)); 
      _antxyz[2] = str2dbl(_line.substr(28,14));
      if( _log ) _log->comment(3,"rinexo","ANTENNA: X/Y/Z: "
                  + dbl2str(_antxyz[0]) + " " + dbl2str(_antxyz[1]) + " " + dbl2str(_antxyz[2]));

    // --------
    }else if( _line.find("ANTENNA: H/E/N",60) != string::npos ){

      _antenu[0] = str2dbl(_line.substr(14,14));  // East
      _antenu[1] = str2dbl(_line.substr(28,14));  // North
      _antenu[2] = str2dbl(_line.substr( 0,14));  // Up
      if( _log ) _log->comment(3,"rinexo","ANTENNA: H/E/N: "
                  + dbl2str(_antenu[0]) + " " + dbl2str(_antenu[1]) + " " + dbl2str(_antenu[2]));

    // --------
    }else if( _line.find("ANTENNA: PHASECENTER",60) != string::npos ){

      t_gtriple ecc;
      ecc[0] = str2dbl(_line.substr( 5,14));      // N or X
      ecc[1] = str2dbl(_line.substr(19,14));      // E or Y
      ecc[2] = str2dbl(_line.substr(33,14));      // U or Z
       
      _pcosat.push_back(_line[0]);                // GNSS (G/R/E/S/...)
      _pcosys.push_back(trim(_line.substr(2,3))); // coordinate system (NEU/XYZ)
      _pcoecc.push_back(ecc);                     // NEU or XYZ phase center vs. ARP eccentricities

      if( _log ) _log->comment(3,"rinexo","ANTENNA: PHASECENTER: ");
//                     + _pcosat[_pcosat.size()-1]
//		 + " " + _pcosys[_pcsoys.size()-1]
//               + " " +  dbl2str(ecc[0])
//		 + " " +  dbl2str(ecc[1]) 
//		 + " " +  dbl2str(ecc[2]));

    // --------
    }else if( _line.find("ANTENNA: B.SIGHT XYZ",60) != string::npos ){

      _recnumb  = trim(_line.substr(0, 19));
      _rectype  = trim(_line.substr(20,19));
      _recvers  = trim(_line.substr(40,19));
      if( _log ) _log->comment(3,"rinexo","ANTENNA: B.SIGHT XYZ: "
		  + _recnumb + " " + _rectype + " " + _recvers);

    // --------
    }else if( _line.find("ANTENNA: ZERODIR AZI",60) != string::npos ){
      if( _log ) _log->comment(3,"rinexo","ANTENNA: ZERODIR AZI: "   + string("not implemented"));

    // --------
    }else if( _line.find("ANTENNA: ZERODIR XYZ",60) != string::npos ){
      if( _log ) _log->comment(3,"rinexo","ANTENNA: ZERODIR XYZ: "   + string("not implemented"));
       
    // --------
    }else if( _line.find("CENTER OF MASS: XYZ", 60) != string::npos ){
      if( _log ) _log->comment(3,"rinexo","CENTER OF MASS: XYZ: "   + string("not implemented"));

       
    // -------- RINEX v2.x specificonly
    }else if( _line.find("# / TYPES OF OBSERV", 60) != string::npos ){
       
      int num = str2int(_line.substr(0,6));
      int ii   = 1;
      int fact = 1;
      string sys(1,_rnxsys);
 
      for(int i=0; i<num; i++, ii++){
	   	   
	int idx = 10+6*(ii-1);
	// make sure that the obs is not already in t_vobstypes
        GOBS obs = t_gobs::str2gobs(_line.substr(idx, 2));
	t_vobstypes::iterator it = _mapobs[sys].begin();
	while( it != _mapobs[sys].end() ){ 
	  if( it->first == obs ){ it->second = fact; break; }
	  it++;
        }
        if( it == _mapobs[sys].end() ) _mapobs[sys].push_back( make_pair(obs, fact) );

        // read newline and check the end of buffer
        if( ii%9 == 0 ){
          ii = 0;

	  int addsize = t_gcoder::_getline(_line, tmpsize);
	  if( addsize < 0 ){ complete = false; break;
	  }else{             tmpsize += addsize; }

        } 
      } // loop over # observations
       
      if( _log && _log->verb() >= 3 ){
        ostringstream lg;
        lg << "# / TYPES OF OBSERV: [" << sys << "] ";
	
   	    t_vobstypes::const_iterator it = _mapobs[sys].begin();
        while( it != _mapobs[sys].end() ){ 
	      lg << " " << t_gobs::gobs2str( it->first );
	      it++;	 
	    }
        _log->comment(3,"rinexo",lg.str());
      }

    // --------
    }else if( _line.find("SYS / # / OBS TYPES", 60) != string::npos ){
             
      // new system (read all lines at once)
      string sys(1,_line[0]);
      int fact = 1;
      if( sys != " " ){
        int num = str2int(_line.substr(3,3));
        int ii = 1;
	for(int i=0; i<num; i++, ii++){
	   
	  // make sure that the obs is not already in t_vobstypes
	  GOBS obs = t_gobs::str2gobs(_line.substr(7 + 4*(ii-1), 3));
	  t_vobstypes::iterator it = _mapobs[sys].begin();
	  while( it != _mapobs[sys].end() ){ 
	    if( it->first == obs ){ it->second = fact; break; }
	    it++;
	  }
	  if( it == _mapobs[sys].end() ) _mapobs[sys].push_back( make_pair(obs, fact) );
	   	   
	  // read next for old system
  	  if( ii%13 == 0 ){
  	    ii = 0;
	     
	    // get newline and check the end of buffer
	    int addsize = t_gcoder::_getline(_line, tmpsize);
	    if( addsize < 0 ){ complete = false; break;
	    }else{             tmpsize += addsize; }

	  }
        } // loop over # observations	 
      }
      if( _log ){
        ostringstream lg;
        lg << "SYS / # / OBS TYPES: [" << sys << "] ";
	 
	t_vobstypes::const_iterator it = _mapobs[sys].begin();
        while( it != _mapobs[sys].end() ){ 
	  lg << " " << t_gobs::gobs2str( it->first );
	  it++;	 
	}
        _log->comment(3,"rinexo",lg.str());
      }

    // --------
    }else if( _line.find("SYS / SCALE FACTOR",  60) != string::npos ){
       
      // new system (read all lines at once)
      string sys(1,_line[0]);
      int fact = str2int(_line.substr(2,4));

      if( sys != " " ){
        int num = str2int(_line.substr(8,2));
        int ii = 1;
	for(int i=0; i<num; i++, ii++){
	    
	  // make sure that the obs is not already in t_vobstypes
	  GOBS obs = t_gobs::str2gobs(_line.substr(11 + 4*(ii-1), 3));          
	  t_vobstypes::iterator it = _mapobs[sys].begin();
	  while( it != _mapobs[sys].end() ){ 
	    if( it->first == obs ){ it->second = fact; break; }
	    it++;
	  }
	  if( it == _mapobs[sys].end() ) _mapobs[sys].push_back( make_pair(obs, fact) );

	  // read next for old system
  	  if( ii%13 == 0 ){
  	    ii = 0;

	    // get newline and check the end of buffer
	    int addsize = t_gcoder::_getline(_line, tmpsize);
	    if( addsize < 0 ){ complete = false; break;
	    }else{             tmpsize += addsize; }	     
	  }
        } // loop over # observations	 
      }

      if( _log && _log->verb() >= 3 ){
        ostringstream lg;
        lg << "SYS / SCALE FACTOR: [" << sys << "] ";
	 
        t_vobstypes::const_iterator it = _mapobs[sys].begin();
        while( it != _mapobs[sys].end() ){ 
	  lg << " " << t_gobs::gobs2str( it->first );
	  it++;	 
	}
        _log->comment(3,"rinexo",lg.str());
      }

      if( _log ) _log->comment(3,"rinexo","SYS / SCALE FACTOR: "     + string("not implemented"));

    // --------
    }else if( _line.find("SIGNAL STRENGTH UNIT",60) != string::npos ){              
       
      _strength = trim(_line.substr(0,20));
      if( _log ) _log->comment(3,"rinexo","SIGNAL STRENGTH UNIT: " + _strength);

    // --------
    }else if( _line.find("INTERVAL",            60) != string::npos ){
       
      _interval = str2dbl(_line.substr(0,9));
      if( _log ) _log->comment(3,"rinexo","INTERVAL: " + dbl2str(_interval));

    // --------
    }else if( _line.find("TIME OF FIRST OBS",   60) != string::npos ){

      string tsys = trim(_line.substr(48,3));
       
      if(      tsys == "GPS" )  _first.tsys(t_gtime::GPS);
      else if( tsys == "GLO" )  _first.tsys(t_gtime::GLO);
      else if( tsys == "GAL" )  _first.tsys(t_gtime::GAL);
      else                      _first.tsys(t_gtime::GPS);
       
      _first.from_ymdhms(str2int(_line.substr( 0, 6)),        // year
                         str2int(_line.substr( 6, 6)),        // month
                         str2int(_line.substr(12, 6)),        // day
                         str2int(_line.substr(18, 6)),        // hours
                         str2int(_line.substr(24, 6)),        // minutes
                         str2dbl(_line.substr(30,13)) );      // double seconds

      if( _log  && _log->verb() >= 3 ) 
         _log->comment(3,"rinexo",_first.str_ymdhms("TIME OF FIRST OBS: ") );

      
    // --------
    }else if( _line.find("TIME OF LAST OBS",    60) != string::npos ){

      string tsys = trim(_line.substr(48,3));
       
      if(      tsys == "GPS" )  _last.tsys(t_gtime::GPS);
      else if( tsys == "GLO" )  _last.tsys(t_gtime::GLO);
      else if( tsys == "GAL" )  _last.tsys(t_gtime::GAL);
      else                      _last.tsys(t_gtime::GPS);

      _last.from_ymdhms(str2int(_line.substr( 0, 6)),        // year
                        str2int(_line.substr( 6, 6)),        // month
                        str2int(_line.substr(12, 6)),        // day
                        str2int(_line.substr(18, 6)),        // hours
                        str2int(_line.substr(24, 6)),        // minutes
                        str2dbl(_line.substr(30,13)) );      // double seconds


    if( _log && _log->verb() >= 3 ) _log->comment(0,"rinexo",_last.str_ymdhms("TIME OF LAST OBS: ") );

    // --------
    }else if( _line.find("RCV CLOCK OFFS APPL", 60) != string::npos ){
      if( _log ) _log->comment(3,"rinexo","RCV CLOCK OFFS APPL: "    + string("not implemented"));

    // --------
    }else if( _line.find("SYS / DCBS APPL",     60) != string::npos ){
      if( _log ) _log->comment(3,"rinexo","SYS / DCBS APPL: "        + string("not implemented"));

    // --------
    }else if( _line.find("SYS / PCVS APPL",     60) != string::npos ){
      if( _log ) _log->comment(3,"rinexo","SYS / PCVS APPL: "        + string("not implemented"));

    // --------
    }else if( _line.find("SYS / PHASE SHIFT",   60) != string::npos ){
      if( _log ) _log->comment(3,"rinexo","SYS / PHASE SHIFT: "      + string("not implemented"));

    // --------
    }else if( _line.find("GLONASS SLOT / FRQ #",60) != string::npos ){
      if( _log ) _log->comment(3,"rinexo","GLONASS SLOT / FRQ #: "   + string("not implemented"));

    // --------
    }else if( _line.find("LEAP SECONDS",        60) != string::npos ){  
       
      // implemented only 1st part - total LEAP SEC.
      _leapsec = str2int(_line.substr(0,6));
      if( _log ) _log->comment(3,"rinexo","LEAP SECONDS: " 
                  + int2str(_leapsec));

    // --------
    }else if( _line.find("# OF SATELLITES",     60) != string::npos ){
       
      _numsats = str2int(_line.substr(0,6));
      if( _log ) _log->comment(3,"rinexo","# OF SATELLITES: "
		  + int2str(_numsats));

    // --------
    }else if( _line.find("PRN / # OF OBS",      60) != string::npos ){
      if( _log ) _log->comment(3,"rinexo","PRN / # OF OBS: "         + string("not implemented"));

    // --------
    }else if( _line.find("END OF HEADER",60) != string::npos ){
      if( _log ) _log->comment(3,"rinexo","END OF HEADER ");
      t_gcoder::_consume(tmpsize);
      return -1;
    }

    if( complete ) t_gcoder::_consume(tmpsize);
  }

  return consume;
}

   
// OBS-RINEX body
//   - read block of lines for each epoch (tmpsize cummulated and then consumed)
// ----------
int t_rinexo::decode_data(char* buff, int sz, int& cnt, vector<string>& errmsg)
{
  // filter-out unknown marker/receiver
  if( _rec.size() == 0 || _rec.find( _markname ) == _rec.end() ) return -1;

  string dummy, sat;
  char  flg = '0';                   // special event
  int nsats = 0;                     // number of satellite
// JD 
//  vector<t_gobsgnss*> data;           // GNSS observations
  vector<t_gobsgnss> data;           // GNSS observations
  _line     = "";                    // working line
  _tmpsize  = 0;                     // currently processing byte counter
  _consume  = 0;                     // total byte counter
  _complete = true;                  // flag for completed block
   
  boost::mutex::scoped_lock lock(_mutex);

  // complete main buffer
  if( t_gcoder::_add2buffer(buff, sz) == 0 ){ return 0; };

  while( _complete && ( ( _tmpsize = t_gcoder::_getline( _line ) ) >= 0 ) ){
    
//    cout << "line DATA = " << _line;
     
    _complete = true;
    data.clear();

/*     
    // JD TEMP
//    cout << " continue: " << _tmpsize << " " <<  _line.size() << " " << _line;
    t_gcoder::_consume(_tmpsize); continue;
    // JD TEMP
*/
     
#ifdef DEBUG
    if( _log ) _log->comment(0,"rinexo", _line + " " + int2str(_tmpsize));
#endif

    // ------------------------------
    // RINEX 2.x decoder
    // ------------------------------
    if( _version < "3.00" ){
       
      // epoch-line contains list of all satellites in RINEX v2.x
      vector<string> satellites;
       
      if( _read_epoch( nsats, flg ) < 0 ) break;
//      cout << _epoch.str_hms() << " " << nsats << " " << flg << endl; 
      if( flg != '0' && flg != '1' ){ // special event handling
     	if( _log ) _log->comment(3,"rinexo",_epoch.str_ymdhms("RINEX 2.x special events not handled after "));
     // else                        cerr << _epoch.str_ymdhms("RINEX 2.x special events not handled after ") << "\n";
	 
        t_gcoder::_consume(_tmpsize);
        _consume += _tmpsize;
        _tmpsize = 0;
	    continue;
      }

      // DATA READING 
      // ------------
      // read data for vector for satellites (RINEX v2.x specific)
      if( _read_satvec( nsats, satellites ) < 0 ) break;

      // loop over satellits records (x-lines)
      for( int i = 0; i < nsats; i++ ){
   
        string sat = satellites[i];	   
        string key = sat;

        // read sat-line and define system/satellite key in _mapobs (RINEX v2.x specific)
        if( _read_syskey( sat, key ) < 0 ) break;

        // loop over all observation types for system/satellite key in _mapobs
        if( _read_obstypes( sat, key, data ) < 0 ) break;
	 
      }
	 
      // fill data - block (epoch-wise) is completed
      if( _complete ){
        if( _log && _log->verb() >= 5 ) _log->comment(5,"rinexo",_epoch.str_ymdhms(sat + " reading finished at "));
        t_gcoder::_consume(_tmpsize);
	  
        cnt += _fill_data(data);
        _consume += _tmpsize;
        _tmpsize = 0;
      }

    // ------------------------------
    // RINEX 3.x decoder
    // ------------------------------
    }else{

      if( _read_epoch( nsats, flg ) < 0 ) break;
            
      if( flg != '0' && flg != '1' ){ // special event handling
	    if( _log ) _log->comment(3,"rinexo",_last.str_ymdhms("RINEX 3.x special events not handled after "));
     // else                        cerr << _last.str_ymdhms("RINEX 3.x special events not handled after ") << "\n";

        t_gcoder::_consume(_tmpsize);
        _consume += _tmpsize;
        _tmpsize = 0;
   	    continue;
      } 
       
      // DATA READING 
      // ------------
      // loop over satellits records (x-lines)
      for( int i = 0; i < nsats; i++ ){

	    string sat = "";
	    string key = "";
      
        // loop over all observation types for system/satellite key in _mapobs
   	    if( _read_obstypes( sat, key, data ) < 0 ) break;
      }       
	 
      if( _complete ){

        if( _log && _log->verb() >= 5 ) _log->comment(5,"rinexo",_epoch.str_ymdhms(sat + " reading finished at "));
        cnt += _fill_data(data);
        t_gcoder::_consume(_tmpsize);
        _consume += _tmpsize;
        _tmpsize = 0;	   
      }
    } // version     
  } // loop over lines
 
  return _consume;
}


// read sat-system key (G,E,R,.. M, or specific satellite for _mapobs)
// only RINEX 2.x version
// ----------
int t_rinexo::_read_syskey( const string& sat, string& key )
{
  // get newline and check the end of buffer	
  int addsize = t_gcoder::_getline( _line, _tmpsize );

  // getline succeed only if 'EOL' found
  if( (int)_line.length() <= 0 ) return _stop_read();
   
  // select proper key (sat/sys) for observation list    
  if(      _mapobs.count(sat.substr(0,3)) != 0 ) key = sat.substr(0,3);
  else if( _mapobs.count(sat.substr(0,1)) != 0 ) key = sat.substr(0,1);
  else                                           key = string(1,_rnxsys);

#ifdef DEBUG
  cout << " _READ_SYSKEY :"  << sat << "|" << sat.substr(0,1) 
       << "[" << key << "] " << " [" << _rnxsys << "]\n";

  map<string,t_vobstypes>::iterator it = _mapobs.begin();
  while( it != _mapobs.end() ){
    cout << " KEY = " << it->first;
    it++;
  }
  cout << endl;
#endif

  // check if enough data (+1 EOL)
  if( addsize <= (int)_mapobs[key].size()*16 + 1 ){ _tmpsize += addsize;
  }else{ 
    if( _log ) _log->comment(0,"rinexo:",_epoch.str_ymdhms(sat + " RINEX 2.x : data-line probably incomplete: "));
    return _stop_read();
  }
  return 1;
}
   

// read epoch & number of satellites, return flag
// ----------
int t_rinexo::_read_epoch(int& nsats, char& flag)
{
  nsats = 0;
  flag  = 'X';
   
  // not enough data to recognize standard epoch or special event
  if( _line.length() < 30 ) return _stop_read();
  flag = _line[28];

  // keep full space from beginning of line!
  if(       _line.find("                            2",  0) != string::npos ){
    if( _log ) _log->comment(2,"rinexo","warning: start moving antenna identified, but not yet implemented!");
 // else              cerr <<  "rinexo - warning: start moving antenna identified, but not yet implemented!\n";
    return 0;
     
  }else if( _line.find("                            3",  0) != string::npos ){
    if( _log ) _log->comment(2,"rinexo","warning: new site occupation identified, but not yet implemented!");
 // else              cerr <<  "rinexo - warning: new site occupation identified, but not yet implemented!\n";
    return 0;

  }else if( _line.find("                            4",  0) != string::npos ){
    if( _log ) _log->comment(2,"rinexo","warning: new header information identified, but not fully implemented!");
 // else              cerr <<  "rinexo - warning: new header information identified, but not fully implemented!\n";
    return 0;

  }else if( _line.find("                            5",  0) != string::npos ){
    if( _log ) _log->comment(2,"rinexo","warning: external event identified, but not yet implemented!");
 // else              cerr <<  "rinexo - warning: external event identified, but not yet implemented!\n";
    return 0;
  }
  
  if( _line.find("COMMENT",  60) != string::npos ){ // comment in header
    return 0;
  }
     
  istringstream is(_line);
  is.clear();

  string dummy;
  int yr,mn,dd,hr,mi;
  double sc;

  if( _version < "3.00" ){
     is          >> yr >> mn >> dd >> hr >> mi >> sc >> flag >> nsats;
  }else{
     is >> dummy >> yr >> mn >> dd >> hr >> mi >> sc >> flag >> nsats;
  // double offs = 0.0; istr >> offs; istr.clear(); // optional receiver offset (currently not used)
  }

  // check success
  if( is.fail() ){ // cout << "rinexo: problem reading EPOCH:" << _line;    
    return _stop_read();   
  }

  _epoch.from_ymdhms(yr, mn, dd, hr, mi, sc);
	 
#ifdef DEBUG	   
  if( _log ) _log->comment(0,"rinexo","reading epoch: " + _epoch.str_ymd() + " " + _epoch.str_hms());
#endif

  return 1;
}


// read satellite list (vector)
// ----------
int t_rinexo::_read_satvec( const int& nsats, vector<string>& satellites )
{  
 
  int ii = 0;
  for( int i = 0; i < nsats; i++, ii++ ){

    // getline succeed only if 'EOL' found
    if( (int)_line.length() <= 0 ) return _stop_read();

    // only for rinex < 2.x
    if( _version < "3.00" ){
   
      // read extra line for each satellite (if complete)
      unsigned int idx = 32 + ii*3;
        
      if( ii > 11 ){
	ii = 0; idx = 32;
       	unsigned int addsize = t_gcoder::_getline( _line, _tmpsize );
	 
	if( addsize < idx + 3 ) return _stop_read();
        else                    _tmpsize += addsize;
      }

      if( _line.length() >= idx + 3 ){

         string sat = t_gsys::eval_sat( _line.substr(idx,3) );
	 
         // need to check system char for GPS/Galileo/GLONASS/COMPASS/ --> enum in gutils/gsystem.h ?
         satellites.push_back(sat);
      }
    }else{
       // not relevant for rinex 3.x!
    }
  }
 
  return _tmpsize;
}


// read satellite observation types
// ----------
// JD
//int t_rinexo::_read_obstypes( const string& sat, const string& sys, vector<t_gobsgnss*>& data )
int t_rinexo::_read_obstypes( const string& sat, const string& sys, vector<t_gobsgnss>& data )
{
  int ii = 0;
  int addsize = 0;
  unsigned int idx = 0;

  // general sys- or sat- specific group of observations
  string key = sat;
  map<string,t_vobstypes>::const_iterator itMAP = _mapobs.find(sat);
  if( itMAP == _mapobs.end() ){ key = sys; }

  t_vobstypes::const_iterator it = _mapobs[key].begin();

  // RINEX 2.x
  // ---------
  if( _version < "3.00" ){
   
// JD     
//    t_gobsgnss* obs = new t_gobsgnss(_markname, sat, _epoch);
    t_gobsgnss obs(_markname, sat, _epoch);

    // loop over observation types
    while( it != _mapobs[key].end() ){

      idx = 16*ii;

      // read new line for satellite observations
      if( ii > 4 ){
        if( ( addsize = t_gcoder::_getline( _line, _tmpsize ) ) >= 0 ){
          _tmpsize += addsize;
          ii  = 0; 
	  idx = 0;
        }else return _stop_read();
      }
              
      // last check (after new getline or for missing record)
      if( (_line.length() < (idx + 14) ) || trim(_line.substr(idx,14)).empty() )
	         _null_log( sat, t_gobs::gobs2str(it->first) );
      else   _read_obs( idx, it, obs );

      ii++; it++;
       
    } // while (over observation types)

    // filter-out GNSS systems
    if( _sys.size() == 0 || _sys.find( t_gsys::char2str( sat[0] )) != _sys.end() ) data.push_back(obs);

  // RINEX 3.x
  // ---------
  }else{
    string tmpsat("");

    // read new line for individual satellite (complete observations)
    if( ( addsize = t_gcoder::_getline( _line, _tmpsize ) ) >= 3 ){
      _tmpsize += addsize;
      tmpsat = _line.substr(0,3);
    }else  return _stop_read();
     
// JD     
//    t_gobsgnss* obs = new t_gobsgnss(_markname, tmpsat, _epoch);
    t_gobsgnss obs(_markname, tmpsat, _epoch);

    // loop over sys-defined observation types
    vector<pair<GOBS,int> >::const_iterator it = _mapobs[tmpsat.substr(0,1)].begin();
    unsigned int len;
    while( _complete && (it != _mapobs[tmpsat.substr(0,1)].end()) ){

      idx = 3 + 16*ii;
      len = _line.length();
       
      // getline succeed only if 'EOL' found
      if( len <= 0 && len < (idx + 14) ) return _stop_read();

      // check completness (excluding last SNR+LLI - sometimes missing)
      if( len < (idx + 14) ) _null_log( tmpsat, t_gobs::gobs2str(it->first) );
      else                   _read_obs( idx, it, obs );

      ii++; it++;

    } // while (over observation types)

    // filter-out GNSS systems
    if( _sys.size() == 0 || _sys.find( t_gsys::char2str( tmpsat[0] )) != _sys.end() ) data.push_back(obs);

  } // version

#ifdef DEBUG
  cout << " RNX: "     << _version 
       << " sys: "     << sys
       << " key: "     << key
       << " code/phase ID"
//       << " - band1:"  << " " << t_gobs::gobs2str( obs.code(1)  )
//                       << " " << t_gobs::gobs2str( obs.phase(1) )
//       << " - band2:"  << " " << t_gobs::gobs2str( obs.code(2)  )
//                       << " " << t_gobs::gobs2str( obs.phase(2) )
       << endl;
#endif     

  return 1;
}


// fill single observation element
// ----------
// JD
//int t_rinexo::_read_obs(const unsigned int& idx,
//                        const t_vobstypes::const_iterator& it,
//                        t_gobsgnss* obs )
int t_rinexo::_read_obs(const unsigned int& idx,
                        const t_vobstypes::const_iterator& it,
                        t_gobsgnss& obs )
{
  GOBS type = it->first;

  // filter-out unknown marker/receiver
   if( _obs.size() > 0 && _obs.find( trim(t_gobs::gobs2str(type)) ) == _obs.end() ) return 0;

//obs.addobs( type, str2dbl(trim(_line.substr(idx,14))) * it->second ); // trim not necessary !
//obs.addobs( type, strtod(_line.substr(idx,14).c_str(), NULL) * it->second ); // use scale factor !
  obs.addobs( type, str2dbl(_line.substr(idx,14)) * it->second ); // use scale factor !

#ifdef DEBUG	       
  cout << obs.sat()
       << " idx = " << setw(2) << idx << fixed << setprecision(3)
       << "  ["     << type
       << "] ["     << t_gobs::gobs2str(type)
       << "] ["     << setw(14) << str2dbl(trim(_line.substr(idx,14))) * it->second
       << "]"       << _epoch.str_ymd() + " " + _epoch.str_hms()
       << " dsize:" << obs.size()
       << endl; cout.flush();
#endif

  // only for C,L,D, but not SNR observations !!
  if( ( GOBS(type) >=  300 && GOBS(type)  < 400 ) ||  // signal-to-noise ratio
      ( GOBS(type) >= 1300 && GOBS(type) < 1400 ) ){  // signal-to-noise ratio
//  if( t_gobs::gobs2str(type).substr(0,1) != "S" ){ // slower alternative!

    // read LLI
    if( _line.length()-1 == (idx + 14+1) )
         obs.addlli( type, str2dbl(_line.substr(idx+14+2,1)));
    else obs.addlli( type, 0);

    // only for phase observations !
    if( GOBS(type) >= 100 && GOBS(type) < 200 ){
          
      // read SNR type instead of original OBS type (always 3-char in G-Nut = RNX2.x 2-char + space)
      GOBS snrtype = t_gobs::pha2snr(type);
      // ---------------------------------
      // slower alternative!
      // GOBS snrtype = t_gobs::str2gobs( "S" + t_gobs::gobs2str(type).substr(1,2));
      // ---------------------------------

      // read SNR observation and convert flag to 
      if( _line.length() -1 == (idx + 14+2) ){
       
        string snr = trim(_line.substr(idx+14+1,1));
        int i = str2int(snr);
        if( snr == ""  )  obs.addobs( snrtype,  0.0);
        else if( i == 1 ) obs.addobs( snrtype,  6.0);
        else if( i == 2 ) obs.addobs( snrtype, 15.0);	 
        else if( i == 3 ) obs.addobs( snrtype, 20.0);
        else if( i == 4 ) obs.addobs( snrtype, 27.0);
        else if( i == 5 ) obs.addobs( snrtype, 33.0);
        else if( i == 6 ) obs.addobs( snrtype, 39.0);
        else if( i == 7 ) obs.addobs( snrtype, 45.0);
        else if( i == 8 ) obs.addobs( snrtype, 50.0);
        else if( i == 9 ) obs.addobs( snrtype, 60.0);
		 
      }else{ obs.addobs( snrtype,  0.0); }
    }
  }

  // verb() condition to reduce non-verbose mode time
  if( _log && _log->verb() >= 4 ){
    ostringstream lg;
    lg << obs.sat() + ": obs added "
       << _epoch.str_ymdhms()
       << fixed    << setprecision(3)
       << setw(3)  << t_gobs::gobs2str(type) 
       << setw(14) << obs.getobs( type );	  
   _log->comment(4,"rinexo",lg.str());   
 }
 return 0;
}
 

// fill header information
// ----------
int t_rinexo::_fill_head()
{
  int cnt = 0;
/*
  map<string,t_gdata*>::iterator itDAT = _data.begin();
  while( itDAT != _data.end() ){
  
    vector<t_gobsgnss>::const_iterator itOBS = v_obs.begin();
    while( itOBS != v_obs.end() ){
	
      if( _log && _log->verb() >= 4 )
          _log->comment(4,"rinexo",itOBS->sat() + " obs filled "
                        + itOBS->epoch().str_ymdhms()
                        + itOBS->site());

      ((t_gallobs*)itDAT->second)->addobs( *itOBS );
       
      itOBS++;
    }
    cnt++;
    itDAT++;  
  }
*/
  return cnt;
}




// fill observation data structure
// ----------
// JD
//int t_rinexo::_fill_data( const vector<t_gobsgnss*>& v_obs )
int t_rinexo::_fill_data( const vector<t_gobsgnss>& v_obs )
{
  // filter out data
  if( _epoch > _end ) return 0;
  if( _epoch < _beg ) return 0;
     
  int cnt = 0;
  map<string,t_gdata*>::iterator itDAT = _data.begin();
  while( itDAT != _data.end() ){
  
// JD     
//    vector<t_gobsgnss*>::const_iterator itOBS = v_obs.begin();	    
    vector<t_gobsgnss>::const_iterator itOBS = v_obs.begin();	    
    while( itOBS != v_obs.end() ){
/*	
      if( _log && _log->verb() >= 4 )
          _log->comment(4,"rinexo",itOBS->sat() + " obs filled "
                        + itOBS->epoch().str_ymdhms()
                        + itOBS->site());
*/
      ((t_gallobs*)itDAT->second)->addobs( *itOBS );
       
      itOBS++;
    }
    cnt++;
    itDAT++;  
  }

  return cnt;
}


// correct reading stop
// ----------
int t_rinexo::_null_log( const string& sat, const string& obstype )
{
  if( _log && _log->verb() >= 4 )
      _log->comment(4,"rinexo",_epoch.str_ymdhms(sat + " null obs [" + obstype + "] "));
  return 0;
}


// correct reading stop
// ----------
int t_rinexo::_stop_read()
{
  _tmpsize = 0;
  _complete = false;
  return -1;

}