
#ifndef RINEXO_H
#define RINEXO_H
 
/* ----------------------------------------------------------------------
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  Purpose: Clock RINEX encoder/decoder
  Version: $Rev:$

  2011-11-04 /JD: created
  2012-09-04 /JD: RINEX 3.01 implemented
  2012-09-24 /JD: RINEX 2.11 implemented
  2012-09-27 /JD: support of multi-GNSS (phase not converted to [m])
  2013-03-08 /JD: filtering via general settings
 
  Todo: header information completed 
        gobj getting information from rinexo header
        encoders implementation
        to finish obs-filtering via settings

-*/

#include <string> 
#include <vector> 

#include "gall/gallobs.h"
#include "gutils/gtime.h"
#include "gutils/gtriple.h"
#include "gutils/gfreq.h"
#include "gutils/gobs.h"
#include "gutils/gtypeconv.h"
#include "gcoders/gcoder.h"

using namespace std;

class t_rinexo : public t_gcoder {

 public:
   t_rinexo( string version = "", int sz = DEFAULT_BUFFER_SIZE );
  ~t_rinexo(){};

  virtual void clear();

  typedef vector< pair<GOBS,int> > t_vobstypes;

  virtual  int decode_head(char* buff, int sz,           vector<string>& errmsg);
  virtual  int decode_data(char* buff, int sz, int& cnt, vector<string>& errmsg);

 protected:
   int _read_syskey(const string& sat,              // read sat-system key (G,E,R,.. M, or
                          string& key);             // .. specific satellite for _mapobs)
   int _read_epoch(int&,char&);                     // read epoch & number of satellites, return flag
   int _read_satvec(const int& nsat,
                          vector<string>& satellites); // read satellite list (vector)
// JD   
//   int _read_obstypes(const string& sat, const string& sys,
//	 	                    vector<t_gobsgnss*>& data); // read single satellite observation types
   int _read_obstypes(const string& sat, const string& sys,
	 	                    vector<t_gobsgnss>& data); // read single satellite observation types
// JD
//   int _read_obs(const unsigned int& idx,
//                 const t_vobstypes::const_iterator& it,
//		               t_gobsgnss* obs);            // fill single observation element
   int _read_obs(const unsigned int& idx,
                 const t_vobstypes::const_iterator& it,
		               t_gobsgnss& obs);            // fill single observation element
// JD   
//   int _fill_data(const vector<t_gobsgnss*>& v_obs); // fill observation data structure   
   int _fill_head();                                // fill header information
   int _fill_data(const vector<t_gobsgnss>& v_obs); // fill observation data structure
   int _null_log(const string& sat, 
                 const string& obstype);            // write log for null observations (sat, obstype)
   int _stop_read();                                // common stop reading used in local subroutines


 private:
   t_gtime                    _epoch;      // working epoch
   string                     _line;       // working line read from
   int                        _tmpsize;    // working amount of bytes processed
   int                        _consume;    // working total amount of bytes (return)
   bool                       _complete;   // working flag for completed epoch decoding

   char                       _rnxsys;     // G=GPS, R=GLO, E=GAL, S=SBAS, M=Mix
   string                     _program;    // name of program creating RINEX file
   string                     _runby;      // name of agency  creating RINEX file
   t_gtime                    _gtime;      // name of date and file of RINEX creation
   vector<string>             _comment;    // vector of comments
   string                     _markname;   // marker name
   string                     _marknumb;   // marker number
   string                     _marktype;   // marker type
   string                     _observer;   // name of observer
   string                     _agency;     // name of agency
   string                     _recnumb;    // receiver number
   string                     _rectype;    // receiver type
   string                     _recvers;    // receiver version
   string                     _antnumb;    // antenna number
   string                     _anttype;    // antenna type
   t_gtriple                  _aprxyz;     // approximate xyz position [m]
   t_gtriple                  _antxyz;     // antenna xyx eccentricities [m]
   t_gtriple                  _antenu;     // antenna east/north/up eccentricities [m]
   
   vector<char>               _pcosat;     // GNSS (G/R/E/S/...) for antenna phase center vs. ARP
   vector<string>             _pcosys;     // XYZ or NEU system for pco
   vector<t_gtriple>          _pcoecc;     // PCO-ARP values in system above (XYZ, NEU) [m]
   map<string, t_vobstypes>   _mapobs;     // GNSS/sat (G/R/E/S/...) observation group
                                           // vector due to saving order in header !
                                           // including scale factor (1000,100,10,1)
 
   string                     _strength;   // signal strength [DBHZ/...]
   double                     _interval;   // interval [sec]
   t_gtime                    _first;      // time of first observation
   t_gtime                    _last;       // time of last observation
   int                        _leapsec;    // leapseconds since 6-Jan-1980
   int                        _numsats;    // number of satellites

};

#endif
