
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <iostream>
#include <iomanip>
#include <sstream>

#include "gdata/gobsgnss.h"
#include "gutils/gconst.h"
#include "gutils/gtypeconv.h"

using namespace std;

// constructor
// -----------
t_gobsgnss::t_gobsgnss(const string& site, const string& sat, const t_gtime& t)
  : t_gdata(),
     _staid(site),
     _satid(sat),
     _epoch(t)
{ 
  id_type(t_gdata::OBSGNSS);
  id_group(t_gdata::GRP_OBSERV);
}

/*
// copy constructor
// -----------
t_gobsgnss::t_gobsgnss(const t_gobsgnss& rf)
  : t_gdata(),
     _staid(rf.site()),
     _satid(rf.sat()),
     _epoch(rf.epoch())
{ 
  id_type(t_gdata::OBSGNSS);
  id_group(t_gdata::GRP_OBSERV);

  vector<GOBS> obs = rf.obs();
  vector<GOBS>::const_iterator it = obs.begin();
  while( it != obs.end() ){
    _gobs[*it] = rf.getobs( *it );
    it++;
  }
}
*/
 
// destructor
// -----------
t_gobsgnss::~t_gobsgnss()
{
#ifdef DEBUG
  cout << "OBS-destruct: " << site() << " " << sat() << " " 
       << epoch().str("  %Y-%m-%d %H:%M:%S[%T] ") << fixed << setprecision(3);
   
  vector<GOBS> v_obs = this->obs();
  vector<GOBS>::iterator itOBS = v_obs.begin();
  for( ; itOBS != v_obs.end(); ++itOBS ) 
      cout << " " << t_gobs::gobs2str( *itOBS ) << ":" << this->getobs( *itOBS );
  cout  << endl; cout.flush();   
#endif
}


// add observation - special map of doubles
// -----------
void t_gobsgnss::addobs(const GOBS& obs, const double& d)
{
  boost::mutex::scoped_lock lock(_mutex);     
  _gobs[obs] = d;
}


// add lost-of-lock indicators (LLI) - special map of integers
// -----------
void t_gobsgnss::addlli(const GOBS& obs, const int& i)
{  
  boost::mutex::scoped_lock lock(_mutex);
  if( i != 0 ) _glli[obs] = i;
}


// get vector of observations types (GOBS)
// -----------
vector<GOBS> t_gobsgnss::obs()
{  
  boost::mutex::scoped_lock lock(_mutex);
  vector<GOBS> tmp;
  map<GOBS,double>::iterator it = _gobs.begin();  
  while( it != _gobs.end() ){
    tmp.push_back( it->first );
    it++;
  }
  return tmp;
}

// add OBS - code [m], phase [whole cycles], dopler[cycles/sec], snr [DBHZ], ...  observations (double)
// -----------
double t_gobsgnss::getobs(const GOBS& obs)
{
  boost::mutex::scoped_lock lock(_mutex);
  if( _gobs.find(obs) == _gobs.end() ) return UNDEFINED_GOBS;
  return _gobs[obs];
}


// add OBS - code [m], phase [whole cycles], dopler[cycles/sec], snr [DBHZ], ...  observations (double)
// -----------
double t_gobsgnss::getobs(const string& obs)
{
  boost::mutex::scoped_lock lock(_mutex);
  if( _gobs.find(t_gobs::str2gobs(obs)) == _gobs.end() ) return UNDEFINED_GOBS;
  return _gobs[t_gobs::str2gobs(obs)];
}


// add LLI - lost of lock indicator (int)
// -----------
int t_gobsgnss::getlli(const GOBS& obs)
{
  boost::mutex::scoped_lock lock(_mutex);
  if( _glli.find(obs) == _glli.end() ) return 0;
  return _glli[obs];
}


// add LLI - lost of lock indicator (int)
// -----------
int t_gobsgnss::getlli(const string& obs)
{
  boost::mutex::scoped_lock lock(_mutex);
  if( _glli.find(t_gobs::str2gobs(obs)) == _glli.end() ) return 0;
  return _glli[t_gobs::str2gobs(obs)];
}


// return # of observations
size_t t_gobsgnss::size() const
{
 return _gobs.size();
}


//  get GNSS system from satellite IDOC
// ----------
GSYS t_gobsgnss::gsys()
{
  return t_gsys::char2gsys( _satid[0] );
}


// get selected code (GOBS)
// ----------
GOBS t_gobsgnss::code(const int& band, const GOBS& obs)
{    
  size_t sz = sizeof(code_choise[band]) / sizeof(GOBS);
   
  // not preffered option !
  if( obs == X ){
    
    GOBS tmp = obs;
    // loop over choices 
    for( size_t i=0; i<sz; i++ ){
      tmp = code_choise[band][i];
      if( tmp == X || _gobs.find( tmp ) == _gobs.end() ) continue;
      
      if( ! double_eq( _gobs[ tmp ], 0.0) ) return tmp;
      // { cout << " found  code ID: [" << band << "] " << t_gobs::gobs2str( tmp ) << endl; return tmp; }
    }

  // preffered option (via obs settings)
  }else if( _gobs.find( obs ) != _gobs.end() && ! double_eq( _gobs[ obs ], 0.0) ) return obs;

  return X;
}


// get selected code (GOBS)
// ----------
GOBS t_gobsgnss::phase(const int& band, const GOBS& obs)
{    
  size_t sz = sizeof(phase_choise[band]) / sizeof(GOBS);
 
  // not preffered option !
  if( obs == X ){
    
    GOBS tmp = obs;
    // loop over choices 
    for( size_t i=0; i<sz; i++ ){
      tmp = phase_choise[band][i];
      if( tmp == X || _gobs.find( tmp ) == _gobs.end() ) continue;

      if( ! double_eq( _gobs[ tmp ], 0.0) ) return tmp;
      // { cout << " found phase ID: [" << band << "] " << t_gobs::gobs2str( tmp ) << endl; return tmp; }
    }

  // preffered option (via obs settings)
  }else if( _gobs.find( obs ) != _gobs.end() && ! double_eq( _gobs[ obs ], 0.0) ) return obs;

  return X;
}


// code observations [m]
// ----------
double t_gobsgnss::obs_C(const int& band, GOBS* obs) // const
{   
//  boost::mutex::scoped_lock lock(_mutex);
  GOBS tmpobs = X;
   
  if( obs )
    if( *obs != X ) tmpobs = code( band, *obs );  // pointer     exists, obs ID requested
    else            tmpobs = code( band       );  // pointer     exists, obs ID searched
  else              tmpobs = code( band       );  // pointer not exists, obs ID searched

  if( tmpobs != X ){                              // search was succesfull !
    if( obs ) *obs = tmpobs;
    return _gobs[tmpobs];
  }

  return 0.0; // not found
}


// phase observations [m] !
// ----------
double t_gobsgnss::obs_L(const int& band, GOBS* obs) // const
{
//  boost::mutex::scoped_lock lock(_mutex);
  GOBS tmpobs = X;
  if( _satid.size() == 0 ) return 0.0;
  t_gsys gsys( _satid[0] );
  
  if( obs )
    if( *obs != X ) tmpobs = phase( band, *obs );  // pointer     exists, obs ID requested
    else            tmpobs = phase( band       );  // pointer     exists, obs ID searched
  else              tmpobs = phase( band       );  // pointer not exists, obs ID searched

#ifdef DEBUG
  cout << " obs_L("  << band << ")  tmpobs = " << t_gobs::gobs2str( tmpobs )
       << " gsys = " << gsys.str() << " prn = " << _satid << " wvlength = " << gsys.wavelength( band ) << endl; 
#endif
   
  // search was succesfull
  if( tmpobs != X ){
    if( obs ) *obs = tmpobs;
    return ( _gobs[tmpobs] * gsys.wavelength(band) ); // transfer from whole cycles to meters!
  }

  return 0.0; // not found
}


// return value of general code linear combination
// for 2 or 3 bands with given coefficients
// ----------
double t_gobsgnss::code_lc(const int& band1, const double& coef1,
		                   const int& band2, const double& coef2,
	 	                   const int& band3, const double& coef3 )
{
  double C1 = 0.0, C2 = 0.0, C3 = 0.0;
  if( ! double_eq( coef1, 0.0 ) ){ C1 = obs_C( band1 ); if( double_eq(C1,0.0) ) return 0.0; }
  if( ! double_eq( coef2, 0.0 ) ){ C2 = obs_C( band2 ); if( double_eq(C2,0.0) ) return 0.0; }
  if( ! double_eq( coef3, 0.0 ) ){ C3 = obs_C( band3 ); if( double_eq(C3,0.0) ) return 0.0; }
  
  double lc =   ( coef1*C1 + coef2*C2 + coef3*C3 );

#ifdef DEBUG
    cout << _satid << " " << _staid
         << "  code_lc BAND(" << band1 << "," << band2 << "," << band3 << ") "
         <<         "  COEF(" << coef1 << "," << coef2 << "," << coef3 << ")  "
         << fixed << setprecision(3) << setw(16) << C1 << setw(16) << C2 << setw(16) << C3 << "  lc = " << setw(16) << lc <<  endl;
#endif
   
  return lc;
}


// return value of general carrier-phase linear combination
// for 2 or 3 bands with given coefficients
// ----------
double t_gobsgnss::phase_lc(const int& band1, const double& coef1,
		                    const int& band2, const double& coef2,
		                    const int& band3, const double& coef3 )
{
  double L1 = 0.0, L2 = 0.0, L3 = 0.0;
  if( ! double_eq( coef1, 0.0 ) ){ L1 = obs_L( band1 ); if( double_eq(L1,0.0) ) return 0.0; }
  if( ! double_eq( coef2, 0.0 ) ){ L2 = obs_L( band2 ); if( double_eq(L2,0.0) ) return 0.0; }
  if( ! double_eq( coef3, 0.0 ) ){ L3 = obs_L( band3 ); if( double_eq(L3,0.0) ) return 0.0; }

  double lc = ( coef1*L1 + coef2*L2 + coef3*L3 );

#ifdef DEBUG
    cout << _satid << " " << _staid
         << " phase_lc BAND(" << band1 << "," << band2 << "," << band3 << ") "
         <<         "  COEF(" << coef1 << "," << coef2 << "," << coef3 << ")  "
         << fixed << setprecision(3) << setw(16) << L1 << setw(16) << L2 << setw(16) << L3 << "  lc = " << setw(16) << lc <<  endl;
#endif
   
  return lc;
}


      // return value of general code linear combination
// for 2 or 3 bands with given coefficients
// ----------
double t_gobsgnss::code_lcf(const int& band1, const int& coef1,
	 	                    const int& band2, const int& coef2,
		                    const int& band3, const int& coef3 )
{
  t_gsys gsys( this->gsys() );

  double C1 = 0.0, C2 = 0.0, C3 = 0.0;
  if( coef1 != 0 ){ C1 = obs_C( band1 ); if( double_eq(C1,0.0) ) return 0.0; }
  if( coef2 != 0 ){ C2 = obs_C( band2 ); if( double_eq(C2,0.0) ) return 0.0; }
  if( coef3 != 0 ){ C3 = obs_C( band3 ); if( double_eq(C3,0.0) ) return 0.0; }
  
  double f1 = gsys.frequency( band1 );
  double f2 = gsys.frequency( band2 );
  double f3 = gsys.frequency( band3 );

  double lc =   ( coef1*f1*C1 + coef2*f2*C2 + coef3*f3*C3 )
              / ( coef1*f1    + coef2*f2    + coef3*f3    );

#ifdef DEBUG
    cout << _satid << " " << _staid
         << "  code_lcf BAND(" << band1 << "," << band2 << "," << band3 << ") "
         <<          "  COEF(" << coef1 << "," << coef2 << "," << coef3 << ")  "
         << fixed << setprecision(3) << setw(16) << C1 << setw(16) << C2 << setw(16) << C3 << "  lc = " << setw(16) << lc <<  endl;
#endif
   
  return lc;
}
   

// return value of general carrier-phase linear combination
// for 2 or 3 bands with given coefficients
// ----------
double t_gobsgnss::phase_lcf(const int& band1, const int& coef1,
		                     const int& band2, const int& coef2,
		                     const int& band3, const int& coef3 )
{
  t_gsys gsys( this->gsys() );
  
  double L1 = 0.0, L2 = 0.0, L3 = 0.0;
  if( coef1 != 0 ){ L1 = obs_L( band1 ); if( double_eq(L1,0.0) ) return 0.0; }
  if( coef2 != 0 ){ L2 = obs_L( band2 ); if( double_eq(L2,0.0) ) return 0.0; }
  if( coef3 != 0 ){ L3 = obs_L( band3 ); if( double_eq(L3,0.0) ) return 0.0; }
      
  double f1 = gsys.frequency( band1 );
  double f2 = gsys.frequency( band2 );
  double f3 = gsys.frequency( band3 );

  double lc =   ( coef1*f1*L1 + coef2*f2*L2 + coef3*f3*L3 )
              / ( coef1*f1    + coef2*f2    + coef3*f3    );

#ifdef DEBUG
    cout << _satid << " " << _staid
         << " phase_lcf BAND(" << band1 << "," << band2 << "," << band3 << ") "
         <<          "  COEF(" << coef1 << "," << coef2 << "," << coef3 << ")  "
         << fixed << setprecision(3) << setw(16) << L1 << setw(16) << L2 << setw(16) << L3 << "  lc = " << setw(16) << lc <<  endl;
#endif
   
  return lc;
}


// return coefficients (c1,c2) of the ionosphere-free linear combination
// ----------
int t_gobsgnss::ionofree_coef( const int& b1, double& c1, 
                               const int& b2, double& c2 )
{
  t_gsys gsys( this->gsys() );

/*
  // COEFFICIENTS FOR c1*f1*O1 + ... / c1*f1 + ... notation
  // ------------------------------------------------------
  // dual-frequency ionosphere-free LC
  double tmp1 =   gsys.factor(b1);
  double tmp2 = - gsys.factor(b2);
   
  c1 = static_cast<int>( tmp1 );
  c2 = static_cast<int>( tmp2 );
   
  // search integer coefficients (for non-integer multipliers)
  int i = 1;
  while( ( fabs( (c1 - tmp1)*1e6) < 0.0  || 
	   fabs( (c2 - tmp2)*1e6) < 0.0     ) && i++ <= 10 ){
    c1 = static_cast<int>( tmp1 * i );
    c2 = static_cast<int>( tmp2 * i );
  }
   
  // multipliers still non-integer
  if( i > 10 ){  c1 = c2 = 0; return -1; }
*/
  double fac1 = gsys.factor( b1 );
  double fac2 = gsys.factor( b2 );
   
  c1 =   fac1*fac1 / (fac1*fac1 - fac2*fac2);
  c2 = - fac2*fac2 / (fac1*fac1 - fac2*fac2);
   
#ifdef DEBUG
  cout << " coefs (D) = " << fixed << setprecision(3)
       << setw(15) << c1
       << setw(15) << c2
       << endl;
#endif

  return 0;
}


// return coefficients (c1,c2) of the geometry-free linear combination
// ----------
/*
int t_gobsgnss::geomfree_coef( int b1, int& c1, int b2, int& c2 )
{   
  c1 =  0; // neni implementovano
  c2 =  0; // neni implementovano

  return 0;
}
*/


// return coefficients (c1,c2) of the geometry-free linear combination
// ----------
int t_gobsgnss::geomfree_coef(const int& b1, double& c1, 
                              const int& b2, double& c2)
{   
  c1 =  1.0;
  c2 = -1.0;

  return 0;
}


// return value [m] and coefficients (c1,c2) of the geometry-free linear combination
// ----------
//double t_gobsgnss::geomfree_lc( int b1, int& c1, int b2, int& c2 )
//{   
//  c1 =  1;
//  c2 = -1;
//  return phase_lc(b1, 1.0, b2, -1.0);
//}


// Ionosphere-linear code combination [m]
// ----------
//double t_gobsgnss::P3() // const
double t_gobsgnss::P3(const int& b1, const int& b2) // const
{
  double coef1, coef2;
  boost::mutex::scoped_lock lock(_mutex);
   
  ionofree_coef( b1, coef1, b2, coef2 );
     
  double lc = code_lc( b1, coef1, b2, coef2 ); // still dual-frequency only
  return lc;
}


// Ionosphere-free linear phase combination [m] !
// ----------
double t_gobsgnss::L3(const int& b1, const int& b2) // const
{
  double coef1, coef2;
  boost::mutex::scoped_lock lock(_mutex);
   
  ionofree_coef( b1, coef1, b2, coef2 );

  double lc = phase_lc( b1, coef1, b2, coef2 ); // still dual-frequency only
  return lc;
}


// get Geometry-free combination [m] !
// ----------
double t_gobsgnss::L4(const int& b1, const int& b2) // const
{
  double coef1, coef2;
  boost::mutex::scoped_lock lock(_mutex);

  // future METHOD !?
  geomfree_coef( b1, coef1, b2, coef2 );
  double lc_new = phase_lc( b1, coef1, b2, coef2 ); // still dual-frequency only
/*
  // (common to new and old)
  t_gsys gsys( this->gsys() );
  GOBS gobs;

  // new METHOD
  size_t i, sz = sizeof(phase_choise[b1]) / sizeof(GOBS); // size is the same for all bands
  double l1_cur = 0.0, l2_cur = 0.0, lc_cur = 0.0;
  for( i=0; i<sz; ++i ){  // get L1 [m]
    gobs = phase_choise[b1][i];
    if( _gobs.find( gobs ) != _gobs.end() && ! double_eq( _gobs[ gobs ], 0.0 ) ){
      l1_cur = _gobs[ gobs ] * gsys.wavelength( b1 );
      break;
    }
  }
  for( i=0; i<sz; ++i ){  // get L2 [m]
    gobs = phase_choise[b2][i]; 
    if( _gobs.find( gobs ) != _gobs.end() && ! double_eq( _gobs[ gobs ], 0.0 ) ){
      l2_cur = _gobs[ gobs ] * gsys.wavelength( b2 );
      break;
    }
  }
  if( ! double_eq(l1_cur, 0.0) &&
      ! double_eq(l2_cur, 0.0)    ) lc_cur = l1_cur - l2_cur;
*/
/*
  // old METHOD
  double l1_old = 0.0, l2_old = 0.0, lc_old = 0.0;
  GOBS l1_enum[] = { L1P, L1W, L1C, L1 };
  GOBS l2_enum[] = { L2P, L2W, L2C, L2 };
  sz = sizeof(l1_enum) / sizeof(GOBS);
   
  for( size_t i=0; i<sz; ++i ){
    gobs = l1_enum[i];
    if( _gobs.find( gobs ) != _gobs.end() && ! double_eq( _gobs[ gobs ], 0.0 ) ){
      l1_old = _gobs[ gobs ] * gsys.wavelength( b1 ); 
   // cout << " found L1 (" << t_gobs::gobs2str( gobs ) << ") : "<< l1 << endl;
      break;
    }
  }
   
  sz = sizeof(l2_enum) / sizeof(GOBS);
  for( size_t i=0; i<sz; ++i ){
    gobs = l2_enum[i];
    if( _gobs.find( gobs ) != _gobs.end() && ! double_eq( _gobs[ gobs ], 0.0 ) ){
      l2_old = _gobs[ gobs ] * gsys.wavelength( b2 );
   // cout << " found L2 (" << t_gobs::gobs2str( gobs ) << ") : "<< l2 << endl;
      break;
    }
  }   

  if( ! double_eq(l1_old, 0.0) &&
      ! double_eq(l2_old, 0.0)    ) lc_old = l1_old - l2_old;
*/

#ifdef DEBUG
//  cout << " L4(cur) = " << fixed << setprecision(3)
//       << setw(16)      << lc_cur
//       << setw(16)      << l1_cur
//       << setw(16)      << l2_cur
//       << endl;
   
  cout << " L4(new) = " << fixed << setprecision(3)
       << setw(16)      << lc_new
       << setw(16)      << phase_lc(b1, 1, b2, 0)  // first  band
       << setw(16)      << phase_lc(b1, 0, b2, 1)  // second band
       << endl;
#endif   

  return lc_new;
}


// Melbourne-Wuebenna combination phase & code [m] !
// ----------
double t_gobsgnss::MW(const int& b1, const int& b2)
{
  double l1 = 0.0, l2 = 0.0, c1 = 0.0, c2 = 0.0;
  boost::mutex::scoped_lock lock(_mutex);
   
  // future METHOD !   need to distinguish 'geometry-free, ionosphere-free' coefficients etc..
  // --------------- 
//  int coef1, coef2;
//  geomfree_coef( b1, coef1, b2, coef2 );
//  double lc_code  =  code_lc( b1, coef1, b2, coef2 ); // still dual-frequency only
//  double lc_phase = phase_lc( b1, coef1, b2, coef2 ); // still dual-frequency only
  
  t_gsys gsys( this->gsys() );
  double f1 = gsys.frequency( b1 ); // first  band frequency
  double f2 = gsys.frequency( b2 ); // second band frequency
  GOBS obs;
  size_t i;
  size_t sz = sizeof(phase_choise[b1]) / sizeof(GOBS); // size is the same for all bands

  // get L1 [m]
  for( i=0; i<sz; ++i ){
    obs = phase_choise[b1][i];
    if( obs == X || _gobs.find( obs ) == _gobs.end() ) continue;
    // cout << "MW (L1) = " << obs << " " << t_gobs::gobs2str( obs ) << " " << fixed << setprecision(3) << setw(20) << _gobs[ obs ] << endl; cout.flush();
    l1 = _gobs[ phase_choise[b1][i] ] * gsys.wavelength( b1 ); 
    break;
   }

  // get L2 [m]
  for( i=0; i<sz; ++i ){
    obs = phase_choise[b2][i];
    if( obs == X || _gobs.find( obs ) == _gobs.end() ) continue;
    // cout << "MW (L2) = " << obs << " " << t_gobs::gobs2str( obs ) << " " << fixed << setprecision(3) << setw(20) << _gobs[ obs ] << endl; cout.flush();
    l2 = _gobs[ phase_choise[b2][i] ] * gsys.wavelength( b2 ); 
    break;
  }

  // get C1 [m]
  for( i=0; i<sz; ++i ){
    obs = code_choise[b1][i];
    if( obs == X || _gobs.find( obs ) == _gobs.end() ) continue;
    // cout << "MW (C1) = " << obs << " " << t_gobs::gobs2str( obs ) << " " << fixed << setprecision(3) << setw(20) << _gobs[ obs ] << endl; cout.flush();
    c1 = _gobs[ code_choise[b1][i] ]; 
    break;
  }
     
  // get C2 [m]
  for( i=0; i<sz; ++i ){
    obs = code_choise[b2][i];
    if( obs == X || _gobs.find( obs ) == _gobs.end() ) continue;
    // cout << "MW (C2) = " << obs << " " << t_gobs::gobs2str( obs ) << " " << fixed << setprecision(3) << setw(20) << _gobs[ obs ] << endl; cout.flush();
    c2 = _gobs[ code_choise[b2][i] ]; 
    break;
  }

#ifdef DEBUG3
  cout << "MW TEST  = " << fixed << setprecision(3)
       << setw(20)      << l1 
       << setw(20)      << l2
       << setw(20)      << c1
       << setw(20)      << c2
       << endl;
#endif
   
  if( double_eq( l1, 0.0 ) ){ return 0.0; }
  if( double_eq( l2, 0.0 ) ){ return 0.0; }
  if( double_eq( c1, 0.0 ) ){ return 0.0; }
  if( double_eq( c2, 0.0 ) ){ return 0.0; }
   
  double lc_old = (f1*l1 - f2*l2) / (f1 - f2) -
                  (f1*c1 + f2*c2) / (f1 + f2);
   

#ifdef DEBUG
//  cout << " MW(cur) = " << fixed << setprecision(3)
//       << setw(16)      << lc_cur
//       << setw(16)      << l1_cur
//       << setw(16)      << l2_cur
//       << endl;

  cout << " MW(new) = " << fixed << setprecision(3)
       << setw(16)      << lc_old
       << setw(16)      << phase_lc(b1, 1, b2, 0)  // first  band
       << setw(16)      << phase_lc(b1, 0, b2, 1)  // second band
       << endl;
#endif   
   
  return lc_old;
}


// valid
// ----------
bool t_gobsgnss::valid()
{
  boost::mutex::scoped_lock lock(_mutex);
  return this->_valid();
}


// clean data
// ----------
void t_gobsgnss::clear()
{ 
  boost::mutex::scoped_lock lock(_mutex);
  this->_clear();
}


// clean internal function
// ----------
void t_gobsgnss::_clear()
{
  _staid.clear();
  _epoch = t_gtime::first_time();
}


// clean internal function
// ----------
bool t_gobsgnss::_valid() const
{   
  if( _staid.empty() || 
      _staid == ""   || 
      _epoch == t_gtime::first_time() ) return false;

  return true; 
}

// Band of code/phase
// ------------------------------
void t_gobsgnss::nbands(pair<int,int>& nb)
{
   for (int i=1; i<=8; i++)
   {
      if (code(i) != X) nb.first++;
      if (phase(i) != X) nb.second++;      
   }
}
