
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <iostream>

#include "gio/giof.h"

using namespace std;


// constructor
// ---------
t_giof::t_giof( string mask )
  : _mask(mask),
    _name(""),
    _omode(ios_base::trunc),
    _repl(false),
    _tsys(t_gtime::UTC)
{
  boost::mutex::scoped_lock lock(_mutex);
  this->exceptions( fstream::failbit | fstream::badbit );
  if( _mask.find('%') != string::npos ) _repl = true;
  _name = _replace();
}


// destructor
// ---------
t_giof::~t_giof()
{
  boost::mutex::scoped_lock lock(_mutex);
  if( is_open() ){
    try{
      close();
    }catch( fstream::failure e ){
      cerr << "Warning: Exception closing file " << _name << ": " << e.what() << endl;
    }
  }
}


// set mask
// ---------
int t_giof::mask( string mask )
{
  boost::mutex::scoped_lock lock(_mutex); 

  this->clear();
  if( is_open() ){
    try{
#ifdef DEBUG
	  cout << "mask: closing log-file: " << _name << endl;
#endif  
      close();
    }catch( fstream::failure e ){
      cerr << "Warning: Exception closing file " << _name << ": " << e.what() << endl;
      return -1;
    }
  }
  if( mask.find('%') != string::npos ) _repl = true;
  _mask = mask;
  _name = _replace();
  return 1;
}


// local log file archive
// ---------
int t_giof::write( const string& s )
{
  return this->write( s.c_str(), s.size() );
}
   

// local log file archive
// ---------
int t_giof::write( const char* buff, int size )
{     
  if( _mask == "" ) return -1;

  boost::mutex::scoped_lock lock(_mutex);
  string name = _replace();

  this->clear();
  // new name or not opened
  if( ! is_open() or name !=  _name ){
     
    if( is_open() ){
      try{ 
        close();
#ifdef DEBUG
        cout << "closing log-file: " << _name << " [" << name << "]\n";
#endif  
      }catch( fstream::failure e ){
        cerr << "Warning: Exception closing file " << _name << ": " << e.what() << endl;
        return -1;
      }
    }
     
    try{ 
      _name = name;
      open( _name.c_str(), ios::binary | ios::out | _omode );
#ifdef DEBUG
     cout << " opening file " << _name << endl;
#endif
    }catch( fstream::failure e ){
      cerr << "Warning: Exception opening file " << _name << ": " << e.what() << endl;
      return -1;
    }
  }

  try{ 
    fstream::write( buff, size );
  }catch( fstream::failure e ){
    cerr << "Warning: Exception writting to file " << _name << ": " << e.what() << endl;
    return -1;
  }
  return size;
}


// local log file source
// ---------
int t_giof::read( char* buff, int size )
{
  if( _mask == ""   ) return -1;

  boost::mutex::scoped_lock lock(_mutex);
  _name = _replace();

  this->clear();
  if( ! is_open() ){
    try{
      open( _name.c_str(), ios::binary | ios::in );
    }catch( fstream::failure e ){
      cerr << "Warning: Exception opening file " << _name << ": " << e.what() << endl;
      return -1;
    }
  }

  int nbytes = size;
  try{	
    fstream::read( buff, size );
  }catch( fstream::failure e ){
    if( eof() ){
      nbytes = gcount();
    }else{	  
      cerr << "Warning: Exception reading from file " << _name << ": " << e.what() << endl;
      return -1;
    }
  }
#ifdef DEBUG
  cout << "giof_read  [" << nbytes << "] " << buff << "\n name "  << _name << "\n"; cout.flush();
#endif

  return nbytes;
}


// append mode
// ---------
void t_giof::append(const bool& b)
{ 
  if( b ) _omode = ios_base::app;
  else    _omode = ios_base::trunc;
}


// set time system for replacement
// ---------
void t_giof::tsys(t_gtime::t_tsys ts)
{ 
  boost::mutex::scoped_lock lock(_mutex); 
  _tsys = ts;
}


// get time system for replacement
// ---------
t_gtime::t_tsys t_giof::tsys()
{ 
  boost::mutex::scoped_lock lock(_mutex);
  return _tsys;
}


// evaluate name
// ---------
string t_giof::_replace()
{ 
  if( ! _repl ) return _mask;
  return t_gtime::current_time(_tsys).str(_mask);
}


