
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <cmath>
#include <iomanip>

#include "gutils/gconst.h"
#include "gutils/gmatrixconv.h"
#include "gutils/gsysconv.h"
#include "gmodels/gbancroft.h"

void gbancroft(const Matrix& BBpass, ColumnVector& pos) {

  if (pos.Nrows() != 4) {
    pos.ReSize(4);
  }
  pos = 0.0;

  for (int iter = 1; iter <= 2; iter++) {
    Matrix BB = BBpass;
    int mm = BB.Nrows();
    for (int ii = 1; ii <= mm; ii++) {
      double xx = BB(ii,1);
      double yy = BB(ii,2);
      double traveltime = 0.072;
      if (iter > 1) {
        double zz  = BB(ii,3);
        double rho = sqrt( (xx-pos(1)) * (xx-pos(1)) + 
                           (yy-pos(2)) * (yy-pos(2)) + 
                           (zz-pos(3)) * (zz-pos(3)) );
        traveltime = rho / CLIGHT;
      }
      double angle = traveltime * OMEGA;
      double cosa  = cos(angle);
      double sina  = sin(angle);
      BB(ii,1) =  cosa * xx + sina * yy;
      BB(ii,2) = -sina * xx + cosa * yy;
    }
    
    Matrix BBB;
    if (mm > 4) {
      SymmetricMatrix hlp; hlp << BB.t() * BB;
      BBB = hlp.i() * BB.t();
    }
    else {
      BBB = BB.i();
    }
    ColumnVector ee(mm); ee = 1.0;
    ColumnVector alpha(mm); alpha = 0.0;
    for (int ii = 1; ii <= mm; ii++) {
      alpha(ii) = lorentz(BB.Row(ii).t(),BB.Row(ii).t())/2.0; 
    }
    ColumnVector BBBe     = BBB * ee;
    ColumnVector BBBalpha = BBB * alpha;
    double aa = lorentz(BBBe, BBBe);
    double bb = lorentz(BBBe, BBBalpha)-1;
    double cc = lorentz(BBBalpha, BBBalpha);
    double root = sqrt(bb*bb-aa*cc);

    Matrix hlpPos(4,2); 
    hlpPos.Column(1) = (-bb-root)/aa * BBBe + BBBalpha;
    hlpPos.Column(2) = (-bb+root)/aa * BBBe + BBBalpha;

    ColumnVector omc(2);
    for (int pp = 1; pp <= 2; pp++) {
      hlpPos(4,pp)      = -hlpPos(4,pp);
      omc(pp) = BB(1,4) - 
                sqrt( (BB(1,1)-hlpPos(1,pp)) * (BB(1,1)-hlpPos(1,pp)) +
                      (BB(1,2)-hlpPos(2,pp)) * (BB(1,2)-hlpPos(2,pp)) +
                      (BB(1,3)-hlpPos(3,pp)) * (BB(1,3)-hlpPos(3,pp)) ) - 
                hlpPos(4,pp);
    }
    if ( fabs(omc(1)) > fabs(omc(2)) ) {
      pos = hlpPos.Column(2);
    }
    else {
      pos = hlpPos.Column(1);
    }
  }
}

double lorentz(const ColumnVector& aa, const ColumnVector& bb) {
  return aa(1)*bb(1) +  aa(2)*bb(2) +  aa(3)*bb(3) -  aa(4)*bb(4);
}


// Class t_gbancroft
// --------------------------
t_gposition::t_gposition(t_gallobs* obs, t_gallnav* nav, t_gallpcv* pcv)
{
   this->_gobs = obs;
   this->_gnav = nav;
   this->_gpcv = pcv;
}

t_gposition::~t_gposition()
{
}

// Aprox station position
// ------------------------------------
int t_gposition::aprPos(vector<t_gsatdata>& satdata, t_gtriple& crd)
{

   if (_gnav == 0) return -1;

   satPos(satdata);  // compute satellite position and clocks   

   if (satdata.size() <= 4) return -1;

  Matrix BB(satdata.size(), 4); BB = 0.0;

  vector<t_gsatdata>::iterator iter = satdata.begin();
  int iobs = 0;
  while(iter != satdata.end())
   {
      if (iter->P3() == 0.0) satdata.erase(iter);
      else
      {
        iobs++;	 
        BB(iobs,1) = iter->satcrd().crd(0);
        BB(iobs,2) = iter->satcrd().crd(1);
        BB(iobs,3) = iter->satcrd().crd(2);
        BB(iobs,4) = iter->P3() + iter->clk();
        iter++;
      }            
   }      
  BB = BB.Rows(1,iobs);    // delete zero rows

 if (BB.Nrows() <= 4) return -1;   
   
#ifdef DEBUG
   cout << "BB: " << BB << endl;
#endif
   
   ColumnVector pos;
   gbancroft(BB, pos);
   crd.set(pos);

#ifdef DEBUG
   cout << "BB: " << BB << endl;
   cout << "Pos:" << pos << endl;
#endif

   return 1;
}

// Satelite position for all t_gsatdata for one epoch
// ----------
int t_gposition::satPos(vector<t_gsatdata>& satdata)
{

   double xyz[3]  = {0.0, 0.0, 0.0};
   double vel[3]  = {0.0, 0.0, 0.0};
   double var[3]  = {0.0, 0.0, 0.0};
   double clk =    0.0;
   double dclk =   0.0;
   double clkrms = 0.0;   

   vector<t_gsatdata>::iterator iter = satdata.begin();
   while (iter != satdata.end())
   {         
     string satname(iter->sat());
     if( satname.substr(0,1) != "G" ) {
	satdata.erase(iter);
	continue;
     }      

     t_gtime epoT(t_gtime::GPS);
     double satclk = 0.0;
     double satclk2 = 1.0;
   
     bool notclk = false;
     while (abs(satclk-satclk2) > 1.e-4/CLIGHT){
       satclk2 = satclk;
       epoT = iter->epoch() - iter->P3()/CLIGHT - satclk;

       if( _gnav->clk( satname, epoT, &clk, &clkrms, &dclk ) < 0 ){
//	 string comment = "Satelite " + satname + " clock not calculated for epoch " + iter->epoch().str(" %H:%M:%S");
//cout << comment << endl;
//         if( _log ) _log->comment(3, "ppp", comment);
         satdata.erase(iter);
	 notclk = true;
	 break;
       }
       satclk = clk;
     }
     if (notclk) continue;

     if( _gnav->pos( satname, epoT, xyz, var, vel ) < 0 ) {
//      string comment = "Satelite " + satname + " coordinate not calculated for epoch " + iter->epoch().str(" %H:%M:%S");
//cout << comment << endl;
	//      if( _log ) _log->comment(3, "ppp", comment);
       satdata.erase(iter);
       continue;
     }  
   
   // return 0 for eclipsed satelite
   // not impleneted yet
   
     t_gpcv* pcv = 0;
     ColumnVector Satcrd = array2colVec(xyz);

     // phase center ofset correction
     if ( _gpcv != NULL && (pcv = _gpcv->find(satname,"*",epoT)) != 0 )
     {  	
//	cout << "PRN " << satname << endl;
       pcv->pco( epoT, Satcrd, G01 );

//	int ooo; cin >> ooo;

     }

   // relativistic correction
     satclk -= 2.0 * ( xyz[0]*vel[0] + xyz[1]*vel[1] + xyz[2]*vel[2] ) /CLIGHT /CLIGHT;

     t_gtriple crd(Satcrd(1),Satcrd(2),Satcrd(3));
     iter->addcrd(crd);
     iter->addclk(satclk*CLIGHT);
//   cout << "process satPos A2-end " << satname << _epoch.str("  %Y-%m-%d %H:%M:%S [%T] \n"); cout.flush();   

#ifdef DEBUG      
   cout << satname 
        << " CRD " << fixed << setprecision(3)
        <<  "  "   << epoT.str("%Y-%m-%d %H:%M:%S")
        << " X "   << setw(14) << Satcrd(1)
        << " Y "   << setw(14) << Satcrd(2)
        << " Z "   << setw(14) << Satcrd(3)
        << " T "   << setw(14) << clk*1000.0*1000.0
        << endl;
#endif

     iter++;
   }
   
   if (satdata.size() <= 4) return -1;
   
   return 1;
}

// Add satellite information such as elevetion, azimuth, rho
// -------------------------
int t_gposition::satData(vector<t_gsatdata>& satdata, t_gtriple& xyz)
{
   boost::mutex::scoped_lock lock(_mutex);
   
   t_gtriple neu;
   
   vector<t_gsatdata>::iterator iter = satdata.begin();
   while (iter != satdata.end())
   {  
      t_gtime epoch = iter->epoch();
      t_gtriple satcrd = iter->satcrd();

      xyz2neu(satcrd, xyz, neu);
     
      ColumnVector xRec(3);
      double rho0 = (satcrd.crd_cvect() - xyz.crd_cvect()).norm_Frobenius();
      double dPhi = OMEGA * rho0 / CLIGHT;
	
      xRec(1) = xyz[0] * cos(dPhi) - xyz[1] * sin(dPhi);
      xRec(2) = xyz[1] * cos(dPhi) + xyz[0] * sin(dPhi);
      xRec(3) = xyz[2];  
		 
      // !!!!!!!!!! for precise app must be added tide model !!!!!!!!!!!!!!!
      
      double tmp = (satcrd.crd_cvect() - xRec).norm_Frobenius();
      iter->addrho( tmp );

      double NE2 = neu[0]*neu[0]+neu[1]*neu[1];
      double ele = acos(sqrt(NE2)/iter->rho());
      if( neu[2]<0.0 ) {
        ele *= - 1.0;
      }
     
      if( sqrt(NE2)/iter->rho() > 1.0 )
         iter->addele(0.0);
      else iter->addele(ele);

      double azi = atan2(neu[1], neu[0]);
      if (azi < 0) azi += 2*M_PI;
      iter->addazi(azi);
      
      iter++;     
   }   
 return 1;
}
