
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <stdlib.h>
#include <iostream>
#include <iomanip>

#include "gmodels/gpcv.h"
#include "gutils/gsysconv.h"
#include "gutils/gtypeconv.h"
#include "gutils/gmatrixconv.h"

using namespace std;

// -------------------------------
// T_GPCV - single antenn patterns
// -------------------------------

// constructor
// ----------
t_gpcv::t_gpcv()
 : _antenn(""),     // antenna name
   _radome(""),     // radome  name
   _sernum(""),     // serial number
   _method(""),     // serial number
   _source(""),     // source of calibration
   _snxcod("")      // sinex code 
{
  _beg.tsys(t_gtime::UTC);
  _end.tsys(t_gtime::UTC);
}

// destructor
// ----------
t_gpcv::~t_gpcv(){}


// return correction ARP/CoM -> phase center
// zen [rad] - for receiver antenna, it is zenith angle to the satellite
//           - for satellite antenna, it is nadir angle (receiver zenith minus
//             geocentric (space) angle btw XYZsat and XYZrec)
// azi [rad] - for receiver antenna -> need to be implemented !
//           - for satellite antenna not interesting (only if horizontal eccentricity exists!)
// ----------
double t_gpcv::pco( double zen, double azi, GFRQ frq )
{
  boost::mutex::scoped_lock lock(_mutex);
  double corr = 0.0;
  if( zen > PI ){
    ostringstream lg;
    lg << "not valid zenith angle:" << zen*R2D << endl;
    if( _log ) _log->comment(0,"gpcv",lg.str());
    return corr;
  }
   
  // satellite only PCO (Z-offset) mapped to rec-sat direction (approximated)
  if( _mappco.find(frq) != _mappco.end() ){
     corr = _mappco[frq].neu[2] * cos(zen);  // for satellite zen should to be zen-alfa
  }
   
  // SATELLITE HORIZONTAL ECCENTRICITIES NOT YET IMPLEMENTED !!!!
  // RECEIVER PCO NOT YET IMPLEMENTED (azimut/zenith dependent)
   
  return corr;
}


// PV - docasna implementace !!!!!
// -------------------------------
int t_gpcv::pco (t_gtime& epo, ColumnVector& Satcrd, GFRQ frq){
  
  boost::mutex::scoped_lock lock(_mutex);
   
  double MJD = epo.mjd(); 
  double pcvNEU[3];

  if( _mappco.find(frq) != _mappco.end() ){
    pcvNEU[0] = _mappco[frq].neu[0];
    pcvNEU[1] = _mappco[frq].neu[1];
    pcvNEU[2] = _mappco[frq].neu[2];     
    for (int i=0; i<=2; i++){
       pcvNEU[i] /= 1000.0;
    }
  }
  else return -1;
   
   // Satelite-Earth unit vector
     ColumnVector k = -Satcrd / Satcrd.NormFrobenius();
   
   // Along solar panel unit vector
     ColumnVector sun = _ephplan.sunPos(MJD).crd_cvect();
     ColumnVector e = (sun-Satcrd) / sqrt(DotProduct(sun-Satcrd, sun-Satcrd));
     ColumnVector j = crossproduct(k,e);
   
   // complete to satelite fixed right-hand coord system
      ColumnVector i = crossproduct(j,k);      
   
   ColumnVector dx(3);
   dx(1) = pcvNEU[0]*i(1) + pcvNEU[1]*j(1) + pcvNEU[2]*k(1);
   dx(2) = pcvNEU[0]*i(2) + pcvNEU[1]*j(2) + pcvNEU[2]*k(2);
   dx(3) = pcvNEU[0]*i(3) + pcvNEU[1]*j(3) + pcvNEU[2]*k(3);   
   
   for (int i=1; i<=3; i++){
     Satcrd(i) += dx(i);
   }

#ifdef DEBUG
   cout << epo.str("EPOCH: %H:%M:%S") << endl
        << fixed << setprecision(4)
        << "PCV neu: " << pcvNEU[0] << " " << pcvNEU[1] << " " << pcvNEU[2] << endl
//        << "Rotation matrix \n" << R
        << "Sat dX:\n" << dx
        << "Corrected Sat coord:\n" << Satcrd << endl;
#endif
   
   return 1;
}




// return type
// ----------
double t_gpcv::pcv( double zen, double azi, GFRQ frq )
{
  boost::mutex::scoped_lock lock(_mutex);
  double corr = 0.0;
   
  // NOT YET IMPLEMENTED !!!!
  return corr;
}



// return type
// ----------
double t_gpcv::pcv( double zen, GFRQ frq )
{
  boost::mutex::scoped_lock lock(_mutex);
  double corr = 0.0;

  // NOT YET IMPLEMENTED !!!!
  return corr;
}

