
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include "gmodels/gpppmodel.h"


// Global variables
// -----------------------------------------------------------
//const unsigned int MINOBS       =      5;
const double   MAXRES_CODE      =   15.0;
const double   MAXRES_PHASE     =   0.04;

// -----------------------------------------------------------


// Constructors
// -------------------
t_gpppmodel::t_gpppmodel(string site, t_glog* glog, t_gsetbase* settings)
{
   _site     = site;
   _settings = settings;
   _log      = glog;
   
   string trpModStr(dynamic_cast<t_gsetproc*>(_settings)->tropo_model() );

   if      (trpModStr.compare("saastamoinen") == 0) tropoModel = new t_saastamoinen();
   else if (trpModStr.compare("davis")        == 0) tropoModel = new t_davis();
   else if (trpModStr.compare("hopfield")     == 0) tropoModel = new t_hopf();
   else {
     if( _log ) _log->comment(0, "gppp","Tropo Model setting is not correct [" + trpModStr + "]! Check XML-config file");
     else                cerr << "gppp - Tropo Model setting is not correct [" + trpModStr + "]! Check XML-config file\n";
   }
   
}

// Destructor
// --------------------------
t_gpppmodel::~t_gpppmodel()
{
   delete tropoModel;   
}


// wind-up correction
// ----------
double t_gpppmodel::windUp(t_gtime& epoch, const string prn, const ColumnVector& rSat, const ColumnVector& rRec)
{
  double Mjd = epoch.mjd(true) + epoch.sod(true) / 86400.0;

  // First time - initialize to zero
  // -------------------------------
  map<string, double>::iterator it = _windUpTime.find(prn);  
  if( it == _windUpTime.end() ){
      _windUpTime[prn] = Mjd;
      _windUpSum[prn]  = 0.0;
     }
 
  // Compute the correction for new time
  // -----------------------------------
  else if (_windUpTime[prn] != Mjd) {
         
    _windUpTime[prn] = Mjd;
      
    // Unit Vector GPS Satellite --> Receiver
    // --------------------------------------
    ColumnVector rho = rRec - rSat;
    rho /= rho.norm_Frobenius();

    // GPS Satellite unit Vectors sz, sy, sx
    // -------------------------------------
    ColumnVector sz = -rSat / rSat.norm_Frobenius();
  
    ColumnVector xSun = _ephplan.sunPos(Mjd).crd_cvect();
    xSun /= xSun.norm_Frobenius();
  
    // Along solar panel unit vector
    ColumnVector e = (xSun-rSat) / sqrt(DotProduct(xSun-rSat, xSun-rSat));
    ColumnVector sy = crossproduct(sz,e);     
     
     
//  ColumnVector sy = crossproduct(sz, xSun);
    ColumnVector sx = crossproduct(sy, sz);

    // Effective Dipole of the GPS Satellite Antenna
    // ---------------------------------------------
    ColumnVector dipSat = sx - rho * DotProduct(rho,sx) - crossproduct(rho, sy);

    // Receiver unit Vectors rx, ry
    // ----------------------------
    ColumnVector rx(3);
    ColumnVector ry(3); 

    double recEll[3]; xyz2ell(rRec.data(), recEll) ;
    double neu[3];

    neu[0] = 1.0;
    neu[1] = 0.0;
    neu[2] = 0.0;
    neu2xyz(recEll, neu, rx.data());

    neu[0] =  0.0;
    neu[1] = -1.0;
    neu[2] =  0.0;
    neu2xyz(recEll, neu, ry.data());
 
    // Effective Dipole of the Receiver Antenna
    // ----------------------------------------
    ColumnVector dipRec = rx - rho * DotProduct(rho,rx) + crossproduct(rho, ry);
 
    // Resulting Effect
    // ----------------
    double alpha = DotProduct(dipSat,dipRec) /(dipSat.norm_Frobenius() * dipRec.norm_Frobenius());

    if (alpha >  1.0) alpha =  1.0;
    if (alpha < -1.0) alpha = -1.0;

    double dphi = acos(alpha) / 2.0 / M_PI;  // in cycles

    if( DotProduct(rho, crossproduct(dipSat, dipRec)) < 0.0 ) dphi = -dphi;

    _windUpSum[prn] = floor(_windUpSum[prn] - dphi + 0.5) + dphi;
  }

  return _windUpSum[prn];
}


// Outliers detection
// ----------
int t_gpppmodel::outlierDetect(   vector<t_gsatdata>& data,
			             SymmetricMatrix& Qx,
			       const SymmetricMatrix& Qsav, 
			       const ColumnVector& v)
{
   vector<t_gsatdata>::iterator it;
   
   double maxVcode  = 0.0;
   double maxVphase = 0.0;
   
   vector<t_gsatdata>::iterator itMaxVcode;
   vector<t_gsatdata>::iterator itMaxVphase;

   int ii = 0;
   for( it = data.begin(); it != data.end(); it++ ){
      
     ++ii;
     if( maxVcode == 0.0 || abs(v(ii)) > maxVcode ){
	
       maxVcode = abs(v(ii));
       itMaxVcode = it;		     
     }	
	
     if( dynamic_cast<t_gsetproc*>(_settings)->phase() ){
	
       ++ii;
       if( maxVphase == 0.0 || abs(v(ii)) > maxVphase ){
       
	 maxVphase = abs(v(ii));
	 itMaxVphase = it;
       }
     }
   }
#ifdef DEBUG
    cout << "Max res range: " << maxVcode << endl;
    cout << "Max res phase: " << maxVphase << endl;
#endif

  if( maxVcode > MAXRES_CODE ){

    if( _log && _log->verb() >= 2 )
        _log->comment(2, "gppp", _site + " outlier (range) " + itMaxVcode->sat() 
		                       + " size:" + int2str(data.size())
	  	                       + " v: "   + dbl2str(maxVcode));
    data.erase(itMaxVcode);
    Qx = Qsav;
    return 1;
  }

  if( maxVphase > MAXRES_PHASE ){

    if( _log && _log->verb() >= 2 ) 
        _log->comment(2, "gppp", _site + " outlier (phase) " + itMaxVphase->sat()
		                       + " size:" + int2str(data.size())
		                       + " v: "   + dbl2str(maxVphase));
    data.erase(itMaxVphase);
    Qx = Qsav;
    return 1;
  }
  return 0;
}


// model computed range value (phase/code)
// ----------
double t_gpppmodel::cmpObs(t_gtime& epoch, t_gallpar& param, t_gsatdata& gsatdata, bool phase)
{   
  // Cartesian coordinates to ellipsodial coordinates
  t_gtriple xyz, ell;
  ColumnVector cRec(3);

  int info = param.getCrdParam(_site, xyz);

  if ( info > 0 ){
    cRec = xyz.crd_cvect();
  }else{
    dynamic_cast<t_gsetrec*>(_settings)->getCRD(xyz);
    cRec = xyz.crd_cvect();
  }	   
  xyz2ell(xyz.crd_array(), ell.crd_array());   

  ColumnVector cSat = gsatdata.satcrd().crd_cvect();     
  
  double wind = windUp(epoch, gsatdata.sat(), cSat, cRec) * (GPSLAMB3);

  double clkRec = 0.0;
  int i;
  i = param.getParam(_site, t_gpar::CLK, "");
  if ( i >= 0 ) clkRec = param[i].value;
  else
    if( _log && _log->verb() >= 1) 
        _log->comment(1, "gppp", _site + " ! warning:  Receiver Clock is not included in parameters!");

  double trpDelay;

  trpDelay = tropoDelay(epoch, param, ell, gsatdata.ele());

#ifdef DEBUG
   cout << gsatdata.sat() << fixed << setprecision(5)
        << "  rho: "      << gsatdata.rho()
        << "  ell: "      << setw(6) << gsatdata.ele()*180.0/PI
        << "  wind: "     << setw(6) << wind
        << "  rec clk: "  << setw(6) << clkRec
        << " ell[2]:   "  << setw(6) << ell[2]
        << " Tropdelay: "  << setw(6) << trpDelay
        << " cmpObs: "    << setw(6) << gsatdata.rho() + clkRec - gsatdata.clk() + trpDelay + wind
        << endl;
#endif     
     
    if( !phase ) return gsatdata.rho() + clkRec - gsatdata.clk() + trpDelay        ;
    else         return gsatdata.rho() + clkRec - gsatdata.clk() + trpDelay + wind ;
}


// Compute troposperic delay
// -----------
double t_gpppmodel::tropoDelay(t_gtime& epoch, t_gallpar& param, t_gtriple ell, double elevation)
{   
   double delay = 0.0;

   int i;
   i = param.getParam(_site, t_gpar::TRP, "");
   if ( i >= 0 ){
     if ( param[i].getMF().compare("gmf") == 0 ){
        double gmfh, gmfw, dgmfh, dgmfw;
        gmf( epoch.mjd(), ell[0], ell[1], ell[2], M_PI/2.0-elevation,
              gmfh, gmfw, dgmfh, dgmfw );
        delay = gmfh * tropoModel->getZHD(ell, epoch) + gmfw * param[i].value;

#ifdef DEBUG
    cout << epoch.str("EPOCH: %H:%M:%S") << endl << fixed << setprecision(3);
    cout << "Ell:" << ell[0] << " " << ell[1] << " " << ell[2]
	 << " Hydrostatic part: " << tropoModel->getZHD(ell, epoch)
	 << " Wet part: "        << param[i].value << endl << endl;
#endif
     }

     if ( param[i].getMF().compare("gmf") == 0 ){
        delay = t_saastamoinen::getSlantDelay(elevation, ell[2]) + 
	            param[i].value / sin(elevation);
     }      
     }else{
       delay = t_saastamoinen::getSlantDelay(elevation, ell[2]);
   }   

   return delay;
}


// Set site name
// -----------
//void t_gpppmodel::setSite(string site)
//{
//   this->_site = site;
//}
