
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <cmath>
#include <iomanip>

#include "gmodels/gtropo.h"
#include "gmodels/gpt.h"
#include "gutils/gconst.h"
#include "gutils/gtypeconv.h"

using namespace std;

// basis class for tropospheric models

t_gtropo::t_gtropo(){}

t_gtropo::~t_gtropo(){}

// Saastamoinen model

double t_saastamoinen::getSlantDelay(double ele, double height)
{      
	     double pp =  1013.25 * pow(1.0 - 2.26e-5 * height, 5.225) ;
	     double TT =  18.0 - height * 0.0065 + 273.15;
	     double hh =  50.0 * exp(-6.396e-4 * height);
	     double ee =  hh / 100.0 * exp(-37.2465 + 0.213166*TT - 0.000256908*TT*TT);

#ifdef DEBUG
    cout << "BNC meteo data: " << fixed << setprecision(3)
         << " pp " << setw(12) << pp
         << " TT " << setw(12) << TT
         << " hh " << setw(12) << hh
         << " ee " << setw(12) << ee << endl;
#endif
      
	     double h_km = height / 1000.0;
	
	     if (h_km < 0.0) h_km = 0.0;
	     if (h_km > 5.0) h_km = 5.0;
	     int    ii   = int(h_km + 1);
	     double href = ii - 1;
	
	     double bCor[6];
	     bCor[0] = 1.156;
	     bCor[1] = 1.006;
	     bCor[2] = 0.874;
	     bCor[3] = 0.757;
	     bCor[4] = 0.654;
	     bCor[5] = 0.563;
	
	     double BB = bCor[ii-1] + (bCor[ii]-bCor[ii-1]) * (h_km - href);
	
	     double zen  = PI/2.0 - ele;
	     return (0.002277/cos(zen)) * (pp + ((1255.0/TT)+0.05)*ee - BB*(tan(zen)*tan(zen)));

}

double t_saastamoinen::getZHD(t_gtriple Ell, t_gtime& epoch)
{
   
      double P, T, N;
      gpt_v1( epoch.mjd(), Ell[0], Ell[1], Ell[2], P,T,N);
   
   double delay = (0.002277 * P) /
                  (1.0 - 0.00266 * cos(2.0*Ell[0]) - 0.00000028 * Ell[2]);

#ifdef DEBUG
    cout << "GPT data: " << fixed << setprecision(3)
         << " P " << setw(12) << P
         << " T " << setw(12) << T
         << " N " << setw(12) << N
         << " ZHD " << setw(12) << delay << endl;           
#endif   
   
   return delay;

}


double t_saastamoinen::getZWD(t_gtriple Ell, t_gtime& epoch)
{
   
      double P, T, N;   
      gpt_v1( epoch.mjd(), Ell[0], Ell[1], Ell[2], P, T, N);

      T += 273.15;   
   
     double hh =  50.0 * exp(-6.396e-4 * Ell[2]);
     double e =  hh / 100.0 * exp(-37.2465 + 0.213166*T - 0.000256908*T*T);   
   
      double delay =  (0.0022768*e*(1255.0/T + 0.05)) /
                     (1.0 - 0.00266 * cos(2.0*Ell[0]) - 0.00000028 * Ell[2]);
      return delay;
}

// Davis model

double t_davis::getZHD(t_gtriple Ell, t_gtime& epoch)
{
      
      double P, T, N;
      gpt_v1( epoch.mjd(), Ell[0], Ell[1], Ell[2], P,T,N);
   
   double delay = (0.0022768 * P) /
                  (1.0 - 0.00266 * cos(2.0*Ell[0]) - 0.00000028 * Ell[2]);

#ifdef DEBUG
    cout << "GPT data: " << fixed << setprecision(3)
         << " P " << setw(12) << P
         << " T " << setw(12) << T
         << " N " << setw(12) << N
         << " ZHD " << setw(12) << delay << endl;           
#endif   
   
   return delay;

}

double t_davis::getZWD(t_gtriple Ell, t_gtime& epoch)
{
   // not implemented yet
   return 0.0;
}

// Hopfield model

double t_hopf::getZHD(t_gtriple Ell, t_gtime& epoch)
{
   // not implemeted yet
   return 0.0;
}

double t_hopf::getZWD(t_gtriple Ell, t_gtime& epoch)
{
   // not implemeted yet
   return 0.0;
}
