
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <iostream>
#include <sstream>

#include "gutils/gconst.h"
#include "gutils/gtriple.h"
#include "gutils/gmatrixconv.h"
#include "gutils/gsysconv.h"
#include "gutils/gtypeconv.h"
#include "gmodels/gpcv.h"
#include "gmodels/gbancroft.h"
#include "gmodels/gmf.h"
#include "gproc/gpreproc.h"
#include "gproc/gpppfilter.h"


// Global variables
// -----------------------------------------------------------
const unsigned int MINOBS = 5;
double clkJump = 0;
// -----------------------------------------------------------

// constructor
// ----------
t_gpppfilter::t_gpppfilter(string mark, t_gsetbase* set)
  : _site(mark),
    _set(set),
    _log(0),
    _flt(0),
    _smt(0),
    _phase(true),
    _smooth(false),
    _running(false),
    _initialized(false)
{
  if( !set ) cerr << "error: setting not available\n\n";

  _setLog();   // set log  (before using it later on!)
  _setOut();   // set output && identify if smoothing requested
   
  _gotl        = 0;
  _tides       = new t_gtide96();
  _gpcv        = 0;
  _gnav        = 0;
  _epodata     = 0;
  _gpppModel   = new t_gpppmodel(_site, _log, _set);
  _cEpo_beg    = _cEpo_end = t_gtime::first_time();   
  _epoch.tsys(t_gtime::GPS);

  _phase       = dynamic_cast<t_gsetproc*>(_set)->phase();
  _sigCode     = dynamic_cast<t_gsetproc*>(_set)->sigma_C();
  _sigPhase    = dynamic_cast<t_gsetproc*>(_set)->sigma_L();
  _sigAmbig    = dynamic_cast<t_gsetproc*>(_set)->sig_init_amb();

  _trpStoModel = new t_randomwalk();   
  _trpStoModel->setq(dynamic_cast<t_gsetflt*>(_set)->rndwk_ztd() );
   
  _clkStoModel = new t_whitenoise(dynamic_cast<t_gsetflt*>(_set)->noise_clk() );
  _crdStoModel = new t_whitenoise(dynamic_cast<t_gsetflt*>(_set)->noise_crd() );

  string fltModStr(dynamic_cast<t_gsetflt*>(_set)->method() );
  if      (fltModStr.compare("kalman") == 0) _filter = new t_kalman();
  else if (fltModStr.compare("srcf")   == 0) _filter = new t_SRF();
  else
    if( _log ) _log->comment(0, "gppp", _site + " not correct filter setting - check XML-config");
    else                        cerr << _site + " not correct filter setting - check XML-config\n";
         
  _vBanc.ReSize(4); _vBanc = 0.0;

  int ipar = 0;   

  string crdConstr(dynamic_cast<t_gsetrec*>(_set)->CRD_constr());
  if( crdConstr.compare("FIX") != 0 ){
     _param.addParam( t_gpar(_site, t_gpar::CRD_X, ++ipar, "") );
     _param.addParam( t_gpar(_site, t_gpar::CRD_Y, ++ipar, "") );
     _param.addParam( t_gpar(_site, t_gpar::CRD_Z, ++ipar, "") );      
   }         

  _param.addParam( t_gpar(_site, t_gpar::CLK, ++ipar, "") );   

  if( dynamic_cast<t_gsetproc*>(_set)->tropo() ){   
     t_gpar par_trp(_site, t_gpar::TRP,     ++ipar, "");
     par_trp.setMF(dynamic_cast<t_gsetproc*>(_set)->tropo_mf());
     _param.addParam(par_trp);
  }

  _Qx.ReSize( _param.parNumber() ); _Qx = 0.0;  
   
 double crdInit = dynamic_cast<t_gsetproc*>(_set)->sig_init_crd();
 double ztdInit = dynamic_cast<t_gsetproc*>(_set)->sig_init_ztd();
 for (unsigned int i=1; i<=_param.parNumber(); i++) {
   if (_param[i-1].parType == t_gpar::CRD_X )	
     _Qx(i,i) = crdInit * crdInit;
   else if (_param[i-1].parType == t_gpar::CRD_Y)
     _Qx(i,i) = crdInit * crdInit;
   else if (_param[i-1].parType == t_gpar::CRD_Z)
     _Qx(i,i) = crdInit * crdInit;
   else if (_param[i-1].parType == t_gpar::CLK)
     _Qx(i,i) = _clkStoModel->getQ() * _clkStoModel->getQ();
   else if (_param[i-1].parType == t_gpar::TRP)
     _Qx(i,i) = ztdInit * ztdInit;
 }   
} // end constructor


// destructor
// ----------
t_gpppfilter::~t_gpppfilter()
{ 
  if( _filter      ) delete _filter;
  if( _gpppModel   ) delete _gpppModel;
  if( _trpStoModel ) delete _trpStoModel;
  if( _clkStoModel ) delete _clkStoModel;
  if( _crdStoModel ) delete _crdStoModel;
  if( _tides       ) delete _tides;
   
  if( _flt && _flt->is_open() ){ _flt->close(); delete _flt; }
  if( _smt && _smt->is_open() ){ _smt->close(); delete _smt; }
}


// Process observation batch
// ----------
void t_gpppfilter::processBatch(const t_gtime& beg, const t_gtime& end, int sampling)
{
  boost::mutex::scoped_lock lock(_mutex);

  if( beg > end ){ cerr << _site << " - processing not started [beg > end]\n"; return; }

  t_gtime begT( beg );
  t_gtime endT( end );
   
  _running = true;
  _param.setSite(_site);
  double ZTD, ZTDrms;   
  t_gtriple xyz, xyz_ref, neu, ell;

  if( dynamic_cast<t_gsetrec*>(_set)->getCRD(xyz_ref) == 0)
    _param.getCrdParam(_site, xyz_ref);

  double subint = pow(10,floor(log10(sampling)));
  bool forw = true;
  int sign  = fabs(sampling)/sampling;

  unsigned int  mxrun = 0;
  if( _smooth ) mxrun = 1;
  
  for( unsigned int run=0; run <= mxrun ; ++run )
  {    
    // smoothing
    if( run ){
      sign = - fabs(sampling)/sampling;
      forw = false;
      begT = end;
      endT = beg;   
    }
     
   t_gtime now( begT );

#ifdef DEBUG
   cerr << begT.str_ymdhms( "Filtering: " ) << endT.str_ymdhms(" ") << endl;
#endif

   while(( forw && (now < end || now == end)) ||
         (!forw && (now > beg || now == beg)))
   {	
     // synchronization
     if( now.sod()%sampling != 0){
       now.add_dsec( sign * subint/100 ); // add_dsec used for synchronization!
       continue;
     }

     // clean/collect/filter epoch data
     _data.erase( _data.begin(), _data.end() );
     _data = _epodata->obs( _site, now );
     _filterData(_data);
      
     if( _data.size() == 0){
       if( _log && _log->verb() >= 1 )
	   _log->comment(1,"gppp", _site + now.str_ymdhms(" no observation found at epoch: "));

       now.add_sec( sign * sampling );  // not yet add_dsec() ! not support >1Hz data
       continue;
     }

     _trpStoModel->updateTime(now);
     _trpStoModel->setTcurr(now);
      
     if( _processEpoch(now) < 0 ) {
       if( _log && _log->verb() >= 1 ) 
	   _log->comment(1,"gppp", _site + now.str_ymdhms(" epoch ") + " was not calculated" );

       now.add_sec( sign * sampling ); // not yet add_dsec() ! not support >1Hz data
       continue;
     }

     if( _log && _log->verb() >= 1 )
	 _log->comment(1,"gppp", _site + now.str_ymdhms(" processing epoch: "));

     if( _param.getCrdParam(_site, xyz) == 0 ) dynamic_cast<t_gsetrec*>(_set)->getCRD(xyz);
     
     if ( double_eq(xyz_ref[0], 0.0) &&
 	  double_eq(xyz_ref[1], 0.0) &&
	  double_eq(xyz_ref[2], 0.0))
     {
       neu[0] = 0.0;
       neu[1] = 0.0; 
       neu[2] = 0.0;
     }else{
       xyz2neu(xyz, xyz_ref, neu); // xyz2neu(xyz.crd_array(), xyz_ref.crd_array(), neu.crd_array());
     }
      
     xyz2ell(xyz, ell);
     
     int idx = _param.getParam(_site, t_gpar::TRP, "");
     if( idx>0 ) ZTDrms = sqrt(_Qx(idx+1,idx+1));     // _param is indexed from 0, but _Qx from 1 !
     else        ZTDrms = 0.0;
      
     ZTD = _gpppModel->tropoDelay(now, _param, ell, M_PI/2.0);

     ostringstream os;
     os       << fixed << setprecision(3)
	      << now.str_ymdhms()
	      << " "   << setw(13) << xyz[0]
              << " "   << setw(13) << xyz[1] 
	      << " "   << setw(13) << xyz[2]
	      << " "   << setw(7)  << ZTD
	      << " "   << setw(13) << neu[0]
              << " "   << setw(13) << neu[1] 
	      << " "   << setw(13) << neu[2]
	      << fixed << setprecision(5)
	      << " "   << setw(9)  << sqrt(_Qx(1,1))
	      << " "   << setw(9)  << sqrt(_Qx(2,2)) 
	      << " "   << setw(9)  << sqrt(_Qx(3,3))
	      << " "   << setw(7)  << ZTDrms
              << " "   << setw(2)  << _data.size()
              << endl;

     if(  forw && _flt ){
       _flt->write(os.str().c_str(), os.str().size());
       _flt->flush();
     }

     if(! forw && _smt ){
       _smt->write(os.str().c_str(), os.str().size());
       _smt->flush();
     }

     _initialized = true;

     now.add_sec( sign * sampling ); // not yet add_dsec() ! not support >1Hz data
   }
  }
  _running = false;
}


// Set obs, nav
// --------------------------------
void t_gpppfilter::setDAT(t_gallobs* gobs, t_gallnav* gnav)
{      
   _epodata = gobs;
   _gnav    = gnav;
}

// Set PCV
// --------------------------------
void t_gpppfilter::setPCV(t_gallpcv* gpcv)
{      
   _gpcv = gpcv;
}


// Set OTL
// --------------------------------
void t_gpppfilter::setOTL(t_gallotl* gotl)
{      
   _gotl = gotl;
   if( _tides ) _tides->setOTL( _gotl );
}


// Set site for processing
// -----------------------------
bool t_gpppfilter::isRunning()
{ 
//  boost::mutex::scoped_lock lock(_mutex);   // not if processBatch
  return _running;
}


// ----------------------------------------------------------------------------------
// PROTECTED
// ----------------------------------------------------------------------------------
// 


// Filter data (del zero obs)
// --------------------
void t_gpppfilter::_filterData(vector<t_gsatdata>& epoData)
{
   vector<t_gsatdata>::iterator it = epoData.begin();
   while (it != epoData.end())
   {
      if ( double_eq(it->P3(), 0.0) || double_eq(it->L3(), 0.0) ) 
      {
#ifdef DEBUG
       cout << "t_gpreproc::filterData(VEC): erasing data, ";
       cout << "epo: " << it->epoch().str_hms() << ", ";
       cout << "prn: " << it->sat() << endl;
#endif	 
       epoData.erase(it); // zveda iterator, takze nakonec ukazuje na end()!
      }
     it++;
   }   
}


// Process one epoch
// -----------------------------
int t_gpppfilter::_processEpoch(const t_gtime& runEpoch)
{ 
#ifdef DEBUG
  cout << "process Epoch - full data size: " << _data.size() << endl;
#endif

  _epoch = runEpoch;

  Matrix A;
  DiagonalMatrix P;
  ColumnVector l, dx, dx_el;
  SymmetricMatrix Qsav;
  ColumnVector    v;
  vector<t_gsatdata>::iterator it;

  _cntrep  = 0;    // number of iterations caused by outliers
  do{
     
    if( _prepareData() < 0 ) return -1;

    _predict();

    t_gtriple groundEll, groundXYZ; 
     
    if ( _param.getCrdParam(_site, groundXYZ) > 0){
    }else{ dynamic_cast<t_gsetrec*>(_set)->getCRD(groundXYZ); }
     
    xyz2ell(groundXYZ.crd_array(), groundEll.crd_array());

    unsigned int nObs;    
    if( ! _phase ) nObs = _data.size();   // P3 used only
    else nObs = 2 * _data.size();        // P3 and L3 used together	     
	
    unsigned int nPar = _param.parNumber();
	
    A.ReSize(nObs, nPar); A = 0.0;	
    l.ReSize(nObs);       l = 0.0;	
    P.ReSize(nObs);       P = 0.0;	 
    dx.ReSize(nPar);     dx = 0.0;	     	
    int iobs = 1;

    // to this value is constrained tropo (wet part)
    double TRP_fix = dynamic_cast<t_gsetrec*>(_set)->ZTD() - _gpppModel->tropoModel->getZHD(groundEll, _epoch);
     
    // Create matrixes and vectors for estimation
    // ----------------------------------------------------------------------------------	     	
    for( it = _data.begin(); it != _data.end(); it++ ){

      // create first design matrix for P3       
      for( unsigned int ipar=1; ipar <= _param.parNumber(); ipar++ ){

	if( _param[ipar-1].parType == t_gpar::AMB )
	     A(iobs, ipar) = 0.0;
        else A(iobs, ipar) = _param[ipar-1].partial( *it, _epoch, groundEll.crd_array() );
      }

      // create reduced measurement, obs. weight matrix for P3
      double weight_coef;       
       
      if (it->ele()*180.0/M_PI > 30) weight_coef = 1.0;
      else weight_coef = cos(it->ele());// * sin(it->ele());
       
      P(iobs,iobs) =  1 / (_sigCode * _sigCode) / weight_coef ;

      double modObsP = _gpppModel->cmpObs(_epoch, _param, *it, false );
      l(iobs) = it->P3() - modObsP;

#ifdef DEBUG       
      cout << it->sat()     << fixed << setprecision(5)
           << "  P3: "      << setw(12) << it->P3()
           << "  cmpObs: "  << setw(12) << modObsP
           << endl;
#endif
       
      if( _phase ){
	 
	iobs++; // count phase obs
		
	// create first design matrix for L3
	for( unsigned int ipar=1; ipar<=_param.parNumber(); ipar++ ){
	   
	  if( _param[ipar-1].parType == t_gpar::AMB && 
	      _param[ipar-1].prn     != it->sat() )
				 
	       A(iobs, ipar) = 0.0;
	  else A(iobs, ipar) = _param[ipar-1].partial( *it, _epoch, groundEll.crd_array());
	}		
	 
	// create reduced measurement, obs. weight matrix for L3

	 P(iobs,iobs) =  1 / ( (_sigPhase * _sigPhase) / weight_coef);
	 
	for( unsigned int iPar=0; iPar<=_param.parNumber()-1; iPar++ ){

	  if( _param[iPar].parType == t_gpar::AMB &&
	      _param[iPar].prn.compare( it->sat() ) == 0 ) {
	
            double modObsL = _gpppModel->cmpObs(_epoch, _param, *it, true );

	    // Check lli for cycle slip	       	       

	     if ( it->getlli(L1P) == 1 || it->getlli(L2P) == 1 )
	       {
//		  _param[iPar].value -= 1e-3*CLIGHT;
		 _param[iPar].value = it->L3() - modObsL;
		 _Qx(iPar+1, iPar+1) =  _sigAmbig * _sigAmbig;
	       }
	     		     
	      // reduced measurement for L3
	      l(iobs) = it->L3() - _param[iPar].value - modObsL;

#ifdef DEBUG	     
	cout << it->sat()    << fixed << setprecision(5)
             << "  epoch"    << _epoch.str_ymdhms()
             << "  L1: "     << setw(12) << it->getobs(L1)
	     << "  L3: "     << setw(12) << it->L3()
	     << "  amb: "    << setw(12) << _param[iPar].value
	     << "  cmpObs: " << setw(12) << modObsL
	     << endl;
#endif
	    }
	  }
       }
       iobs++; // count code obs

    } // end loop over all measurement in current epoch

// end of creation objects for estimation
// ------------------------------------------------------------------------------------------------     

// Add pseudomeasurments as constrain
// --------------------------------------
    // Design Matrix
    if( _epoch > _cEpo_beg && _epoch < _cEpo_end )
    {
#ifdef DEBUG       
      cout << _epoch.str_ymdhms(" epoch: ") << endl;
      cout << _cEpo_beg.str_ymdhms(" cEpo_beg: ") << endl;
      cout << _cEpo_end.str_ymdhms("_cEpo_end: ") << endl;
#endif

      Matrix_addRC(A, A.Nrows()+1, 0);
      int i = _param.getParam(_site, t_gpar::TRP, "");
      A(A.Nrows(), i+1) = 1;
    
      // Reduced measurement
      Vector_add(l, l.Nrows()+1);
      l(l.Nrows()) = _param[i].value - TRP_fix;
     
       // weight matrix
      Matrix_addRC(P, P.Nrows()+1);
      P(P.Nrows(), P.Ncols()) = 99999;
    }
// -------------------------------------     

    Qsav =  _Qx;

    _filter->update(A, P, l, dx, _Qx);

     // Test if _Qx is positive definite
     try {
       LowerTriangularMatrix L = Cholesky(_Qx);
     }     
     catch (Exception NPDException) {
       if( _log && _log->verb() >= 1 )
          _log->comment(1, "gppp", _site + _epoch.str_ymdhms(" epoch ")
                                         + " variance-covariance matrix is not positive definite.\n");
     }
//     int i = _param.getParam(_site, t_gpar::CLK, "");    

     v = l - A*dx;
    
#ifdef DEBUG     
    cout << fixed   << setprecision(5)
         << _epoch.str_ymdhms() << endl
         << " l: "  << setw(7) << l 
         << " dx: " << setw(7) << dx
         << " v.Nrows(): " << v.Nrows() 
         << " v: "  << setw(7) << v 
         << " A: "  << setw(7) << A
         << " P: "  << setw(7) << P
         << " Q: "  << setw(7) << _Qx
         << endl; cout.flush();

#endif
  }while( _gpppModel->outlierDetect(_data, _Qx, Qsav, v) != 0 );

  if( _data.size() <= static_cast<unsigned int>(MINOBS) ){
     
    if( _log && _log->verb() >= 3 )
	_log->comment(3, "gppp", _site + _epoch.str_ymdhms(" epoch ") 
		                       + " skipped (_data.size < MINOBS)");
    return -1;
  }

  // Edit and save parematres
  for( unsigned int iPar=0; iPar<=_param.parNumber()-1; iPar++ ) {
     _param[iPar].value += dx(_param[iPar].index);
  }   

  return 1;
}


// Compute Bancroft
// ----------
int t_gpppfilter::_prepareData()
{   
  Matrix BB;

  BB.ReSize(_data.size(), 4); /// HOW TO HANDLE TIME !!!
  BB = 0.0;   
  int iobs = 0;
  vector<t_gsatdata>::iterator iter = _data.begin();

  while( iter != _data.end() ){       
    if( _satPos( _epoch, *iter ) < 0 || iter->P3() == 0.0 || iter->L3() == 0.0 ){

#ifdef DEBUG
    cout << "process cmpBnc - erase data " 
	 << _epoch.str_ymdhms(_site+" "+iter->sat()+" "+int2str(_data.size())+" ") 
	 << endl; cout.flush();
#endif       
      _data.erase( iter ); // !!!! OK u vektoru funguje bez ++ (erase zvysuje pointer !)

    }else{
      iobs++;
      BB(iobs,1) = iter->satcrd().crd(0);
      BB(iobs,2) = iter->satcrd().crd(1);
      BB(iobs,3) = iter->satcrd().crd(2);
      BB(iobs,4) = iter->P3() + iter->clk();
      iter++;
    }
  }

   if( _data.size() < static_cast<int>(MINOBS) ) {
      if( _log && _log->verb() >= 2 )
	  _log->comment(2, "gppp", _epoch.str_ymdhms( _site + " epoch ") + int2str(_data.size())
			          + " skipped (Bancroft not calculated: _data.size < MINOBS)");
      return -1;
   }
   
  BB = BB.Rows(1,iobs);    // delete zero rows
   
  if( BB.Nrows() < static_cast<int>(MINOBS) ) {
      if( _log && _log->verb() >= 2 )
          _log->comment(2, "gppp", _epoch.str_ymdhms(_site + " epoch ")
			         + " skipped (Bancroft not calculated: BB.Nrows < MINOBS)");
    return -1;
  }
   
  _vBanc = 0.0;
  gbancroft(BB, _vBanc);

   if (dynamic_cast<t_gsetrec*>(_set)->valid_crd() && 
       dynamic_cast<t_gsetrec*>(_set)->CRD_constr().compare("APRIORI") == 0 ) {      
      _vBanc(1) = dynamic_cast<t_gsetrec*>(_set)->X();
      _vBanc(2) = dynamic_cast<t_gsetrec*>(_set)->Y();
      _vBanc(3) = dynamic_cast<t_gsetrec*>(_set)->Z();
   }   
   
#ifdef DEBUG
   cout << _epoch.str_ymdhms() << endl;;
   cout << "\nBB\n" << fixed << setprecision(3) << BB << endl;
   cout << "vBanc\n"  << fixed << setprecision(3)<< _vBanc << endl;
   cout << "Clk [mili sec]" << fixed << setprecision(10) << _vBanc(4)*1000/CLIGHT << endl;
   cout << "_Qx\n"  << fixed << setprecision(3)<< _Qx << endl;    cout.flush();
#endif
      
  //Compute sat elevation and rho
  iter = _data.begin();
  while( iter != _data.end() ){

     t_gtriple xyz_r, xyz_s, neu_r;

     xyz_s = iter->satcrd();

     if ( dynamic_cast<t_gsetrec*>(_set)->CRD_constr().compare("FIX") == 0 ){
	dynamic_cast<t_gsetrec*>(_set)->getCRD(xyz_r);
     }else if( !_initialized ){
	xyz_r.set(_vBanc.Rows(1,3));
    }else{
       _param.getCrdParam(_site, xyz_r);
    }

#ifdef DEBUG
    cout << "Coordinates: " << fixed << setprecision(3)
         << " X_R " << setw(12) << xyz_r[0]
         << " Y_R " << setw(12) << xyz_r[1]
         << " Z_R " << setw(12) << xyz_r[2] << endl
         << " X_S " << setw(12) << xyz_s[0]
         << " Y_S " << setw(12) << xyz_s[1]
         << " Z_S " << setw(12) << xyz_s[2] << endl; cout.flush();
#endif

    xyz2neu(xyz_s.crd_array(), xyz_r.crd_array(), neu_r.crd_array());
	
    t_gtriple xRec;
    double rho0 = sqrt(pow(xyz_r[0]-xyz_s[0],2) + pow(xyz_r[1]-xyz_s[1],2) + pow(xyz_r[2]-xyz_s[2],2));
    double dPhi = OMEGA * rho0 / CLIGHT;
	
    xRec[0] = xyz_r[0] * cos(dPhi) - xyz_r[1] * sin(dPhi);
    xRec[1] = xyz_r[1] * cos(dPhi) + xyz_r[0] * sin(dPhi);
    xRec[2] = xyz_r[2];  

    if( _tides ){
      xRec = xRec + _tides->tide_searth(_epoch, xRec)
                  + _tides->load_ocean(_epoch, _site)
     	          + _tides->tide_pole()
	          + _tides->load_atmosph();
    }
   
    double tmp = (iter->satcrd().crd_cvect() - xRec.crd_cvect()).norm_Frobenius();
    iter->addrho( tmp );    
     
    double NE2 = neu_r[0]*neu_r[0]+neu_r[1]*neu_r[1];
    double ele = acos(sqrt(NE2)/iter->rho());
    if( neu_r[2]<0.0 ) {
       ele *= - 1.0;
    }
     
    if( sqrt(NE2)/iter->rho() > 1.0 )
         iter->addele(0.0);
    else iter->addele(ele);
     
    if( iter->ele()*180.0/PI < dynamic_cast<t_gsetproc*>(_set)->minimum_elev() ){
      if( _log && _log->verb() >= 2 )
	  _log->comment(2, "gppp", _epoch.str_ymdhms(_site + " epoch ") + " PRN " + iter->sat()
	                   + " erased due to small elevation (" + dbl2str(iter->ele()*180/M_PI) + ")");

      _data.erase(iter); // !!!! OK u vektoru funguje bez ++ (erase zvysuje pointer !)
      if( _data.size() < static_cast<int>(MINOBS) ){
        if( _log && _log->verb() >= 2 )
  	  _log->comment(2, "gppp", _site + _epoch.str_ymdhms(" epoch ") + int2str(_data.size())
			                + " skipped (Bancroft not calculated: _data.size < MINOBS)");
        return -1;
      }       
    }
    else iter++;
  }
  return 1;
}
// --------------------------------------------------------


// Predict state vector and covariance matrix
// =======================================================
void t_gpppfilter::_predict()
{
   
  _cntrep += 1;   

  // Predict coordinates, clock and troposphere
  // state vector is the same, covariance matrix is predicted with noise   
   int i;
   
   i = _param.getParam(_site, t_gpar::CRD_X, "");
   if ( i >= 0 )  {
     if ( !_initialized ) _param[i].value = _vBanc(1);
     if (_cntrep == 1   ) _Qx(i+1,i+1) += _crdStoModel->getQ() * _crdStoModel->getQ();
   }

   i = _param.getParam(_site, t_gpar::CRD_Y, "");
   if ( i >= 0 )  {
     if ( !_initialized ) _param[i].value = _vBanc(2);
     if (_cntrep == 1   ) _Qx(i+1,i+1) += _crdStoModel->getQ() * _crdStoModel->getQ();
   }   
   
   i = _param.getParam(_site, t_gpar::CRD_Z, "");
   if ( i >= 0 )  {
     if ( !_initialized ) _param[i].value = _vBanc(3);
     if (_cntrep == 1   ) _Qx(i+1,i+1) += _crdStoModel->getQ() * _crdStoModel->getQ();
   }
   
   i = _param.getParam(_site, t_gpar::CLK, "");
   if ( i >= 0 )  {
     _param[i].value = _vBanc(4);
     for( unsigned int jj = 1; jj <= _param.parNumber(); jj++ ) _Qx(i+1, jj) = 0.0;      
     _Qx(i+1,i+1) = _clkStoModel->getQ() * _clkStoModel->getQ();
   }   

   i = _param.getParam(_site, t_gpar::TRP, "");
   if ( i >= 0 )  {
     if (!_initialized){
	// for determination elipsoidal crd of receiver
	t_gtriple Ell, XYZ;
	if (_param.getCrdParam(_site, XYZ) > 0){
	}else{ dynamic_cast<t_gsetrec*>(_set)->getCRD(XYZ); }
	xyz2ell( XYZ.crd_array(), Ell.crd_array());

        if ( dynamic_cast<t_gsetrec*>(_set)->valid_ztd() ) {
	   _param[i].value = dynamic_cast<t_gsetrec*>(_set)->ZTD() - _gpppModel->tropoModel->getZHD(Ell, _epoch);
        }  else _param[i].value = _gpppModel->tropoModel->getZWD(Ell, _epoch);
	
     }else if(_cntrep == 1) _Qx(i+1,i+1) += _trpStoModel->getQ();
   }                                
//      cout << "TrpStoModel->getQ " << fixed << setprecision(10) << _trpStoModel->getQ() << endl;
  // Predict ambiguity 
  // Add ambiguity parameter and appropriate rows/columns covar. matrix

  if( _phase )  _syncAmb();
   
} // end prediction method
// -------------------------------------------------------


// Satelite position
// ----------
int t_gpppfilter::_satPos(t_gtime& epo, t_gsatdata& gsatdata)
{
   
   double xyz[3]  = {0.0, 0.0, 0.0};
   double vel[3]  = {0.0, 0.0, 0.0};
   double var[3]  = {0.0, 0.0, 0.0};
   double clk =    0.0;
   double dclk =   0.0;
   double clkrms = 0.0;   

   string satname(gsatdata.sat());
   if( satname.substr(0,1) != "G" ) return -1;

   t_gtime epoT(t_gtime::GPS);
   double satclk = 0.0;
   double satclk2 = 1.0;
   
   while (abs(satclk-satclk2) > 1.e-4/CLIGHT){
      satclk2 = satclk;
      epoT = epo - gsatdata.P3()/CLIGHT - satclk;

      if( _gnav->clk( satname, epoT, &clk, &clkrms, &dclk ) < 0 ){

         if( _log && _log->verb() >= 3 )
	     _log->comment(3, "gppp", _site + " satelite " + satname 
			                    + epo.str_ymdhms(" clock not calculated for epoch "));
	 return -1;
      }
      satclk = clk;
   }
   if( _gnav->pos( satname, epoT, xyz, var, vel ) < 0 ) {
      if( _log && _log->verb() >= 3 )
	  _log->comment(3, "gppp", _site + " satelite " + satname
			                 + epo.str_ymdhms(" coordinate not calculated for epoch "));
      return -1;
   }  
   
   // return 0 for eclipsed satelite
   // not impleneted yet
   
   t_gpcv* pcv = 0;
   ColumnVector Satcrd = array2colVec(xyz);

   // phase center ofset correction
   if( _gpcv && (pcv = _gpcv->find(satname,"*",epoT)) != 0 )
   {
     pcv->pco( epoT, Satcrd, G01 );
   }

   // relativistic correction
   satclk -= 2.0 * ( xyz[0]*vel[0] + xyz[1]*vel[1] + xyz[2]*vel[2] ) /CLIGHT /CLIGHT;

   t_gtriple crd(Satcrd(1),Satcrd(2),Satcrd(3));
   gsatdata.addcrd(crd);
   gsatdata.addclk(satclk*CLIGHT);
//   cout << "process satPos A2-end " << satname << _epoch.str_ymdhms() << endl; cout.flush();   
   
#ifdef DEBUG      
   cout << satname 
        << " CRD " << fixed << setprecision(3)
        <<  "  "   << epoT.str_ymdhms()
        << " X "   << setw(14) << Satcrd(1)
        << " Y "   << setw(14) << Satcrd(2)
        << " Z "   << setw(14) << Satcrd(3)
        << " T "   << setw(14) << clk*1000.0*1000.0
        << endl;
#endif

   return 1;
}


// Synchronize ambiguity with respect to observation data.
// First new ambiguity parameters are added.
// Second old ambiguity parameters are deleted;
// -----------------------------------------------------
void t_gpppfilter::_syncAmb()
{
   
  // Add ambiguity parameter and appropriate rows/columns covar. matrix   
  map<string, double> mapPRN; 

  vector<t_gsatdata>::iterator it;
  for( it = _data.begin(); it != _data.end(); it++){       // loop over all observations
    mapPRN[it->sat()]=0.0; 

    if ( _param.getParam(_site, t_gpar::AMB, it->sat()) < 0 ) {
      t_gpar newPar(it->site(), t_gpar::AMB, _param.parNumber()+1, it->sat());

      newPar.value   = it->L3() - _gpppModel->cmpObs(_epoch, _param, *it, true );      // first ambiguity value

      newPar.MWWL    = it->MW()/(LAMBWL);
      newPar.L4      = it->L4();

      _param.addParam(newPar);

      Matrix_addRC(_Qx, _param.parNumber(), _param.parNumber());
       double sigAmb = dynamic_cast<t_gsetproc*>(_set)->sig_init_amb();
      _Qx(_param.parNumber(), _param.parNumber()) = sigAmb * sigAmb;

#ifdef DEBUG	
       cout << "AMB was added! For Sat PRN " << it->sat() << " Epoch: " 
            << _epoch.str_ymdhms() << endl;
#endif	 
      
    }
  }  // end loop over all observations

 // Remove ambiguity parameter and appropriate rows/columns covar. matrix

   for (unsigned int i=0; i<=_param.parNumber()-1; i++)
     {
	if (_param[i].parType == t_gpar::AMB) {

	   map<string,double>::iterator prnITER = mapPRN.find(_param[i].prn);
	   if (prnITER == mapPRN.end()) {

#ifdef DEBUG
   cout << "AMB will be removed! For Sat PRN " << _param[i].prn
        << " Epoch: " << _epoch.str_ymdhms() << endl;
#endif	      

      	      Matrix_remRC(_Qx, _param[i].index, _param[i].index);
	      _param.delParam(i);
	      _param.reIndex();
	   }	  
	}	
     }
}


// Set log
// -----------------------------
void t_gpppfilter::_setLog()
{
  string tmp( dynamic_cast<t_gsetout*>(_set)->outputs("ppp") );
  if( !tmp.empty() ){
    substitute(tmp,"$(rec)",_site, false);
    _log = new t_glog;
    _log->tsys(t_gtime::GPS);
    _log->mask(tmp);
    _log->append( dynamic_cast<t_gsetout*>(_set)->append() );
    _log->verb(   dynamic_cast<t_gsetout*>(_set)->verb() );
  }
}


// Set output & return if smooth
// -----------------------------
void t_gpppfilter::_setOut()
{
  string tmp;
  tmp = dynamic_cast<t_gsetout*>(_set)->outputs("flt");
  if( ! tmp.empty() ){
    substitute(tmp,"$(rec)",_site, false);
    _flt = new t_giof;
    _flt->tsys(t_gtime::GPS);
    _flt->mask(tmp);
    _flt->append( dynamic_cast<t_gsetout*>(_set)->append() );
  }

  tmp = dynamic_cast<t_gsetout*>(_set)->outputs("smt");
  if( ! tmp.empty() ){
    substitute(tmp,"$(rec)",_site, false);
    _smt = new t_giof;
    _smt->tsys(t_gtime::GPS);
    _smt->mask(tmp);
    _smt->append( dynamic_cast<t_gsetout*>(_set)->append() );

    // smoothing
    _smooth=true;
  }   
}
