
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <iomanip>
#include <memory>

#include "gproc/gpreproc.h"
#include "gutils/gtypeconv.h"
#include "gmodels/gbancroft.h"

// Constructor
// ----------
t_gpreproc::t_gpreproc(t_gallobs* obs, t_gsetbase* settings)
: t_gmonit("gpreproc"),
  _log(0)
{
   _obs = obs;
   _settings = settings;
   
   _msoffset.clear();
   
   _sigCode  = dynamic_cast<t_gsetproc*>(_settings)->sigma_C();
   _sigPhase = dynamic_cast<t_gsetproc*>(_settings)->sigma_L();
}

// Destructor
// ----------
t_gpreproc::~t_gpreproc()
{ 
}


// run preprocessing from time beg until time end (main method)
// --------------------------------------------------------------
int t_gpreproc::preprocess(string site, const t_gtime& beg, const t_gtime& end, int sampl)
{
   this->_site = site;
//   cout << "start prep for " << _site << endl;
  if ( _obs->nepochs(_site) <= 1) return -1;   

   bool firstEpo = true;
   map<string, t_gobsgnss> epoDataPre;

   double CJ = 0;   // integer ms clk jump   
   
   for(t_gtime epoch = beg; epoch < end || epoch == end; epoch = epoch+sampl) {
     map<string, t_gobsgnss>* epoData;      
//     unique_ptr<map<string, t_gobsgnss> > epoData;
     epoData = _obs->find(_site, epoch);
     if (epoData == 0) continue;
      
     filterData(epoData);
      
     if (firstEpo == true) {
       epoDataPre = *(epoData);
       firstEpo = false;
     }	     

     double Ns = epoData->size();   // total number of viewed satellites
     double Na = Ns;                // number of satellite involved into clk jump detection
     double n  = 0;                 // number of detected jumps in particular epoch
     _sumS = 0;
      
     for (map<string, t_gobsgnss>::iterator it = epoData->begin(); it != epoData->end(); it++) {
	map<string, t_gobsgnss>::iterator itPre = epoDataPre.find(it->first);

	if (itPre == epoDataPre.end()) {Na--; continue;}
	t_gobsgnss* obs1 = &itPre->second;
	t_gobsgnss* obs2 = &it->second;	

#ifdef DEBUG
   cout << "Last epoch: "      << obs2->epoch().str_ymdhms()
        << " PRN: "            << obs2->sat() << fixed << setprecision(3)
	<< " L4: " << setw(14) << obs2->L4()
	<< " MW: " << setw(14) << obs2->MW()
	<< endl;
   cout << "Prev epoch: "      << obs1->epoch().str_ymdhms()
	<< " PRN: "            << obs1->sat() << fixed << setprecision(3)
	<< " L4: " << setw(14) << obs1->L4()
	<< " MW: " << setw(14) << obs1->MW()
	<< endl << endl;
#endif
	 // Detection cycle slip		 
	 if ( _slips(obs1, obs2) != 0) {                    // slip or not handled event
	    Na--;
	    continue;
	 }	
	
	 // clock jump detection
	 if ( _jumps(obs1, obs2) == 1) n++;		 
     }

     if (n == Na) {                                         // clk jump effect at all sats in the same way
	double M = (1e3 * _sumS) / (CLIGHT * n);
	double Js = abs( M-floor(M+0.5) );
	if (Js <= 1e-5) {                                   // it is clock jump instead of common slip
	  CJ += floor(M+0.5);                               // ms jump
          _mbreaks[epoch] = CJ;                             // store for logging	   
	}	
     }
      
     epoDataPre = *epoData;
     _repair(epoData, 1e-3*CJ*CLIGHT);                      // corrected carrier phase
//     cout << "Opravil jsem epoData: " << epoch.str_hms() << " o hodnotu " << CJ << " ms" << endl;
   } 
   
//   cout << "Preprocess for " << site << " finished" << endl;
   return 1;   
}

// check slips
// ---------------------
int t_gpreproc::_slips(t_gobsgnss* gobs1, t_gobsgnss* gobs2)
{  
   
   t_gsys gsys(gobs1->gsys());
   
   bool useGlo = dynamic_cast<t_gsetproc*>(_settings)->useGLONASS();
   if (!useGlo && gobs1->sat().substr(0,1).compare("R") == 0) {
      return -1;
   }
      
   double diff_L4 = abs(gobs2->L4() - gobs1->L4());
//   double diff_MW = abs(gobs2->MW() - gobs1->MW());

#ifdef DEBUG
   cout << "Cycle slip DEBUG" << endl;
   cout << "Last epoch: "      << gobs2->epoch().str_ymdhms()
        << " PRN: "            << gobs2->sat() << fixed << setprecision(3)
	<< " L4: " << setw(14) << gobs2->L4()
	<< " MW: " << setw(14) << gobs2->MW()
	<< endl;
   cout << "Prev epoch: "      << gobs1->epoch().str_ymdhms()
	<< " PRN: "            << gobs1->sat() << fixed << setprecision(3)
	<< " L4: " << setw(14) << gobs1->L4()
	<< " MW: " << setw(14) << gobs1->MW()
	<< endl;
   cout << "diff L4: " << setw(14) << diff_L4
        << " treshold L4: " << setw(14) << _slipThreshold("L4", &gsys)
        << endl << endl;;
#endif   
   
//   if ( diff_L4 > _slipThreshold("L4", &gsys) || diff_MW > _slipThreshold("MW", &gsys) ) {
     if ( diff_L4 > _slipThreshold("L4", &gsys)) {
      
      // set lli if slip detected
      gobs2->addlli( L1, 1 );
      gobs2->addlli( L2, 1 );            
      
      // calculation of cycle slip value 
     double dL1 = gobs2->obs_L(1) - gobs1->obs_L(1); // [m]
     double dL2 = gobs2->obs_L(2) - gobs1->obs_L(2); // [m]
     double dP1 = gobs2->obs_C(1) - gobs1->obs_C(1); // [m]
     double dP2 = gobs2->obs_C(2) - gobs1->obs_C(2); // [m]
      
     double float_slip_L1 = (dP1 - dL1)/(gsys.wavelength(1));   // in cycles
     double float_slip_L2 = (dP2 - dL2)/(gsys.wavelength(2));   // in cycles
     double bound = floor(4*(_sigCode-_sigPhase)/gsys.wavelength(1) + 0.5); // search space

     double minRes = 9999;
     double Res;
		    
     double corr_L1 = 0.0;
     double corr_L2 = 0.0;
     int candidate_slip_L1 = 0;
     int candidate_slip_L2 = 0;
     int int_slip_L1, int_slip_L2;
     for (int_slip_L1 = floor(float_slip_L1 - bound + 0.5); int_slip_L1 <= floor(float_slip_L1 + bound + 0.5); int_slip_L1++) {
       for (int_slip_L2 = floor(float_slip_L2 - bound + 0.5); int_slip_L2 <= floor(float_slip_L2 + bound + 0.5); int_slip_L2++) {
     	 corr_L1 = dL1/gsys.wavelength(1) - int_slip_L1;
	 corr_L2 = dL2/gsys.wavelength(2) - int_slip_L2;
   	 Res = abs(gsys.wavelength(1)*corr_L1 - gsys.wavelength(2)*corr_L2);

	 if (Res < minRes) {
	   minRes = Res;
	   candidate_slip_L1 = int_slip_L1;
	   candidate_slip_L2 = int_slip_L2;
	 }
	  
       }	// end for (determine int slip)	      
     } // end for (determine int slip)
		    
  // if min residuum les then treshold correct phase data
  // if not lli is set
     if (minRes < _slipThreshold("L4", &gsys)) {
       // set estimated slip to t_gobsgnss (NOT IMPLEMENTED YET IN t_gobsgnss)
//       gobs2->setSlip(L1, candidate_slip_L1); 
//       gobs2->setSlip(L2, candidate_slip_L2);	
 
     }
      		    		    		    
#ifdef DEBUG
 cout << obs.epoch().str_ymdhms("EPOCH: ") << " PRN " << prn << endl;
 cout << fixed << setprecision(3);		    
 cout << "dP1: " << dP1 << endl;
 cout << "dP2: " << dP2 << endl;
 cout << "dP4: " << dP1-dP2<< endl;		    
 cout << "float slip L1: " << float_slip_L1 << endl;
 cout << "float slip L2: " << float_slip_L2 << endl;
 cout << "bound:" << bound << endl;
 cout << "Min res = " << minRes << endl;		    
 cout << "Int slip L1: " << candidate_slip_L1 << endl;
 cout << "Int slip L2: " << candidate_slip_L2 << endl;
 cout << endl;		    
#endif		    		    		            
   
      pair<string,string> sl;
      if (candidate_slip_L1 != 0)
	   sl.first  = int2str(candidate_slip_L1);
      else sl.first  = "n/a";
      if (candidate_slip_L2 != 0)
	   sl.second  = int2str(candidate_slip_L2);
      else sl.second  = "n/a";      
      _mslips[gobs2->epoch()][gobs2->sat()] = sl;

     return 1;      
   } // end if
   else return 0;
}

// Compute threshold for cycle slip
// -------------------------------------
double t_gpreproc::_slipThreshold(string LC, t_gsys* gsys, double sampl)
{
  double iono = 0.4/60;    // maximum iono jump/sec  ??? set via XML ???

  double f1 = gsys->frequency(1); // first  band frequency
  double f2 = gsys->frequency(2); // second band frequency
  double k1 = f1/(f1-f2);
  double k2 = f2/(f1-f2);
  double k3 = f1/(f1+f2);
  double k4 = f2/(f1+f2);

  double alfa = (f1*f1)/(f1*f1 - f2*f2);
   
  if (LC.compare("L4") == 0) {
     return 4*sqrt(_sigPhase*_sigPhase + _sigPhase*_sigPhase) + alfa*iono*sampl;
  } else if (LC.compare("MW") == 0) {
     return 4*sqrt(k1*k1*_sigPhase*_sigPhase +
		   k2*k2*_sigPhase*_sigPhase +
		   k3*k3*_sigCode *_sigCode  +
		   k4*k4*_sigCode *_sigCode); 
  } else return 0.0;      
}
	  
	  
// check coherency between range and phase caused by clock jump
// --------------------------------------------------------------
int t_gpreproc::_jumps(t_gobsgnss* gobs1, t_gobsgnss* gobs2)
{              
  t_gsys gsys(gobs1->gsys());
  string prn = gobs2->sat();
   
   bool useGlo = dynamic_cast<t_gsetproc*>(_settings)->useGLONASS();
   if (!useGlo && gobs2->sat().substr(0,1).compare("R") == 0) {
      return -1;
   }   
   
  double dL1 = gobs2->obs_L(1) - gobs1->obs_L(1); // [m]
  double dL2 = gobs2->obs_L(2) - gobs1->obs_L(2); // [m]
  double dP1 = gobs2->obs_C(1) - gobs1->obs_C(1); // [m]
  double dP2 = gobs2->obs_C(2) - gobs1->obs_C(2); // [m]      
  double dp, dl; 
   
  if (!double_eq(dP1,0.0) && !double_eq(dP2,0.0)) dp = (dP1 + dP2)/2;
  else if (!double_eq(dP1,0.0)) dp = dP1;
  else if (!double_eq(dP2,0.0)) dp = dP2;
  else return -1;
   
  if (!double_eq(dL1,0.0) && !double_eq(dL2,0.0)) dl = (dL1 + dL2)/2;
  else if (!double_eq(dL1,0.0)) dl = dL1;
  else if (!double_eq(dL2,0.0)) dl = dL2;
  else return -1;
   
  double S = dp - dl;              // jump detection observable
  double sig = 5;                  // sigma of S
  double k = 1e-3 * CLIGHT - 3*sig; // jump threshold

  if (abs(S) >= k) {   // candidate of clock jump
    _sumS += S; 
    return 1;   
  }
   return 0;
}

// Filter data (del zero obs)
// --------------------
void t_gpreproc::filterData(map<string, t_gobsgnss>* epoData)
{
  map<string, t_gobsgnss>::iterator it = epoData->begin();
  while(it != epoData->end()) 
  {
     if ( double_eq(it->second.P3(), 0.0) || double_eq(it->second.L3(), 0.0) ) 
     {

#ifdef DEBUG
       cout << "t_gpreproc::filterData(VEC): erasing data, ";
       cout << "epo: " << it->second.epoch().str_hms() << ", ";
       cout << "prn: " << it->second.sat() << endl; cout.flush();
#endif

     epoData->erase(it++);
     }     
       
     if( it == epoData->end() ) break;     
     it++;
  }
}


// save int log-file
// -----------
void t_gpreproc::jumpLog(t_glog* glog)
{

   ostringstream os;
   os << endl;
   os << "Receiver clk. jump (" << _site << ")" << endl;
   os << "--------------------\n" << endl;
   os << "YYYY-MM-DD HH:MM:SS[TIM]" << setw(10) << right << "Phase[ms]" << endl;
   glog->write( os.str() ); os.str(""); os.clear();

   map<t_gtime, int>::iterator itBr;
   for(itBr = _mbreaks.begin(); itBr != _mbreaks.end(); itBr++)
   {
  	  os << itBr->first.str_ymdhms() << setw(10) << right << itBr->second << endl;
	
  	  glog->write( os.str() ); os.str(""); os.clear();
   }   

   // clear _mbreaks for next station
   _mbreaks.clear();
}


// save int log-file
// -----------
void t_gpreproc::slipLog(t_glog* glog)
{

   ostringstream os, oss, osss;
   int total = 0;
   
   os << endl;
   os << "\nCycle slips" << endl;
   os << "------------------\n" << endl;
   os << "Total slips: ";
   oss << "YYYY-MM-DD HH:MM:SS[TIM]" << setw(10) << right << "PRN"
                                     << setw(10) << right << "elev"
                                     << setw(10) << right << "Slip L1"
                                     << setw(10) << right << "Slip L2" << endl;

   map<t_gtime, map<string,pair<string,string> > >::iterator itSl;
   for (itSl = _mslips.begin(); itSl != _mslips.end(); itSl++)
   {
     total += itSl->second.size();
     osss << itSl->first.str_ymdhms();
     map<string,pair<string,string> >::iterator it;
     for (it = itSl->second.begin(); it != itSl->second.end(); it++)
     {
	   osss << setw(10) << right << it->first   // PRN
	        << setw(10) << right << fixed << setprecision(2);

	   if (double_eq(_elevation(_site, itSl->first, it->first), 0.0)) osss << "n/a";
	   else osss << _elevation(_site, itSl->first, it->first);

	   osss << setw(10) << right << it->second.first   // slip L1
   	        << setw(10) << right << it->second.second;   // slip L2	
     }
     osss << endl;
   }
   os << total  << endl;

   glog->write( os.str()   );
   glog->write( oss.str()  );
   glog->write( osss.str() );


   // clear _mslips for next stations
   _mslips.clear();
}

// Logging
// -----------------
void t_gpreproc::logging(t_glog* glog)
{
   jumpLog(glog);
   slipLog(glog);
}



// Set navigation file (rinexn or sp3)
// --------------------------
void t_gpreproc::setNav(t_gallnav* nav)
{
   this->_nav = nav;
}



// Compute elevation
// --------------------
double t_gpreproc::_elevation(const string& site, t_gtime epoch, string prn)
{

   if (_nav == 0) return 0;
   
  t_gposition banc(_obs, _nav);

  t_gtriple pos;
  vector<t_gsatdata> satdata = _obs->obs(site, epoch);

//  bancx.satPos(satdata);  // compute satellite position and clocks

  if (banc.aprPos(satdata, _pos) <= 0) return 0.0;   // compute aprox position

  banc.satData(satdata, _pos);            // add ele, rho, azimuth

  vector<t_gsatdata>::iterator iter;
  for (iter = satdata.begin(); iter != satdata.end(); iter++)
  {
     if (iter->sat().compare(prn) == 0) return iter->ele()*180/M_PI;
     else continue;
  }  
  return 0.0;
}

// Repair phase due to clk jump
// -----------------------------------
void t_gpreproc::_repair(map<string, t_gobsgnss>* epoData, double dL)
{
  for (map<string, t_gobsgnss>::iterator it = epoData->begin(); it != epoData->end(); it++)
  {
      t_gsys gsys(it->second.gsys());
      it->second.addobs(t_gobs::str2gobs("L1"), it->second.getobs(L1) + dL/gsys.wavelength(1));
      it->second.addobs(t_gobs::str2gobs("L2"), it->second.getobs(L2) + dL/gsys.wavelength(2));     
  }
   
}

// Set site name
// -------------------
void t_gpreproc::setSite(string site)
{
   this->_site = site;
}
