
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <iomanip>
#include <sstream>

#include "gset/gsetbase.h"

using namespace std;
using namespace pugi;

// Constructor
// ----------
t_gsetbase::t_gsetbase()
{
  _log       = 0;
  _name      = "";
  _delimiter = "  "; // only for nodes/elements
}


// Destructor
// ----------
t_gsetbase::~t_gsetbase()
{}


// Read/set glog
int t_gsetbase::glog(t_glog* glog )
{
  _log = glog;

  return 0;
}


// Read from file
// ----------
int t_gsetbase::read(const string& file)
{
  boost::mutex::scoped_lock lock(_mutex);
  _name = file;

  if( ! (_irc = _doc.load_file( _name.c_str() )) ){ 
    if( _log ) _log->comment(0, "gsetbase" , "XML-config not read file " + _name + " " + string(_irc.description()) );
    else                             cerr << "XML-config not read file " + _name + " " + string(_irc.description()) << endl;
    return -1;
  }

  if( _log ) _log->comment(0, "gsetbase" , "XML-config read from file " + _name );

  lock.unlock();
  this->check();

  return 0;
}


// Read from istream
// ----------
int t_gsetbase::read_istream(istream& is)
{
  boost::mutex::scoped_lock lock(_mutex);

  if( ! (_irc = _doc.load( is ) ) ){
    if( _log ) _log->comment(0, "gsetbase" , "XML-config not read istream: " + string(_irc.description()) );
    else                             cerr << "XML-config not read istream: " + string(_irc.description()) << endl;
    return -1;
  }
   
  if( _log ) _log->comment(0, "gsetbase" , "XML-config read from istream");

  lock.unlock();
  this->check();

  return 0;
}


// Write to file
// ----------
int t_gsetbase::write(const string& file)
{
  boost::mutex::scoped_lock lock(_mutex);

  ofstream of;
  string name(file);   // if( name.empty() ) name = _name;  // ???
  try{
      of.open(name.c_str());
  }catch( fstream::failure e ){
    cerr << "Warning: Exception opening file " << name << ": " << e.what() << endl;
    return -1;
  }
   
  _doc.save_file(name.c_str(), _delimiter.c_str(), pugi::format_indent);

  of.close();

  if( _log ) _log->comment(0,"gsetbase","XML-config saved: " + name );
  else               cout << "gsetbase:  XML-config saved: " + name << endl;

  return 0;
}


// Write to ostream
// ----------
int t_gsetbase::write_ostream(ostream& os)
{
  boost::mutex::scoped_lock lock(_mutex);
  _doc.save(os, _delimiter.c_str(), pugi::format_indent);
  
  return 0;
}

// Return first string of value based on key
// ----------
string t_gsetbase::strval( const string& elem, const string& subelem )
{
  boost::mutex::scoped_lock lock(_mutex);
  return _strval( elem, subelem );
}


// Return set of value based on key
// ----------
set<string> t_gsetbase::setval( const string& elem, const string& subelem )
{
  boost::mutex::scoped_lock lock(_mutex);
  return _setval( elem, subelem );
}


// Return vector of value based on key
// ----------
vector<string> t_gsetbase::vecval( const string& elem, const string& subelem )
{
  boost::mutex::scoped_lock lock(_mutex);
  return _vecval( elem, subelem );
}


// Return first string of value based on key
// ----------
string t_gsetbase::_strval( const string& elem, const string& subelem )
{
  string word;

  istringstream is( _doc.child(XMLKEY_ROOT).child( elem.c_str() ).child_value( subelem.c_str() ));
   
  if( is >> word ) return word;
  return "";
}


// Return set of value based on key
// ----------
set<string> t_gsetbase::_setval( const string& elem, const string& subelem )
{
  set<string> vals;
  string word;

  istringstream is( _doc.child(XMLKEY_ROOT).child( elem.c_str() ).child_value( subelem.c_str() ));
  while( is >> word ) vals.insert( word );
               
  return vals;
}

// Return vector of value based on key
// ----------
vector<string> t_gsetbase::_vecval( const string& elem, const string& subelem )
{
  vector<string> vec;
  string str;

  istringstream is( _doc.child(XMLKEY_ROOT).child( elem.c_str() ).child_value( subelem.c_str() ));
  while( is >> str ) vec.push_back( str );
               
  return vec;
}

   
// check/set node
// ----------
/*
xml_node t_gsetbase::_default_node(xml_node& node, const char* n)
{ 
  string s(n);
  transform(s.begin(), s.end(), s.begin(), ::tolower);
   
  if( ! node ) cerr << " NODE[" << node.name() << "] " << node << " does'not exists !\n";
  else         cerr << " NODE[" << node.name() << "] " << node << " does exists .... ok \n";
   
  xml_node elem = node.child(s.c_str());
   
  if( ! elem ) cerr << " ELEM[" << s << "] " << elem << " does'not exists !\n";
  else         cerr << " ELEM[" << s << "] " << elem << " does exists .... ok \n";
   
   
  if( ! elem ){ cerr << "create ELEM " << s << " " << elem << endl;  elem = node.append_child(s.c_str()); }
  else          cerr << "exists ELEM " << s << " " << elem << endl;
   
  if( ! elem ){
    cout << "NODE " << elem << " could not have been created :" << n << "\n";
    if( _log ) _log->comment(0,"gsetbase","warning - cannot create element " + s);
    else               cerr << "gsetbase:  warning - cannot create element " + s + "\n";
  }else{
    cerr << "succesfully created ELEM " << s << " " << elem << endl;
  }
   
  return elem;
}
*/

// check/set node
// ----------
xml_node t_gsetbase::_default_node(xml_node& node, const char* n, const char* val, bool reset)
{
#ifdef DEBUG
  cout << "creating node [" << n << "]  for parent [" << node.name() << "]  set with value [" << val << "]\n";
#endif

  string s(n);
  transform(s.begin(), s.end(), s.begin(), ::tolower);
   
  // remove if to be reset
  if( strcmp(val,"") && reset ) node.remove_child(s.c_str());

#ifdef DEBUG
  if( ! node ) cerr << " NODE[" << node.name() << "] " << node << " does'not exists !\n";
  else         cerr << " NODE[" << node.name() << "] " << node << " does exists .... ok \n";
#endif
   
  xml_node elem = node.child(s.c_str());

#ifdef DEBUG
  if( ! elem ) cerr << " ELEM[" << s << "] " << elem << " does'not exists !\n";
  else         cerr << " ELEM[" << s << "] " << elem << " does exists .... ok \n";
#endif

  if( ! elem ) elem = node.append_child(s.c_str());


  if( ! elem ){
    if( _log ) _log->comment(0,"gsetbase","warning - cannot create element " + s);
    else               cerr << "gsetbase:  warning - cannot create element " + s + "\n";
  }

  if( elem && strcmp(val,"") && reset ){
    elem.append_child(pugi::node_pcdata).set_value(val);
  }

  return elem;
}


// check attributes & set default
// ----------
void t_gsetbase::_default_attr(xml_node& node, const char* n, const string& v, bool reset)
{ 
  if( node.attribute(n).empty() ) node.append_attribute(n);
  if( strlen(node.attribute(n).value()) == 0 || reset ) node.attribute(n).set_value(v.c_str());
   
  string s = node.attribute(n).value();
  transform(s.begin(), s.end(), s.begin(), ::tolower);
  node.attribute(n).set_value(s.c_str());
}


// check attributes & set default
// ----------
void t_gsetbase::_default_attr(xml_node& node, const char* n, const bool& v, bool reset)
{ 
  if( node.attribute(n).empty() ) node.append_attribute(n);
  if( strlen(node.attribute(n).value()) == 0 || reset ) node.attribute(n).set_value(v);
}


// check attributes & set default
// ----------
void t_gsetbase::_default_attr(xml_node& node, const char* n, const int& v, bool reset)
{ 
  if( node.attribute(n).empty() ) node.append_attribute(n);
  if( strlen(node.attribute(n).value()) == 0 || reset ) node.attribute(n).set_value(v);
}


// check attributes & set default
// ----------
void t_gsetbase::_default_attr(xml_node& node, const char* n, const double& v, bool reset)
{ 
  if( node.attribute(n).empty() ) node.append_attribute(n);
  if( strlen(node.attribute(n).value()) == 0 || reset ) node.attribute(n).set_value(v);
}


// XML help header (optional text comment)
// ----------
void t_gsetbase::help_header()
{
  cerr << "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\" ?> \n"
       << "<!DOCTYPE config>\n\n"
       << "<config>\n";
}


// XML help footer (optional text comment)
// ----------
void t_gsetbase::help_footer()
{
  cerr << "</config>\n";
}
