
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <iomanip>
#include <sstream>
#include <boost/thread/mutex.hpp>

#include "gset/gsetflt.h"

using namespace std;
using namespace pugi;

// Constructor
// ----------
t_gsetflt::t_gsetflt() 
 : t_gsetbase()
{
  _noise_clk      = 1000.0;
  _noise_crd      = 100.0;
  _rndwk_ztd      = 3;
  _method         = "SRCF";
}


// Destructor
// ----------
t_gsetflt::~t_gsetflt()
{}


// Return value
// ----------
string t_gsetflt::method()
{
   boost::mutex::scoped_lock lock(_mutex);    
   return _doc.child(XMLKEY_ROOT).child(XMLKEY_FLT).attribute("method").value();
}


// Return value
// ----------
double t_gsetflt::noise_clk()
{
   boost::mutex::scoped_lock lock(_mutex);  
   return _doc.child(XMLKEY_ROOT).child(XMLKEY_FLT).attribute("noise_clk").as_double();
}


// Return value
// ----------
double t_gsetflt::noise_crd()
{
   boost::mutex::scoped_lock lock(_mutex);  
   return _doc.child(XMLKEY_ROOT).child(XMLKEY_FLT).attribute("noise_crd").as_double();
}


// Return value
// ----------
double t_gsetflt::rndwk_ztd()
{
   boost::mutex::scoped_lock lock(_mutex);  
   return _doc.child(XMLKEY_ROOT).child(XMLKEY_FLT).attribute("rndwk_ztd").as_double();
}


// settings check
// ----------
void t_gsetflt::check()
{
  boost::mutex::scoped_lock lock(_mutex);

  // check existence of nodes/attributes
  xml_node parent = _doc.child(XMLKEY_ROOT);
  xml_node node = _default_node(parent, XMLKEY_FLT);

  // check existence of attributes
  _default_attr(node,"method",    _method    );
  _default_attr(node,"noise_clk", _noise_clk );
  _default_attr(node,"noise_crd", _noise_crd );
  _default_attr(node,"rndwk_ztd", _rndwk_ztd );

  lock.unlock();
}


// help body
// ----------
void t_gsetflt::help()
{
  boost::mutex::scoped_lock lock(_mutex);
  cerr << " <!-- filter description:\n"
       << "\t method        .. type of filtering method (SRCF, Kalman)\n"
       << "\t noise_clk     .. white noise for clocks \n"
       << "\t noise_crd     .. white noise for coordinates \n"
       << "\t rndwk_ztd     .. random walk process for ZTD [mm/sqrt(hour)] \n"
       << "   -->\n";

  cerr << " <filter \n"
       << "   method=\""         <<  _method      << "\" \n"
       << "   noise_clk=\""      <<  _noise_clk   << "\" \n"
       << "   noise_crd=\""      <<  _noise_crd   << "\" \n"
       << "   rndwk_ztd=\""      <<  _rndwk_ztd   << "\" \n"
       << "  /> \n\n";

  lock.unlock();
}
