
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <iomanip>
#include <sstream>
#include <boost/thread/mutex.hpp>

#include "gset/gsetgen.h"

using namespace std;
using namespace pugi;

// Constructor
// ----------
t_gsetgen::t_gsetgen() 
 : t_gsetbase()
{  
}

// Destructor
// ----------
t_gsetgen::~t_gsetgen()
{}


// Return gtime
// ----------
t_gtime t_gsetgen::beg()
{
  boost::mutex::scoped_lock lock(_mutex);
  string str = _doc.child(XMLKEY_ROOT).child(XMLKEY_GEN).child_value( "beg" );
  substitute(str,"\"","");
  t_gtime gt(t_gtime::GPS);
   
  if( str.empty() ) 
    return t_gtime::first_time(t_gtime::GPS);
   
  gt.from_str( "%Y-%m-%d %H:%M:%S", trim(str) );
  return gt;
}


// Return gtime
// ----------
t_gtime t_gsetgen::end()
{
  boost::mutex::scoped_lock lock(_mutex);
  string str = _doc.child(XMLKEY_ROOT).child(XMLKEY_GEN).child_value( "end" );
  substitute(str,"\"","");
  t_gtime gt(t_gtime::GPS);
   
  if( str.empty() ) 
    return t_gtime::last_time(t_gtime::GPS);
   
  gt.from_str( "%Y-%m-%d %H:%M:%S", trim(str) );
  return gt;
}


// Return sampling
// ----------
int t_gsetgen::sampling()
{
  // only >=1Hz sampling supported !
  boost::mutex::scoped_lock lock(_mutex);
  string str  = _doc.child(XMLKEY_ROOT).child(XMLKEY_GEN).child_value( "int" );
  return str2int(str);
}
   

// Return set
// ----------
set<string> t_gsetgen::sys()
{
  boost::mutex::scoped_lock lock(_mutex);
  return t_gsetbase::_setval( XMLKEY_GEN, "sys" );
}


// Return set
// ----------
set<string> t_gsetgen::rec()
{
  boost::mutex::scoped_lock lock(_mutex);               
  return t_gsetbase::_setval( XMLKEY_GEN, "rec" );
}


// Return set
// ----------
set<string> t_gsetgen::sat()
{
  boost::mutex::scoped_lock lock(_mutex);               
  return t_gsetbase::_setval( XMLKEY_GEN, "sat" );
}


// Return set
// ----------
set<string> t_gsetgen::obs()
{
  boost::mutex::scoped_lock lock(_mutex);
  return t_gsetbase::_setval( XMLKEY_GEN, "obs" );
}


// settings check
// ----------
void t_gsetgen::check()
{
  boost::mutex::scoped_lock lock(_mutex);

  // check existence of nodes/attributes
  xml_node parent = _doc.child(XMLKEY_ROOT);
  xml_node node = _default_node(parent, XMLKEY_GEN);

  _default_node(node, "BEG", "");            // all!
  _default_node(node, "END", "");	     // all!
  _default_node(node, "SYS", "");	     // all!
  _default_node(node, "SAT", "");	     // all!
  _default_node(node, "REC", "");	     // none

  xml_node child = _default_node(node, "INT", int2str(10).c_str());   // 30 s
   
  lock.unlock();

  // TO CHECK USER-SETUP CONSISTENCY
  // support >= 1Hz
  if( floor(this->sampling()) < 1 ){
    _default_node(node, "INT", "30", true); // reset!
    cerr << "gsetgen: sampling rate settings above 1Hz is not supported: reset to "
         << this->sampling() << ".\n";
  }
   
}


// help body
// ----------
void t_gsetgen::help()
{
  boost::mutex::scoped_lock lock(_mutex);
  cerr << " <!-- general description:\n"
       << "\t beg    .. beg time          (default: all)\n"
       << "\t end    .. end time          (default: all)\n"
       << "\t int    .. data sampling     (default: 30s)\n"
       << "\t sys    .. GNSS systems      (default: all)\n"
       << "\t rec    .. GNSS receivers    (default: all)\n"
       << "\t sat    .. GNSS satellites   (default: all)\n"
//       << "\t obs    .. GNSS observations (default: all)\n"
       << "   -->\n";
   
  cerr << " <gen>\n"
//       << "   <beg>" << _beg.str("%Y-%m-%d %H:%M:%S") << "</beg>\n"
//       << "   <end>" << _end.str("%Y-%m-%d %H:%M:%S") << "</end>\n"
       << "   <sys>" << DEF_GNSSSYST << "</sys>\n"
       << "   <sat>" << DEF_SATELLIT << "</sat>\n"
       << "   <rec>" << DEF_RECEIVER << "</rec>\n"
       << "   <int>" << DEF_SAMPLING << "</int>\n"
//       << "   <obs>" << DEF_OBSERVAT << "</obs>\n"
       << " </gen>\n\n";

  lock.unlock();
}
