
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <iostream>
#include <iomanip>
#include <sstream>
#include <boost/thread/mutex.hpp>

#include "gset/gsetproc.h"

using namespace std;
using namespace pugi;

// Constructor
// ----------
t_gsetproc::t_gsetproc() 
 : t_gsetbase()
{
  _phase          = true;
  _tropo          = true;
  _useGLONASS     = false;
  _tropo_model    = "saastamoinen";
  _tropo_mf       = "gmf";
  _sigma_C        = 5;
  _sigma_L        = 0.01;
  _sig_init_ztd   = 0.1;
  _sig_init_amb   = 1000.0;
  _sig_init_crd   = 100.0;
  _minimum_elev   = 10;
}


// Destructor
// ----------
t_gsetproc::~t_gsetproc()
{}


// Return value
// ----------
bool   t_gsetproc::tropo()
{
  boost::mutex::scoped_lock lock(_mutex);  
  return _doc.child(XMLKEY_ROOT).child(XMLKEY_PROC).attribute("tropo").as_bool();
}

// Return value
// ----------
bool   t_gsetproc::useGLONASS()
{
  boost::mutex::scoped_lock lock(_mutex);  
  return _doc.child(XMLKEY_ROOT).child(XMLKEY_PROC).attribute("GLONASS").as_bool();
}

// Return value
// ----------
bool   t_gsetproc::phase()
{
  boost::mutex::scoped_lock lock(_mutex);  
  return _doc.child(XMLKEY_ROOT).child(XMLKEY_PROC).attribute("phase").as_bool();
}


// Return value
// ----------
double t_gsetproc::sigma_C()
{
   boost::mutex::scoped_lock lock(_mutex);  
   return _doc.child(XMLKEY_ROOT).child(XMLKEY_PROC).attribute("sigma_C").as_double();
}


// Return value
// ----------
double t_gsetproc::sigma_L()
{
   boost::mutex::scoped_lock lock(_mutex);  
   return _doc.child(XMLKEY_ROOT).child(XMLKEY_PROC).attribute("sigma_L").as_double();
}


// Return value
// ----------
double t_gsetproc::sig_init_ztd()
{
   boost::mutex::scoped_lock lock(_mutex);  
   return _doc.child(XMLKEY_ROOT).child(XMLKEY_PROC).attribute("sig_init_ztd").as_double();
}


// Return value
// ----------
double t_gsetproc::sig_init_crd()
{
   boost::mutex::scoped_lock lock(_mutex);
   return _doc.child(XMLKEY_ROOT).child(XMLKEY_PROC).attribute("sig_init_crd").as_double();
}

// Return value
// ----------
double t_gsetproc::sig_init_amb()
{
   boost::mutex::scoped_lock lock(_mutex);
   return _doc.child(XMLKEY_ROOT).child(XMLKEY_PROC).attribute("sig_init_amb").as_double();
}

// Return value
// ----------
double t_gsetproc::minimum_elev()
{
   boost::mutex::scoped_lock lock(_mutex);  
   return _doc.child(XMLKEY_ROOT).child(XMLKEY_PROC).attribute("minimum_elev").as_double();
}


// Return value
// ----------
string t_gsetproc::tropo_model()
{
   boost::mutex::scoped_lock lock(_mutex);  
   return _doc.child(XMLKEY_ROOT).child(XMLKEY_PROC).attribute("tropo_model").value();
}


// Return value
// ----------
string t_gsetproc::tropo_mf()
{
   boost::mutex::scoped_lock lock(_mutex);  
   return _doc.child(XMLKEY_ROOT).child(XMLKEY_PROC).attribute("tropo_mf").value();
}


// settings check
// ----------
void t_gsetproc::check()
{
  boost::mutex::scoped_lock lock(_mutex);

  // check existence of nodes/attributes
  xml_node parent = _doc.child(XMLKEY_ROOT);
  xml_node node = _default_node(parent, XMLKEY_PROC);

  // check existence of attributes
  _default_attr(node,"phase", _phase);
  _default_attr(node,"tropo", _tropo);
  _default_attr(node,"tropo_model", _tropo_model);
  _default_attr(node,"tropo_mf", _tropo_mf);
  _default_attr(node,"sigma_C", _sigma_C);
  _default_attr(node,"sigma_L", _sigma_L);
  _default_attr(node,"sig_init_crd", _sig_init_crd);
  _default_attr(node,"sig_init_ztd", _sig_init_ztd);
  _default_attr(node,"sig_init_amb", _sig_init_amb);   
  _default_attr(node,"minimum_elev", _minimum_elev);
  _default_attr(node,"GLONASS", _useGLONASS);
  
  lock.unlock();
}


// help body
// ----------
void t_gsetproc::help()
{
  boost::mutex::scoped_lock lock(_mutex);
  cerr << " <!-- process description:\n"
       << "\t GLONASS [0,1] .. use GLONASS observations\n"
       << "\t phase [0,1]   .. use carrier phase data\n"
       << "\t tropo [0,1]   .. estimate troposphere\n"
       << "\t tropo_model   .. tropospheric model (SAAS, GPT, ...)\n"
       << "\t tropo_mf      .. tropospheric mapping function (COSZ, GMF, ...)\n"
       << "\t sigma_C       .. C/code measurements precision [m]\n"
       << "\t sigma_L       .. L/phase measurements precision\n"
       << "\t sig_init_crd  .. accuracy of initial coordinates [m]\n"
       << "\t sig_init_ztd  .. accuracy of initial zenith path delay [m]\n"
       << "\t minimum_elev  .. elevation angle cut-off [degree]\n"
       << "   -->\n";
   
  cerr << " <process \n"
       << "   phase=\""          <<  _phase          << "\" \n"
       << "   GLONASS=\""        <<  _useGLONASS     << "\" \n"
       << "   tropo=\""          <<  _tropo          << "\" \n"
       << "   tropo_model=\""    <<  _tropo_model    << "\" \n"
       << "   tropo_mf=\""       <<  _tropo_mf       << "\" \n"
       << "   sigma_C=\""        <<  _sigma_C        << "\" \n"
       << "   sigma_L=\""        <<  _sigma_L        << "\" \n"
       << "   sig_init_crd=\""   <<  _sig_init_crd   << "\" \n"
       << "   sig_init_ztd=\""   <<  _sig_init_ztd   << "\" \n"
       << "   minimum_elev=\""   <<  _minimum_elev   << "\" \n"
       << " />\n\n";
  lock.unlock();
}
