
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <iomanip>
#include <sstream>
#include <boost/thread/mutex.hpp>

#include "gset/gsetrec.h"

using namespace std;
using namespace pugi;

// Constructor
// ----------
t_gsetrec::t_gsetrec() 
 : t_gsetbase()
{
  _id             = "";
  _name           = "";
  _X              = 0.0;
  _Y              = 0.0;
  _Z              = 0.0;
  _dX             = 0.0;
  _dY             = 0.0;
  _dZ             = 0.0;
  _ZTD            = 0.0;
  _ZTD_constr     = 0.0;
  _CRD_constr     = "";
}


// Destructor
// ----------
t_gsetrec::~t_gsetrec()
{}


// Return value
// ----------
string t_gsetrec::id()
{
  boost::mutex::scoped_lock lock(_mutex);
  return _doc.child(XMLKEY_ROOT).child(XMLKEY_REC).attribute("id").value();
}

double t_gsetrec::X()
{
  boost::mutex::scoped_lock lock(_mutex);
  return _doc.child(XMLKEY_ROOT).child(XMLKEY_REC).attribute("X").as_double();
}

double t_gsetrec::Y()
{
  boost::mutex::scoped_lock lock(_mutex);
  return _doc.child(XMLKEY_ROOT).child(XMLKEY_REC).attribute("Y").as_double();
}

double t_gsetrec::Z()
{
  boost::mutex::scoped_lock lock(_mutex);
  return _doc.child(XMLKEY_ROOT).child(XMLKEY_REC).attribute("Z").as_double();
}

double t_gsetrec::dX()
{
  boost::mutex::scoped_lock lock(_mutex);
  return _doc.child(XMLKEY_ROOT).child(XMLKEY_REC).attribute("dX").as_double();
}

double t_gsetrec::dY()
{
  boost::mutex::scoped_lock lock(_mutex);
  return _doc.child(XMLKEY_ROOT).child(XMLKEY_REC).attribute("dY").as_double();
}

double t_gsetrec::dZ()
{
  boost::mutex::scoped_lock lock(_mutex);
  return _doc.child(XMLKEY_ROOT).child(XMLKEY_REC).attribute("dZ").as_double();
}


int t_gsetrec::getCRD(t_gtriple& crd)
{
   if (double_eq(X(), 0.0) &&
       double_eq(Y(), 0.0) &&
       double_eq(Z(), 0.0))
     return 0;
   else { crd[0] = X();
          crd[1] = Y();
          crd[2] = Z();
   return 1;}
   
}

bool t_gsetrec::valid_crd()
{
  if ( double_eq ( X(), 0.0 ) &&
       double_eq ( Y(), 0.0 ) &&
       double_eq ( Z(), 0.0 ) )
        return false;
   else return true;
}

bool t_gsetrec::valid_ztd()
{
  if ( ZTD() > 0.5 && ZTD() < 5.0 ) return true;
  else return false;
}


double t_gsetrec::ZTD()
{
  boost::mutex::scoped_lock lock(_mutex);
  return _doc.child(XMLKEY_ROOT).child(XMLKEY_REC).attribute("ZTD").as_double();
}

double t_gsetrec::ZTD_constr()
{
  boost::mutex::scoped_lock lock(_mutex);
  return _doc.child(XMLKEY_ROOT).child(XMLKEY_REC).attribute("ZTD_constr").as_double();
}

string t_gsetrec::CRD_constr()
{
  boost::mutex::scoped_lock lock(_mutex);
  return _doc.child(XMLKEY_ROOT).child(XMLKEY_REC).attribute("CRD_constr").value();
}


// settings check
// ----------
void t_gsetrec::check()
{
  boost::mutex::scoped_lock lock(_mutex);

  // check existence of nodes/attributes
  xml_node parent = _doc.child(XMLKEY_ROOT);
  xml_node node = _default_node(parent, XMLKEY_REC);

  // check existence of attributes
  _default_attr(node,"X", _X);
  _default_attr(node,"Y", _Y);
  _default_attr(node,"Z", _Z);
  _default_attr(node,"dX", _dX);
  _default_attr(node,"dY", _dY);
  _default_attr(node,"dZ", _dZ);
  _default_attr(node,"ZTD", _ZTD);
  _default_attr(node,"ZTD_constr", _ZTD_constr);
  _default_attr(node,"CRD_constr", _CRD_constr);

  // PROBABLY LATER CHECK USER-SETUP CONSISTENCY

  lock.unlock();
}


// help body
// ----------
void t_gsetrec::help()
{
  boost::mutex::scoped_lock lock(_mutex);
  cerr << " <!-- receiver description:\n"
       << "\t id     .. ID name\n"
       << "\t name   .. full name\n"
       << "\t desc   .. description\n"
       << "\t X[YZ]  .. X,Y,Z-coordinate [m]\n"
       << "\t dX[YZ] .. X,Y,Z-eccentricity [m]\n"
       << "   -->\n";
   
  cerr << " <rec"
       <<        " id=\""   << _id   << "\""
       <<        " name=\"" << _name << "\""
       <<        " desc=\"" << ""    << "\" >\n"
       << "  <set"
       <<        " beg=\"" << ""     << "\""
       <<        " end=\"" << ""     << "\""
       <<        " rec=\"" << ""     << "\""
       <<        " ant=\"" << ""     << "\""
       <<        " X=\""   << _X     << "\""
       <<        " Y=\""   << _Y     << "\""
       <<        " Z=\""   << _Z     << "\""
       <<        " dX=\""  << _dX    << "\""
       <<        " dY=\""  << _dY    << "\""
       <<        " dZ=\""  << _dZ    << "\""
       <<  "  />\n"
       <<  " </rec>\n";

  lock.unlock();
}
