
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <iostream>
#include <iomanip>

#include "gutils/gfreq.h"

using namespace std;


// ---------------------------
// class GSYS
// ---------------------------

// constructor
// ---------
t_gsys::t_gsys(GSYS sys) : _gsys(GPS)
{}

// constructor
// ---------
t_gsys::t_gsys(string sys)
 : _gsys(LAST_GSYS)
{
 if( sys.size() > 0 ) _gsys = str2gsys(sys);
}

// constructor
// ---------
t_gsys::t_gsys(char c)
{
  _gsys = char2gsys(c);
}


// convert GSYS enum to GSYS string
// ---------- 
string t_gsys::gsys2str( GSYS sys )
{
// PROBLEM WITH STATIC FUNCTION
//  boost::mutex::scoped_lock lock(_mutex);

  switch( sys ){
   case GPS  : return "GPS";
   case GLO  : return "GLO";
   case GAL  : return "GAL";
   case BDS  : return "BDS";
   case SBAS : return "SBAS";
   case QZSS : return "QZSS";
   case LAST_GSYS : return "LAST_GSYS";
     
   default :  cerr << "*** warning: unknown GNSS system \n";
        // cannot be while static function !!!
        // if( _log ) _log->comment(0,"gobsgnss","warning: unknown frequency code ");
        //     else 
  }
   
  return "LAST_GSYS";
}


// convert GSYS enum to GSYS char
// ---------- 
char t_gsys::gsys2char( GSYS sys )
{
// PROBLEM WITH STATIC FUNCTION
//  boost::mutex::scoped_lock lock(_mutex);

  switch( sys ){
   case GPS  : return 'G';
   case GLO  : return 'R';
   case GAL  : return 'E';
   case BDS  : return 'C';
   case SBAS : return 'S';
   case QZSS : return 'J';
   case LAST_GSYS : return 'X';
     
   default :  cerr << "*** warning: unknown GNSS system \n";
        // cannot be while static function !!!
        // if( _log ) _log->comment(0,"gobsgnss","warning: unknown frequency code ");
        //     else 
  }
   
  return 'X';
}


// convert GSYS string to GSYS enum
// ---------- 
GSYS t_gsys::str2gsys( string s )
{
// PROBLEM WITH STATIC FUNCTION
//  boost::mutex::scoped_lock lock(_mutex);

  if( s.size() == 0 ){
    cout << "*** warning: not defined GNSS system code [NULL]\n";
    return LAST_GSYS;
  }

  transform(s.begin(), s.end(), s.begin(), ::toupper);
  
  if(      s == "GPS"  || s == "NAVSTAR"   ) return GPS;
  else if( s == "GLO"  || s == "GLONASS"   ) return GLO;
  else if( s == "GAL"  || s == "GALILEO"   ) return GAL;
  else if( s == "COMP" || s == "COMPASS"   ) return BDS;
  else if( s == "BDS"  || s == "BEIDOU"    ) return BDS;
  else if( s == "SBAS" || s == "EGNOS"     ) return SBAS;
  else if( s == "QZSS" || s == "JAXA"      ) return QZSS;
  else if( s == "GXXX" || s == "LAST_GSYS" ) return LAST_GSYS;
  else
    cerr << "*** warning: not defined GNSS system code [" + s[0] << "]\n";

  return LAST_GSYS;
}


// convert GSYS string to GSYS enum
// ---------- 
char t_gsys::str2char(string s)
{
  if( s.size() > 0 ) gsys2char( str2gsys(s) );
  return 'X';
}	


// convert GSYS char to GSYS enum
// ---------- 
GSYS t_gsys::char2gsys(char c)
{
  if(      c == 'G' ) return GPS;
  else if( c == 'R' ) return GLO;
  else if( c == 'E' ) return GAL;
  else if( c == 'C' ) return BDS;
  else if( c == 'S' ) return SBAS;
  else if( c == 'J' ) return QZSS;
  else if( c == 'X' ) return LAST_GSYS;
  else  
    cerr << "*** warning: not defined GNSS system char [" + c << "]\n";

  return LAST_GSYS;
}


// convert GSYS char to GSYS enum
// ---------- 
string t_gsys::char2str(char c)
{
  return gsys2str( char2gsys(c) );
}	


// convert satellite name
// ---------- 
string t_gsys::eval_sat(string sat, GSYS sys )
{
  istringstream is( sat ); // .substr(0,3) NE!
  GSYS gnss = sys;
  int  svn  =  0;
  char chr  = 'G';        // if empty, use GPS
  size_t l  = is.str().length();
   
  if( l == 0 ) return "X00";

  if( l < 3 or sat[0] == ' ' )  is        >> svn;
  else                          is >> chr >> svn;

  if( is.fail() ){ return "X00"; }
	
  if( chr != 'X' ) gnss = char2gsys( chr );

  switch( gnss ){
    case GPS       : chr = 'G'; break;
    case GLO       : chr = 'R'; break;
    case GAL       : chr = 'E'; break;
    case BDS       : chr = 'C'; break;
    case SBAS      : chr = 'S'; break;
    case QZSS      : chr = 'J'; break;
    case LAST_GSYS : chr = 'X'; break;
    default        : chr = 'X'; break;
  }

  ostringstream os;
  os << setw(1) << chr << setfill('0') << setw(2) << svn;

#ifdef DEBUG3
  cout << "  ARGM: [" << sat << "] " << gsys2str( sys )
       << "  EVAL: " << gsys2str( gnss ) << " [" << chr << "] " << svn
       << "  RESU: " << os.str() << " " << l << endl;
#endif
  return os.str();
}


/*
// get system-specific wavelength
// ---------- 
double t_gsys::wavelength( string prn )
{
  cout << " t_gsys::wavelenght not yet implemented !\n";
  return 0;
//  if( str2gsys(prn) )
//  return ...;
}


// // get system-specific frequency
// ---------- 
double t_gsys::freq( string prn )
{
  cout << " t_gsys::freq not yet implemented !\n";
  return 0;
}
*/


// // get system-specific frequency for band i
// ---------- 
double t_gsys::frequency( int i )
{
  switch( _gsys ){
     
   case GPS  : switch( i ){
                 case 1 : return G01_F;
                 case 2 : return G02_F;
                 case 5 : return G05_F;
                default : return 0.0; }
     
   case GLO  : switch( i ){
                 case 1 : return R01_F(0); // channel need to be implemented !!!
                 case 2 : return R02_F(0); // channel need to be implemented !!!
                default : return 0.0; }
     
   case GAL  : switch( i ){
                 case 1 : return G01_F;
                 case 5 : return E05_F;
                 case 6 : return E06_F;
                 case 7 : return E07_F;
                 case 8 : return E08_F;
                default : return 0.0; }
     
   case BDS  : switch( i ){
                 case 1 : return C01_F;
                 case 2 : return C02_F;
                 case 6 : return C06_F;
                 case 7 : return C07_F;
                default : return 0.0; }
     
   case QZSS : switch( i ){
                 case 1 : return J01_F;
                 case 2 : return J02_F;
                 case 5 : return J05_F;
                 case 6 : return J06_F;
                default : return 0.0; }
     
   case SBAS : switch( i ){
                 case 1 : return S01_F;
                 case 5 : return S05_F;
                default : return 0.0; }
     
   case LAST_GSYS : return 0.0;
  }

  return 0.0;
}


// return system-dependenty frequency multiplier factor
// ---------- 
double t_gsys::factor( int i )
{
  switch( _gsys ){
     
   case GPS  : switch( i ){
                 case 1 : return G01_MLT;
                 case 2 : return G02_MLT;
                 case 5 : return G05_MLT;
                default : return 0.0; }
     
   case GLO  : switch( i ){
                 case 1 : return R01_MLT;
                 case 2 : return R02_MLT;
                default : return 0.0; }
     
   case GAL  : switch( i ){
                 case 1 : return G01_MLT;
                 case 5 : return E05_MLT;
                 case 6 : return E06_MLT;
                 case 7 : return E07_MLT;
                 case 8 : return E08_MLT;
                default : return 0.0; }
     
   case BDS  : switch( i ){
                 case 1 : return C01_MLT;
                 case 2 : return C02_MLT;
                 case 6 : return C06_MLT;
                 case 7 : return C07_MLT;
                default : return 0.0; }
     
   case SBAS : switch( i ){
                 case 1 : return J01_MLT;
                 case 2 : return J02_MLT;
                 case 5 : return J05_MLT;
                 case 6 : return J06_MLT;
                default : return 0.0; }
     
   case QZSS : switch( i ){
                 case 1 : return S01_MLT;
                 case 5 : return S05_MLT;
                default : return 0.0; }
     
   case LAST_GSYS : return 0.0;
  }

  return 0.0;
}


// get system-specific wavelength for band i
// ---------- 
double t_gsys::wavelength( int i )
{
  double frq = this->frequency( i );
  if( frq != 0.0 ) return CLIGHT/frq;
     
  return 0.0;
}


// set GSYS from string
// ---------- 
void t_gsys::from_string( string sys )
{
  boost::mutex::scoped_lock lock(_mutex);
  _gsys = str2gsys( sys );
}


// set GSYS from enum
// ---------- 
void t_gsys::from_gsys( GSYS sys )
{
  boost::mutex::scoped_lock lock(_mutex);
  _gsys = sys;
}


// overloaded equivalence operator
// ----------
bool t_gsys::operator==(const string& sys) const
{
//  transform(sys.begin(), sys.end(), sys.begin(), ::toupper);
  if( _gsys == str2gsys(sys)  ) return true;
  return false;
}


// overloaded equivalence operator
// ----------
bool t_gsys::operator==(const GSYS& sys) const
{
//  if( _gsys == sys.gsys() ) return true;
  return false;
}


// internal conversion to GOBS
// ---------- 
/*
GSYS t_gsys::_to_string( string sys )
{
  transform(sys.begin(), sys.end(), sys.begin(), ::toupper);

  if(      sys == "G" || sys == "GPS"  || sys == "NAVSTAR" ) return GPS;
  else if( sys == "R" || sys == "GLO"  || sys == "GLONASS" ) return GLO;
  else if( sys == "E" || sys == "GAL"  || sys == "GALILEO" ) return GAL;
  else if( sys == "C" || sys == "COMP" || sys == "COMPASS" ) return BDS;  
  else if( sys == "C" || sys == "BDS"  || sys == "BEIDOU" ) return BDS;
  else if( sys == "S" || sys == "SBAS" || sys == "EGNOS"   ) return SBAS;
  else if( sys == "J" || sys == "QZSS" || sys == "JAXA"    ) return QZSS;
  else if( sys == "X" || sys == "GXXX" || sys == "LAST_GSYS" ) return LAST_GSYS;
  else  
    cerr << "*** warning: not defined GNSS system code " + sys << endl;

  return LAST_GSYS;
}


// internal conversion to string
// ---------- 
string t_gsys::_to_string( GOBS GSYS )
{
  switch( sys ){
   case GPS  : return "GPS";
   case GLO  : return "GLO";
   case GAL  : return "GAL";
   case BDS  : return "BDS";
   case SBAS : return "SBAS";
   case QZSS : return "QZSS";
   case LAST_GSYS : return "LAST_GSYS";
     
   default :  cerr << "*** warning: unknown frequency code \n";
        // cannot be while static function !!!
        // if( _log ) _log->comment(0,"gobsgnss","warning: unknown frequency code ");
        //     else 
  }

  return "LAST_GSYS";
}
*/


// ---------------------------
// class FREQ
// ---------------------------

// get GFRQ enum from string
// ----------
GFRQ t_gfreq::str2gfreq( string freq )
{
   
// PROBLEM WITH STATIC FUNCTION
//  boost::mutex::scoped_lock lock(_mutex);

  transform(freq.begin(), freq.end(), freq.begin(), ::toupper);
   
  if(      freq == "G01" ) return G01;
  else if( freq == "G02" ) return G02;
  else if( freq == "G05" ) return G05;
  else if( freq == "R01" ) return R01;
  else if( freq == "R02" ) return R02;
  else if( freq == "E01" ) return E01;
  else if( freq == "E05" ) return E05;
  else if( freq == "E07" ) return E07;
  else if( freq == "E08" ) return E08;
  else if( freq == "E06" ) return E06;
  else if( freq == "C01" ) return C01;
  else if( freq == "C02" ) return C02;
  else if( freq == "C06" ) return C07;
  else if( freq == "C07" ) return C06;
  else if( freq == "J01" ) return J01;
  else if( freq == "J02" ) return J02;
  else if( freq == "J05" ) return J05;
  else if( freq == "J06" ) return J06;
  else if( freq == "S01" ) return S01;
  else if( freq == "S05" ) return S05;
  else if( freq == "LAST_GFRQ" ) return LAST_GFRQ;
  
  cerr << "*** warning: not defined frequency code [" + freq << "]\n";

  // cannot be while static function !!!
  //  if( _log ) _log->comment(0,"gobsgnss","warning: not defined frequency code " + freq);
  //  else                      cerr << "*** warning: not defined frequency code " + freq << endl;

  return LAST_GFRQ;
}


// get string from GFRQ enum
// ---------- 
string t_gfreq::gfreq2str( GFRQ freq )
{
// PROBLEM WITH STATIC FUNCTION
//  boost::mutex::scoped_lock lock(_mutex); 
   
  switch( freq ){
   case G01 : return "G01";
   case G02 : return "G02";
   case G05 : return "G05";
   case R01 : return "R01";
   case R02 : return "R02";
   case E01 : return "E01";
   case E05 : return "E05";
   case E07 : return "E07";
   case E08 : return "E08";
   case E06 : return "E06";
   case C01 : return "C01";
   case C02 : return "C02";
   case C06 : return "C07";
   case C07 : return "C06";
   case J01 : return "J01";
   case J02 : return "J02";
   case J05 : return "J05";
   case J06 : return "J06";
   case S01 : return "S01";
   case S05 : return "S05";
   case LAST_GFRQ : return "LAST_GFRQ";
     
   default :  cerr << "*** warning: unknown frequency code \n";
        // cannot be while static function !!!
        // if( _log ) _log->comment(0,"gobsgnss","warning: unknown frequency code ");
        //     else 
  }
   
  return "LAST_GFRQ";
}

