
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <cmath>
#include <iostream>
#include <iomanip>

#include "gutils/gconst.h"
#include "gutils/gsysconv.h"

using namespace std;


// XYZ to Ellipsoidal coordinates
// ----------
int xyz2ell(const double* XYZ, double* Ell) 
{    
  const double bell = Aell*(1.0-1.0/Finv) ;
  const double e2   = (Aell*Aell-bell*bell)/(Aell*Aell) ;
  const double e2c  = (Aell*Aell-bell*bell)/(bell*bell) ;
                            
  double nn, ss, zps, hOld, phiOld, theta, sin3, cos3;
   
  ss    = sqrt(XYZ[0]*XYZ[0]+XYZ[1]*XYZ[1]) ;
  zps   = XYZ[2]/ss ;
  theta = atan( (XYZ[2]*Aell) / (ss*bell) );
  sin3  = sin(theta) * sin(theta) * sin(theta);
  cos3  = cos(theta) * cos(theta) * cos(theta);
   
  // Closed formula
  Ell[0] = atan( (XYZ[2] + e2c * bell * sin3) / (ss - e2 * Aell * cos3) );
  Ell[1] = atan2(XYZ[1],XYZ[0]) ;
  nn = Aell/sqrt(1.0-e2*sin(Ell[0])*sin(Ell[0])) ;
  Ell[2] = ss / cos(Ell[0]) - nn;
      
  const int MAXITER = 100;
  for( int ii = 1; ii <= MAXITER; ii++ ){
    nn     = Aell/sqrt(1.0-e2*sin(Ell[0])*sin(Ell[0])) ;
    hOld   = Ell[2] ;
    phiOld = Ell[0] ;
    Ell[2] = ss/cos(Ell[0])-nn ;
    Ell[0] = atan(zps/(1.0-e2*nn/(nn+Ell[2]))) ;
    if ( fabs(phiOld-Ell[0]) <= 1.0e-11 && fabs(hOld-Ell[2]) <= 1.0e-5 ) {
      return 0;
    }
  }
     
  return 1;
}


// XYZ to Topocentric/Local (North, East, Up) coordinates
// ----------
int xyz2neu(const double* XYZ, const double* XYZ_Ref, double* neu)
{
  double ele[3];
  xyz2ell(XYZ_Ref, ele);
   
  ColumnVector r(3);
   
  r << XYZ[0]-XYZ_Ref[0] << XYZ[1]-XYZ_Ref[1] << XYZ[2]-XYZ_Ref[2];
   
  double sinPhi = sin(ele[0]);
  double cosPhi = cos(ele[0]);
  double sinLam = sin(ele[1]);
  double cosLam = cos(ele[1]);
  
  neu[0] = - sinPhi*cosLam * r(1)
           - sinPhi*sinLam * r(2)
           + cosPhi        * r(3);
  
  neu[1] = - sinLam * r(1)
           + cosLam * r(2);
  
  neu[2] = + cosPhi*cosLam * r(1)
           + cosPhi*sinLam * r(2)
           + sinPhi        * r(3);
 
  return 1;
}


// Topocentric/Local (NEU) to XYZ coordinates
// ----------
int neu2xyz(const double* Ell, const double* neu,     double* xyz)
{
   double sinPhi = sin(Ell[0]);
   double cosPhi = cos(Ell[0]);
   double sinLam = sin(Ell[1]);
   double cosLam = cos(Ell[1]);
   
   xyz[0] = - sinPhi*cosLam * neu[0]
            - sinLam        * neu[1]
            + cosPhi*cosLam * neu[2];
   
   xyz[1] = - sinPhi*sinLam * neu[0]
            + cosLam        * neu[1]
            + cosPhi*sinLam * neu[2];
   
   xyz[2] = + cosPhi        * neu[0]
            + sinPhi        * neu[2];

  return 1;
}


// Radial/Along-track/Out-of-plane (Satell-system) to XYZ coordinates
// ---------------
int rao2xyz(const ColumnVector& pos, const ColumnVector& vel,
            const ColumnVector& rao, ColumnVector& xyz)
{                
  ColumnVector along  = vel / vel.norm_Frobenius();
  ColumnVector cross  = crossproduct(pos, vel); cross /= cross.norm_Frobenius();
  ColumnVector radial = crossproduct(along, cross);
                      
  Matrix RR(3,3);
  RR.Column(1) = radial;
  RR.Column(2) = along;
  RR.Column(3) = cross;
  
  xyz = RR * rao;

  return 1;
}

int xyz2ell(t_gtriple &crd, t_gtriple &ell)
{
    const double bell = Aell*(1.0-1.0/Finv) ;
    const double e2   = (Aell*Aell-bell*bell)/(Aell*Aell) ;
    const double e2c  = (Aell*Aell-bell*bell)/(bell*bell) ;

    double nn, ss, zps, hOld, phiOld, theta, sin3, cos3;

    ss    = sqrt(crd[0]*crd[0]+crd[1]*crd[1]) ;
    zps   = crd[2]/ss ;
    theta = atan( (crd[2]*Aell) / (ss*bell) );
    sin3  = sin(theta) * sin(theta) * sin(theta);
    cos3  = cos(theta) * cos(theta) * cos(theta);

    ell.set(0, atan( (crd[2] + e2c * bell * sin3) / (ss - e2 * Aell * cos3) ));
    ell.set(1, atan2(crd[1],crd[0])) ;
    nn     = Aell/sqrt(1.0-e2*sin(ell[0])*sin(ell[0])) ;
    ell.set(2, ss / cos(ell[0]) - nn);

    const int MAXITER = 100;
    for (int ii = 1; ii <= MAXITER; ii++)
    {
        nn     = Aell/sqrt(1.0-e2*sin(ell[0])*sin(ell[0])) ;
        hOld   = ell[2] ;
        phiOld = ell[0] ;
        ell.set(2, ss/cos(ell[0])-nn);
        ell.set(0, atan(zps/(1.0-e2*nn/(nn+ell[2])))) ;

        if ( fabs(phiOld-ell[0]) <= 1.0e-11 && fabs(hOld-ell[2]) <= 1.0e-5 )
        {
            return 0;
        }
    }
    return 1;
}

void xyz2neu(t_gtriple &crd, t_gtriple &crd_site, t_gtriple &neu)
{
    t_gtriple ele(0,0,0);
    xyz2ell(crd_site, ele);

    double r[3];

    r[0] = crd[0] - crd_site[0];
    r[1] = crd[1] - crd_site[1];
    r[2] = crd[2] - crd_site[2];

    double sinPhi = sin(ele[0]);
    double cosPhi = cos(ele[0]);
    double sinLam = sin(ele[1]);
    double cosLam = cos(ele[1]);

    neu.set(0, - sinPhi*cosLam * r[0]
               - sinPhi*sinLam * r[1]
               + cosPhi        * r[2]);

    neu.set(1, - sinLam * r[0]
               + cosLam * r[1]);

    neu.set(2, + cosPhi*cosLam * r[0]
               + cosPhi*sinLam * r[1]
               + sinPhi        * r[2]);
}

void neu2xyz(t_gtriple &ell, t_gtriple &neu, t_gtriple &crd)
{
    double sinPhi = sin(ell[0]);
    double cosPhi = cos(ell[0]);
    double sinLam = sin(ell[1]);
    double cosLam = cos(ell[1]);

    crd.set(0, - sinPhi*cosLam * neu[0]
               - sinLam        * neu[1]
               + cosPhi*cosLam * neu[2]);

    crd.set(1, - sinPhi*sinLam * neu[0]
               + cosLam        * neu[1]
               + cosPhi*sinLam * neu[2]);

    crd.set(2, + cosPhi        * neu[0]
               + sinPhi        * neu[2]);
}
