
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sstream>
#include <iostream>
#include <iomanip>

#include "gutils/gtime.h"
#include "gutils/gtypeconv.h"

using namespace std;

// copy constructor
// ----------
/* NOT NECESSARY
t_gtime::t_gtime( const t_gtime& gt )
  : _mjd(0), _sod(0), _dsec(0.0), _tsys(TAI)
{
  _tsys = gt.tsys();       // original TS
  _mjd  = gt.mjd(false);   // original TS
  _sod  = gt.sod(false);   // original TS
  _dsec = gt.dsec(false);  // original TS
}
*/

// constructor 
// ----------
t_gtime::t_gtime(const t_tsys& ts)
  : _mjd(0), _sod(0), _dsec(0.0), _tsys(UTC)
{
  from_time(time(NULL), 0.0, true);  // input UTC time-system
  _tsys = ts;                        // now switch to required TS
}


// constructor 
// ----------
t_gtime::t_gtime(const time_t& tt, const double& ds, const t_tsys& ts )
  : _mjd(0), _sod(0), _dsec(0.0), _tsys(UTC)
{
  from_time(tt, ds, true); // input UTC time-system
  _tsys = ts;              // now switch to required TS
}


// constructor 
// ----------
t_gtime::t_gtime(const int& yr, const int& mn, 
                 const int& dd, const int& hr, 
                 const int& mi, const int& sc, 
                 const double& ds, const t_tsys& ts )
  : _mjd(0), _sod(0), _dsec(0.0), _tsys(ts)
{
  from_ymdhms(yr,mn,dd,hr,mi,sc + ds,true);
  // old style
  // from_ymd(yr,mn,dd,static_cast<int>(hr*3600 + mi*60 + sc),ds,true);
}


// constructor
// ----------
t_gtime::t_gtime(const int& gw, const int& dw, const int& sd, 
                 const double& ds, const t_tsys& ts)
  : _mjd(0), _sod(0), _dsec(0.0), _tsys(ts)
{
  from_gwd(gw, dw, sd, ds, true);
}


// constructor
// ----------
t_gtime::t_gtime(const int& gw, const double& sow, const t_tsys& ts)
  : _mjd(0), _sod(0), _dsec(0.0), _tsys(ts)
{
  from_gws(gw, sow, true);
}


// constructor
// ----------
t_gtime::t_gtime(const int& mjd, const int& sd,
                 const double& ds, const t_tsys& ts)
  : _mjd(0), _sod(0), _dsec(0.0), _tsys(ts)
{ 
 from_mjd(mjd, sd, ds, true);
}


// destructor
// ----------
t_gtime::~t_gtime()
{}


// set first time
// ----------
t_gtime t_gtime::first_time(const t_tsys& ts)
{
  boost::mutex::scoped_lock lock(_static_mutex);
  t_gtime tmp( ts );
  tmp.from_str(BEG_FORMAT, BEG_TIME, false);
  return tmp;
}


// set last time
// ----------
t_gtime t_gtime::last_time(const t_tsys& ts)
{
  boost::mutex::scoped_lock lock(_static_mutex);
  t_gtime tmp( ts );
  tmp.from_str(END_FORMAT, END_TIME, false);
  return tmp;
}


// set current time // acccuracy of 1 sec!
// ----------
t_gtime t_gtime::current_time(const t_tsys& ts)
{
  boost::mutex::scoped_lock lock(_static_mutex);
  t_gtime tmp(UTC);
  tmp.from_time(time(NULL), 0.0, true);
  tmp.tsys( ts );
//  cout << tmp.str(" gtime::current_time: %Y-%m-%d %H:%M:%S[%T] %v") << endl; cout.flush();
  return tmp;
}


// set from time
// -----------
int t_gtime::from_time(const time_t& tt, const double& ds, const bool& conv )
{
  struct tm *tm = gmtime( &tt );
  from_ymd( tm->tm_year+1900,  tm->tm_mon+1,   tm->tm_mday,
	    tm->tm_hour*3600 + tm->tm_min*60 + tm->tm_sec,
	    ds, conv );
  _norm_dsec();
  _norm_sod();
  return 0;
}


// set from GPS week and second of week
// -----------
int t_gtime::from_gws(const int& gw, const double& sow, const bool& conv)
{ 
  int dw = floor(sow/86400);
  _mjd = 44244 + 7*gw + dw;
  _sod = floor(sow - dw*86400);
  _dsec = sow - dw*86400 - _sod;
  if(conv) _to_tai(_tsys);
  _norm_dsec();
  _norm_sod();
//  cout << " test: " << gw << " " << sow << " ----> " << this->gwk() << " " << this->sow() << endl; cout.flush();
  return 0;
}


// set from GPS week and dow of week
// -----------
int t_gtime::from_gwd(const int& gw, const int& dw, const int& sd, 
                      const double& ds, const bool& conv )
{
  _mjd = 44244 + 7*gw + dw;
  _sod = sd;
  _dsec = ds;
  if(conv) _to_tai(_tsys);
  _norm_dsec();
  _norm_sod();
  return 0;
}


// set from year, month, day, second of day, fractional seconds
// -----------
int t_gtime::from_ymd(const int& yr, const int& mn, const int& dd, 
                      const int& sd, const double& ds, const bool& conv )
{
  int year(yr);
  _norm_year(year);
  _mjd  = _ymd_mjd(year,mn,dd);
  _sod  = sd;
  _dsec = ds;
  if(conv) _to_tai(_tsys);
  _norm_dsec();
  _norm_sod();
//  cout  << " mjd = " << _mjd << " sod = " << _sod << " dsec = " << _dsec << endl;
  return 0;
}


// set from year, month, day, hour, minute, float seconds
// -----------
int t_gtime::from_ymdhms(const int& yr, const int& mn, 
                         const int& dd, const int& h, 
                         const int& m, const double& s, 
                         const bool& conv )
{
  int year(yr);
  _norm_year(year);
  _mjd  = _ymd_mjd(year,mn,dd);
   
  _sod  = (h*3600 + m*60) + floor(s);
  _dsec = s - 1.0*(floor(s));

  if(conv) _to_tai(_tsys);
  _norm_dsec();
  _norm_sod();
//  cout  << " mjd = " << _mjd << " sod = " << _sod << " dsec = " << _dsec << endl;
  return 0;
}


// set from MJD
// -----------
int t_gtime::from_mjd(const int& mjd, const int& sd, 
                      const double& ds, const bool& conv)
{
  _mjd  = mjd;
  _sod  = sd;
  _dsec = ds;
  if(conv) _to_tai(_tsys);
  _norm_dsec();
  _norm_sod();
  return 0;
}


// from string
// ----------
int t_gtime::from_str(const string& ifmt, const string& idat, const bool& conv )
{
  int cYMD = _ymd;
  int cHMS = _hms;
  int y,b,d,h,m,s;   y = b = d = h = m = s = 0;
  size_t idx;
  size_t fmtadd = 0;
  size_t datadd = 0;
  size_t fmtpos = 0;
  size_t datpos = 0;

  // search and process keys from left to right
  while( (fmtpos < ifmt.length()) and ( idx = ifmt.find('%',fmtpos) ) != string::npos ){
    
    fmtpos++; // a priori encrease
     
    for( int i=0; i < MAX_DT; ++i ){
      string tmp = TD[i];
 
      // too short ifmt format string, skip!
      if( ifmt.length() - idx - 1 < tmp.length() ) continue;

      // dat string not identified, skipped !
      if( ifmt.substr(idx+1,tmp.length()).compare( tmp ) != 0 ) continue;

      fmtpos  = idx + tmp.length() + 1; // end of format reading
      datpos  = idx - fmtadd + datadd;  // sum of all idat characters
      fmtadd +=       tmp.length() + 1; // sum of all ifmt characters

      if(      !tmp.compare("Y")){ y =   str2int(idat.substr(datpos,4));  datadd+=4; cYMD+=t_tymd(_year);
      }else if(!tmp.compare("y")){ y =   str2int(idat.substr(datpos,2));  datadd+=2; cYMD+=t_tymd(_year);
      }else if(!tmp.compare("b")){ b =       mon(idat.substr(datpos,3));  datadd+=3; cYMD+=t_tymd(_mon);
      }else if(!tmp.compare("m")){ b =   str2int(idat.substr(datpos,2));  datadd+=2; cYMD+=t_tymd(_mon);
      }else if(!tmp.compare("d")){ d =   str2int(idat.substr(datpos,2));  datadd+=2; cYMD+=t_tymd(_day);
      }else if(!tmp.compare("j")){ b = 1;                                            cYMD+=t_tymd(_mon);
	                               d =   str2int(idat.substr(datpos,3));  datadd+=3; cYMD+=t_tymd(_day);
      }else if(!tmp.compare("H")){ h =   str2int(idat.substr(datpos,2));  datadd+=2; cHMS+=t_thms(_hour);
      }else if(!tmp.compare("M")){ m =   str2int(idat.substr(datpos,2));  datadd+=2; cHMS+=t_thms(_min);
      }else if(!tmp.compare("S")){ s =   str2int(idat.substr(datpos,2));  datadd+=2; cHMS+=t_thms(_sec);
      }else if(!tmp.compare("W")){ y = 1980;                                         cYMD+=t_tymd(_year);
                                   b = 1;                                            cYMD+=t_tymd(_mon);
  	                               d+= 6+str2int(idat.substr(datpos,4))*7;datadd+=4;
      }else if(!tmp.compare("w")){ d+=   str2int(idat.substr(datpos,1));  datadd+=1; cYMD+=t_tymd(_day);
      }else if(!tmp.compare("v")){ s+=   str2int(idat.substr(datpos,6));  datadd+=6; cYMD+=t_tymd(_day);
	                                                                                 cHMS+=t_thms(_hour)
	                                                                                      +t_thms(_min)
                                                                                          +t_thms(_sec);
      }else if(!tmp.compare("s")){ s =   str2int(idat.substr(datpos,5));  datadd+=5; cHMS+=t_thms(_hour)
	                                                                                      +t_thms(_min)
	                                                                                      +t_thms(_sec);
      }else if(!tmp.compare("I")){ y = 2000;                                         cYMD+=t_tymd(_year);
	                           b = 1;                                                cYMD+=t_tymd(_mon);
	                           d = 1-51544+(str2int(idat.substr(datpos,11)));
	                                                                                 cYMD+=t_tymd(_day);
	                           s = 86400*(  str2int(idat.substr(datpos,11))
					                      -(str2int(idat.substr(datpos,11))) );
	                                                                      datadd+=11;cHMS+=t_thms(_hour)
	                                                                                      +t_thms(_min) 
	                                                                                      +t_thms(_sec);
      }else if(!tmp.compare("J")){ y = 2000;                                         cYMD+=t_tymd(_year);
 	                               b = 1;                                            cYMD+=t_tymd(_mon);
                                   d = 1-51544.0 + str2int(idat.substr(datpos,5));
	                                                                      datadd+=5; cYMD+=t_tymd(_day);
      }else if(!tmp.compare("T")){ cerr << "input time-system conversion not yet supported here!\n";

      }else{ cerr << " warning : t_gtime - unknown date/time identifier [" << tmp << "]\n";
      }
    }
  }

  if( cYMD == ( _year | _mon | _day ) and cHMS == ( _hour | _min | _sec ) ){
     
    from_ymd( y, b, d, h*3600+m*60+s, 0.0, conv );
     
#ifdef DEBUG
  cout << " from_str = " << y
                  << " " << b
                  << " " << d
                  << " " << h
                  << " " << m 
                  << " " << s
                  << " " << _mjd
                  << " " << _sod
                  << "\n";
#endif
     
    return 0;
  }
   
  if( cYMD == ( _year | _mon | _day ) ){

#ifdef DEBUG
  cout << " from_str = " << y
                  << " " << b
                  << " " << d
                  << " " << h
                  << " " << m 
                  << " " << s
                  << "\n";
#endif

    from_ymd( y, b, d, 0, 0.0, conv );
    return 0;
  }

  return 1;
}


// convert from TAI to LOC
// ----------
t_gtime t_gtime::local() const
{
  t_gtime tmp(*this);
  tmp.tsys(LOC);
  return tmp;
}


// convert to TAI
// ----------
t_gtime t_gtime::tai() const
{
  t_gtime tmp(*this);
  tmp.tsys(TAI);
  return tmp;
}


// convert from TAI to UTC
// ----------
t_gtime t_gtime::utc() const
{
  t_gtime tmp(*this);
  tmp.tsys(UTC);
  return tmp;
}


// convert from TAI to GPS
// ----------
t_gtime t_gtime::gps() const
{
  t_gtime tmp(*this);
  tmp.tsys(GPS);
  return tmp;
}


// time zone difference (LOC - UTC) [sec]
// ----------
int t_gtime::tzdiff() const
{
  switch( _tsys ){
    case USER  : return 0;
    case TAI   : return 0;
    case UTC   : return 0;
    case GPS   : return 0;
    case LOC   : { time_t local, utc;
                   local =   time(NULL);
                   utc   = mktime(gmtime(&local));
                   return int(local - utc);
                 }
     default    : cerr << " warning : t_tsys not recognized !\n";
  }
  return 0;
}


// day saving time (summer - winter time, e.g. CEST - CET) [sec]
// ----------
int t_gtime::dstime() const
{  
  switch( _tsys ){
    case USER  : return 0;
    case TAI   : return 0;
    case UTC   : return 0;
    case GPS   : return 0;
    case LOC   : { 
                   time_t tmp;
                   struct tm tm;
                   ymd(tm.tm_year, tm.tm_mon, tm.tm_mday, false); // TAI !
                   hms(tm.tm_hour, tm.tm_min, tm.tm_sec,  false); // TAI !

                   tm.tm_year  -=  1900;
                   tm.tm_mon   -=  1;
                   tm.tm_sec   -=  leapsec() + tzdiff();  // convert tai to LOC !
                   tm.tm_isdst  = -1;
       
                   tmp = mktime(&tm);
                   return int(tm.tm_isdst * 3600 ); // [sec]
                 }
    default    : cerr << " warning : t_tsys not recognized !\n";
  }
  return 0;
}


// get TAI-UTC leap seconds [sec]
// ----------
int t_gtime::leapsec() const
{  
  int leap = 0.0;
  for( int i = sizeof(leapseconds)/sizeof(*leapseconds) -1; i >= 0; --i ){
    double mjd_time = _mjd + _sod/86400.0 - leapseconds[i]._leap/86400.0;
    double mjd_leap = static_cast<double>( _ymd_mjd( leapseconds[i]._year,
			                             leapseconds[i]._mon,
		 	                             leapseconds[i]._day  ));

    if( mjd_leap+1.0 <= mjd_time ){
      leap = leapseconds[i]._leap;
      break;
    }
  } 
  return leap;
}


// getleap year
// ----------
int t_gtime::leapyear(const int& y) const
{
  if( (y%100 + y%400) == 0.0 ) return 1;
  return 0;
}


// get modified julian date
// b = true (convert to TSYS)
// ----------
int t_gtime::mjd(const bool& conv) const
{
  // do not convert
  if( !conv || _tsys == USER || _tsys == TAI) return _mjd;

  // do convert
  t_gtime gt(*this);
  gt._tai_to(_tsys);                // set new TSYS
  return gt.mjd(false);             // == gt._mjd (new TSYS)!
}

// get double modified jujian date
// -----------------------------------
double t_gtime::dmjd(const bool& conv) const
{
  double ret;
  ret = (double)mjd() +
        (double)sod()/24/60/60 + dsec()/24/60/60;
   
  return ret;
}


// get seconds of day
// b = true (convert to TSYS)
// ----------
int t_gtime::sod(const bool& conv) const
{
  // do not convert
  if( !conv || _tsys == USER || _tsys == TAI ) return _sod;

  // do convert
  t_gtime gt(*this);
  gt._tai_to(_tsys);                // set new TSYS
  return gt.sod(false);             // == gt._sod (new TSYS)!
}


// get dsec
// b = true (convert to TSYS)
// ----------
double t_gtime::dsec(const bool& conv) const
{
  // do not convert
  if( !conv || _tsys == USER || _tsys == TAI ) return _dsec;

  // do convert
  t_gtime gt(*this);
  gt._tai_to(_tsys);                // set new TSYS
  return gt.dsec(false);            // == gt._dsec (new TSYS)!
}


// get time_t
// ----------
time_t t_gtime::tim() const
{
  int y,b,d,h,m,s;
  ymd(y,b,d,true);  // convert to UTC ?
  hms(h,m,s,true);  // convert to TAI UTC ?

  struct tm t;
  t.tm_year  = y - 1900;
  t.tm_mon   = b - 1;
  t.tm_mday  = d;
  t.tm_hour  = h;
  t.tm_min   = m;
  t.tm_sec   = s;
  t.tm_isdst = -1;

  return mktime(&t);
}


// get day of year
// ----------
int t_gtime::doy() const
{
  int y, m, d;
  ymd(y,m,d,true);

  int doy = d + leapyear(y);
  for( int i=0; i<m and i<13; ++i)  doy += monthdays[i];
  return doy;
}


// get GPS week
// ----------
int t_gtime::gwk() const
{
  return static_cast<int>((mjd(true)-44244.0)/7.0);  // return true TS
}


// get day of GPS week
// ----------
int t_gtime::dow() const
{
  return static_cast<int>( mjd(true) - 44244.0 - gwk()*7 );
}


// get seconds of GPS week
// ----------
int t_gtime::sow() const
{
  return static_cast<int>( dow()*86400 + sod() );
}


// get year
// ----------
int t_gtime::year() const
{  
  int y, m, d;
  ymd( y, m, d, true);

  return static_cast<int>( y );
}


// get month
// ----------
int t_gtime::mon() const
{  
  int y, m, d;
  ymd( y, m, d, true);

  return static_cast<int>( m );
}


// get day
// ----------
int t_gtime::day() const
{  
  int y, m, d;
  ymd( y, m, d, true);

  return static_cast<int>( d );
}


// get hour
// ----------
int t_gtime::hour() const
{ 
  return static_cast<int>( (sod(true)%86400)/3600 );
}


// get minutes
// ----------
int t_gtime::min() const
{ 
  return static_cast<int>( (sod(true)%3600)/60 );
}


// get seconds
// ----------
int t_gtime::sec() const
{ 
  return static_cast<int>( sod(true)%60 );
}


// get hour
// ----------
void t_gtime::hms( int& h, int& m, int& s, bool conv ) const
{
  h = static_cast<int>( ( sod(conv)%86400 )/3600 );
  m = static_cast<int>( ( sod(conv)%3600  )/60 );
  s = static_cast<int>( ( sod(conv)%60    ) );

//cout << "HMS = " << h <<  " " << m << " " << s << endl;
}

   
// MJD -> YMD
// ----------
void t_gtime::ymd( int& y, int& m, int& d, bool conv ) const
{
  int    jj,mm,dd;
  long   ih, ih1, ih2 ;
  double t1, t2,  t3, t4;
  double mjd = 0.0;

  mjd = this->mjd( conv ) + (this->sod( conv ))/86400;

  //  DO NOT USE THIS DUE TO UNDERFLOW
  //mjd = this->mjd( conv ) + (this->sod( conv ) + this->dsec( conv ))/86400; 
   
  t1  =  1.0 + mjd - fmod( mjd, 1.0 ) + 2400000.0;
  t4  =  fmod( mjd, 1.0 );
  ih  =  int( (t1 - 1867216.25)/36524.25 );
  t2  =  t1 + 1 + ih - ih/4;
  t3  =  t2 - 1720995.0;
  ih1 =  int( (t3 - 122.1)/365.25 );
  t1  =  ih1*365.25 - fmod( ih1*365.25, 1.0 );
  ih2 =  int( (t3 - t1)/30.6001 );
  dd  =  t3 - t1 - int( ih2*30.6001 ) + t4;
  mm  =  ih2 - 1;
  if( ih2 > 13 ) mm = ih2 - 13;
  jj  = ih1;
  if( mm <= 2 ) jj = jj + 1;
  y=jj;
  m=mm;
  d=dd;
//cout << "YMD = " << y <<  " " << m << " " << d << endl;
}


// to string
// ----------
string t_gtime::str(const string& ofmt, const bool& conv ) const
{
  t_gtime gt(*this);

  char cstr[12];
  string str = ofmt; // copy of requested format
  size_t idx =  0;
  int y,b,d,h,m,s;
  gt.ymd( y, b, d, conv );
  gt.hms( h, m, s, conv );

  int y2 = gt.yr(y);

  // search and process all % identificators
  while( ( idx = str.find('%') ) != string::npos and idx+1 <= str.length() ){
   
    bool replace = false;
    for( int i=0; i < MAX_DT; ++i ){
      string tmp = TD[i];

      // dat string not identified, skipped !
      if( str.substr(idx+1,tmp.length()).compare( tmp ) != 0 ) continue;

//    cout << "substitute " << tmp << " " << gt.tsys()<<endl;
      if(      !tmp.compare("Y")){ sprintf(cstr,"%04i",y);
      }else if(!tmp.compare("y")){ sprintf(cstr,"%02i",y2);
      }else if(!tmp.compare("b")){ sprintf(cstr,"%3.3s", gt.mon(b).c_str());
      }else if(!tmp.compare("m")){ sprintf(cstr,"%02i",b);
      }else if(!tmp.compare("d")){ sprintf(cstr,"%02i",d);
      }else if(!tmp.compare("j")){ sprintf(cstr,"%03i",gt.doy());
      }else if(!tmp.compare("H")){ sprintf(cstr,"%02i",h);
      }else if(!tmp.compare("M")){ sprintf(cstr,"%02i",m);
      }else if(!tmp.compare("S")){ sprintf(cstr,"%02i",s);
      }else if(!tmp.compare("W")){ sprintf(cstr,"%04i",gt.gwk());
      }else if(!tmp.compare("w")){ sprintf(cstr,"%01i",gt.dow());
      }else if(!tmp.compare("v")){ sprintf(cstr,"%6i", gt.sow());
      }else if(!tmp.compare("s")){ sprintf(cstr,"%5i", gt.sod( conv ));
      }else if(!tmp.compare("J")){ sprintf(cstr,"%5i", gt.mjd( conv ));
      }else if(!tmp.compare("I")){ sprintf(cstr,"%11.5f", gt.mjd( conv )
					                + gt.sod( conv )/86400.0
					                + gt.dsec( conv )/86400.0);
      }else if(!tmp.compare("T")){ sprintf(cstr,"%3s", sys().c_str() );
      }else{ cerr << " warning : t_gtime - unknown date/time identifier [" << tmp << "]\n";
      }

      str.replace(idx,2,cstr);
      replace = true;
    }
    // replace unknown % occurance
    if( !replace ) str.replace(idx,1,"");
  }
  return str;
}


// to string
// ----------
string t_gtime::str_ymd(const string& str, const bool& conv) const
{
  char cstr[12];
  int y=0, b=0, d=0;
  this->ymd( y, b, d, conv );
  sprintf(cstr,"%04i-%02i-%02i",y,b,d);
  return str + string(cstr);
}


// to string
// ----------
string t_gtime::str_hms(const string& str, const bool& conv) const
{
  char cstr[12];
  int h=0, m=0, s=0;
  this->hms( h, m, s, conv );
  sprintf(cstr,"%02i:%02i:%02i",h,m,s);
  return str + string(cstr);
}


// to string
// ----------
string t_gtime::str_ymdhms(const string& str, 
			   const bool& ts,
			   const bool& conv) const
{
  char cstr[25];
  int y=0, b=0, d=0, h=0, m=0, s=0;
  this->ymd( y, b, d, conv );
  this->hms( h, m, s, conv );
  if( ts ) sprintf(cstr,"%04i-%02i-%02i %02i:%02i:%02i[%3s]",y,b,d,h,m,s, sys().c_str());
  else     sprintf(cstr,"%04i-%02i-%02i %02i:%02i:%02i",y,b,d,h,m,s);
  return str + string(cstr);
  // return str_ymd() + " " + str_hms();
}


// get yr (1980-2079 only)
// ----------
int t_gtime::yr(const int& y) const
{  
  if( y <= 2079 and y >= 2000 ) return y-2000;
  if( y <  2000 and y >= 1980 ) return y-1900;
   
  return -1;
}


// get month 3-char string
// ----------
string t_gtime::mon(const int& m) const
{ 
  if( m < 1 or m > 12 ) return "XXX";
  return MON[m];
}


// get month number
// ----------
int t_gtime::mon(const string& str) const
{ 
  for(int i = 0; i<12; ++i){
     if( str.compare(MON[i]) == 0 )
       return i+1; }
  return 0;
}


// get system identifier
// ----------
string t_gtime::sys() const
{ 
  string ts( t_tstr[_tsys] );
  return ts;
}


// operator reduce [sec] [TAI]
// ----------
double t_gtime::operator-(const t_gtime& t) const
{
  return ( this->diff(t) );
}


// operator minus [TAI]
// ----------
t_gtime t_gtime::operator-(const double& sec) const
{
  t_gtime tmp(*this);
  tmp.add_sec(  - static_cast<int>(sec));
  tmp.add_dsec( - static_cast<double>( sec - static_cast<int>(sec) ));  // already normed
  return tmp;
}


// operator add [TAI]
// ----------
t_gtime t_gtime::operator+(const double& sec) const
{
  t_gtime tmp(*this);
  tmp.add_sec( + static_cast<int>(sec));
  tmp.add_dsec(+ static_cast<double>( sec - static_cast<int>(sec) ));  // already normed
  return tmp;
}


// operator comp [TAI]
// ----------
bool t_gtime::operator<(const t_gtime& t) const
{
  return (    ( mjd(false)  < t.mjd(false) )
           || ( mjd(false) == t.mjd(false) && sod(false)  <  t.sod(false) )
	   || ( mjd(false) == t.mjd(false) && sod(false) ==  t.sod(false) && dsec(false) < t.dsec(false) )
	 );
//  return ( ( ( mjd(false)*86400 + sod(false) == t.mjd(false)*86400 + t.sod(false) ) && ( dsec(false) < t.dsec(false) ) )
//           || ( mjd(false)*86400 + sod(false) <  t.mjd(false)*86400 + t.sod(false) ) );
}


// operator comp [TAI]
// ----------
bool t_gtime::operator>(const t_gtime& t) const
{
  return (    ( mjd(false)  > t.mjd(false) )
           || ( mjd(false) == t.mjd(false) && sod(false)  >  t.sod(false) )
	   || ( mjd(false) == t.mjd(false) && sod(false) ==  t.sod(false) && dsec(false) > t.dsec(false) )
	 );
}


// operator equiv [TAI]
// ----------
bool t_gtime::operator==(const t_gtime& t) const	        
{
    return ( mjd(false) == t.mjd(false) && sod(false) == t.sod(false) && dsec(false) == t.dsec(false) );
//  return ( mjd(false)*86400 + sod(false) == t.mjd(false)*86400 + t.sod(false) && dsec(false) == t.dsec(false) );   
}

// operator not equiv [TAI]
// ----------
bool t_gtime::operator!=(const t_gtime& t) const	        
{
    return ( mjd(false) != t.mjd(false) && sod(false) != t.sod(false) && dsec(false) != t.dsec(false) );
//  return ( mjd(false)*86400 + sod(false) == t.mjd(false)*86400 + t.sod(false) && dsec(false) == t.dsec(false) );   
}

// operator =
// ----------
t_gtime t_gtime::operator=(const t_gtime& t)
{
  _mjd  = t.mjd(false);
  _sod  = t.sod(false);
  _dsec = t.dsec(false);
  tsys( t.tsys() );

  return *this;
}

// add seconds
// ----------
void t_gtime::add_sec(const int& sec)
{
  _sod += sec;
  _norm_sod();
//  cout << "adding sec = " << sec << " this = " << this->str("%Y-%m-%d %H:%M:%S[%T] %v ") << endl;
}


// add dseconds
// ----------
void t_gtime::add_dsec(const double& dsec)
{
  _dsec += dsec;
  _norm_dsec();
  _norm_sod();
}


// OBSOLETE, because via operator - exists
// time difference (this - t) [s]
// ----------
double t_gtime::diff(const t_gtime& t) const
{   
  bool b = false; // false = compare in TAI
   
  return (      mjd(b)*86400.0 +   sod(b)  +  dsec(b)
	        - t.mjd(b)*86400.0 - t.sod(b) - t.dsec(b) );
}


// -----------------------------------------------------------------------------------
// private functions
// -----------------------------------------------------------------------------------

// convert to TAI from requested time
// ----------
void t_gtime::_to_tai(const t_tsys& ts)
{
  _norm_dsec();
  _norm_sod();
  switch(ts){
   case USER  : break;;
   case TAI   : break;;
   case UTC   : _sod += leapsec(); break;;
   case GPS   : _sod += TAI_GPS;   break;;
   case LOC   : _sod += leapsec(); _norm_sod();
                _sod -= tzdiff();  _norm_sod();
                _sod -= dstime();  break;;
   default    : break;;
  }
  _norm_dsec();
  _norm_sod();
  _tsys = ts; // always set TS
}


// convert to tsys and return
// -----------
t_gtime t_gtime::_tai_to(const t_tsys& ts)
{
  switch(ts){
   case USER  : break;;
   case TAI   : break;;
   case UTC   : _sod -= leapsec(); break;;
   case GPS   : _sod -= TAI_GPS;   break;;
   case LOC   : _sod -= leapsec(); _norm_sod();
                _sod += tzdiff();  _norm_sod();
                _sod += dstime();  break;;
   default    : break;;
  }
  _norm_dsec();
  _norm_sod();
  _tsys = ts; // always set new TS

  return *this;
}


// normalize dsec
// ----------
void t_gtime::_norm_dsec()
{
  while( _dsec > 1 ){ _dsec -= 1; _sod += 1; }
  while( _dsec < 0 ){ _dsec += 1; _sod -= 1; }
}


// normalize sod
// ----------
void t_gtime::_norm_sod()
{
  while( _sod >= 86400 ){ _sod -= 86400; _mjd += 1; }
  while( _sod <  0     ){ _sod += 86400; _mjd -= 1; }
}


// normalize year
// ----------
void t_gtime::_norm_year(int& year) const
{
  if( year < 100.0 ) year += year < 80.0 ? 2000.0 : 1900.0;
}


// Year,Mon,Day -> MJD
// ----------
int t_gtime::_ymd_mjd(const int& yr, const int& mn, const int& dd) const
{
  int year(yr);
  int mon(mn);
  _norm_year(year);

  double mjd = 0.0;
  if( mon <= 2 ){ mon = mon + 12; year = year - 1; }
   
  mjd  = 365.25*year - fmod(365.25*year, 1.0) - 679006.0;
  mjd += floor(30.6001*(mon+1)) + 2 - floor(year/100) + floor(year/400) + dd;
  return static_cast<int>(mjd);
}
	      
/*	      
// initialize leap seconds
// ----------
void t_gtime::_leapseconds()
{		 
 static struct _leap[MAXLEAP] =
 { 
  { 31, 12, 1971, 11 },
  { 31, 12, 1972, 12 },
  { 31, 12, 1973, 13 },
  { 31, 12, 1974, 14 },
  { 31, 12, 1975, 15 },
  { 31, 12, 1976, 16 },
  { 31, 12, 1977, 17 },
  { 31, 12, 1978, 18 },
  { 31, 12, 1979, 19 },
  { 30, 06, 1981, 20 },
  { 30, 06, 1982, 21 },
  { 30, 06, 1983, 22 },
  { 30, 06, 1985, 23 },
  { 31, 12, 1987, 24 },
  { 31, 12, 1989, 25 },
  { 31, 12, 1990, 26 },
  { 30, 06, 1992, 27 },
  { 30, 06, 1993, 28 },
  { 30, 06, 1994, 29 },
  { 31, 12, 1995, 30 },
  { 30, 06, 1997, 31 },
  { 31, 12, 1998, 32 },
  { 31, 12, 2005, 33 },
  { 31, 12, 2008, 34 }

}
*/
